#ifndef BS_SCORED_SEQ_H
#define BS_SCORED_SEQ_H

#include "ScoredSeq.h"

class BSScoredSeq : public ScoredSeq {

	public:
		BSScoredSeq() :ScoredSeq() {}
	
		BSScoredSeq(string seq, double as, unsigned long pos, bool us) 
			: ScoredSeq(seq,as,pos,us) {}
		
		BSScoredSeq(const BSScoredSeq &other) : ScoredSeq(other) {}
		
		BSScoredSeq& operator =(const BSScoredSeq &other) {
			Init(other);
			return *this;
		}
		
		virtual ~BSScoredSeq() {}
		
		/*
		 * print_full_w_score will print the full score of this sequence, returning
		 * the sequence information with its output.
		 * It will NOT modify the Genome object.
		 * 
		 * @return A string representing the Sequence
		 */
		virtual string print_full(double denom, Genome &gen, Read &read) const;
		
		/*
		 * print_and_score will print the full score of this sequence, returning the
		 * sequence information for its output.
		 * This function WILL add its value to the Genome object
		 * 
		 * @return A string representing the Sequence
		 */
		// For now, we don't want to override this
		//string print_and_score(double denom, Genome &gen) const;
		
		/*
		 * score will add the log-odds score (adjusted by denom) to each
		 * position on the Genome gen.  Additionally, this can be passed
		 * a flag to specify the number of positions after the start position
		 * to add to (for adding to different lengths of sequences).
		 */
		virtual void score(double, Genome &, unsigned int, Read &, pthread_mutex_t&);
		
	private:
		inline string GetConsensus(const Read &read) const ;
		
		/*
		 * Because this is used to produce the consensus sequence, we need ambiguity characters...
		 */
		inline char max_char(const float* chr) const;
};

#endif

