#ifndef HMM_SCORED_SEQ_H
#define HMM_SCORED_SEQ_H

#include "ScoredSeq.h"
#include "Defaults.h"
#include "SafeVector.h"
#include "ProbabilisticModel.h"

class HMMScoredSeq : public ScoredSeq {
	
	public:
		HMMScoredSeq() :ScoredSeq() { Init(); }
	
		HMMScoredSeq(string seq, double as, unsigned long pos, bool us) 
			: ScoredSeq(seq,as,pos,us) { Init(); }
		
		HMMScoredSeq(const ScoredSeq &other) : ScoredSeq(other) { Init(); }
		
		HMMScoredSeq& operator =(const ScoredSeq &other) {
			ScoredSeq::Init(other);
			return *this;
		}

		void Init();
		
		virtual ~HMMScoredSeq() {
			if(model)
				delete model;
		}
	
		/*
		 * print_full_w_score will print the full score of this sequence, returning
		 * the sequence information with its output.
		 * It will NOT modify the Genome object.
		 * 
		 * @return A string representing the Sequence
		 */
		virtual string print_full(double denom, Genome &gen, Read &read) const;
		
		/*
		 * print_and_score will print the full score of this sequence, returning the
		 * sequence information for its output.
		 * This function WILL add its value to the Genome object
		 * 
		 * @return A string representing the Sequence
		 */
		//virtual string print_and_score(double denom, Genome &gen) const;
		
		/*
		 * score will add the log-odds score (adjusted by denom) to each
		 * position on the Genome gen.  Additionally, this can be passed
		 * a flag to specify the number of positions after the start position
		 * to add to (for adding to different lengths of sequences).
		 */
		virtual void score(double, Genome &, unsigned int, Read &, pthread_mutex_t&);
		
	private:

		const static int NumInsertStates = 2;

		ProbabilisticModel *model;

		inline string GetConsensus(const Read &read) const {

			string seq = "";
		
			for(unsigned int i=0; i<read.length; i++) {
				seq += max_char(read.pwm[i]);
			}
		
			return seq;
		}
		
		/*
		 * Because this is used to produce the consensus sequence, we need ambiguity characters...
		 */
		inline char max_char(const float* chr) const {
			if((chr[0] == chr[1]) && (chr[0] == chr[2]) && (chr[0] == chr[3]))
				return 'n';
			if(chr[0] >= chr[1]) {
				if(chr[0] >= chr[2]) {
					if(chr[0] >= chr[3])
						return 'a';
					else
						return 't';
					}
				else {
					if(chr[2] >= chr[3])
						return 'g';
					else
						return 't';
				}
			}
			else {
				if(chr[1] >= chr[2]) {
					if(chr[1] >= chr[3])
						return 'c';
					else
						return 't';
				}
				else {
					if(chr[2] >= chr[3])
						return 'g';
					else return 't';
				}
			}
		
			return 'n';
		}


};

#endif
