#ifndef VECTOR_H
#define VECOTR_H

#include <cassert>
#include <cstdlib>
#include <execinfo.h>
#include "const_include.h"
#include "Exception.h"

#define GROW_SIZE	1.5

template <typename TYPE>
class gvector {
	public:
		gvector<TYPE>():
			_cur_pos(0), _max_size(DEFAULT_SIZE), _long(false) {
			_vector = NULL;
			Init(DEFAULT_SIZE);
		}
		
		gvector<TYPE>(unsigned long size) :
			_cur_pos(0), _max_size(size), _long(false) {
			//_vector = new MAP_TYPE[size];
			_vector = NULL;
			Init(size);
		}

		~gvector<TYPE>(){
			if(_vector)
				//delete[] _vector;
				free(_vector);
		}

		TYPE& operator[](unsigned long v) {
			// Just throw an error.  You shouldn't overpass the bounds
			if(v > _max_size) {
				void* bt[10];
				size_t size = backtrace(bt,10);
				fprintf(stderr, "Error: overstepped bounds: %lu vs %lu\n",v,_max_size);
				backtrace_symbols_fd(bt, size, 2);
			}
			assert(v <= _max_size);

			//if(hash_spot > _max_size)
			//	resize();
		
			if(_long) {

			}
			return _vector[v];
		}

		void set_size(unsigned long new_size) {
			resize(new_size);
		}
		
		unsigned long size() {
			return _cur_pos;
		}

		void clear() {
			Init(DEFAULT_SIZE);
		}

		/**
		 * The push_back() function appends val to the end of the vector.
		 */
		void push_back(const TYPE &e) {

			if(_cur_pos+1 > _max_size)
				resize();

			_vector[_cur_pos++] = e;
		}

		/**
		 * The pop_back method removes the last element of the vector
		 */
		void pop_back() {
			_cur_pos--;
		}

		void resize(unsigned long new_size) {
			//fprintf(stderr,"Resizing Vector to %lu elements\n",new_size);
			
			// Allocate a new array
			TYPE* temp_vec = (TYPE*)malloc(sizeof(TYPE) * new_size);

			// Means it didn't malloc--it's too big
			if(!temp_vec) {
				//_long = true;
				//unsigned long inc_size = MAX_LONG;
				printf("ERROR: Cannot handle a vector of size %lu\n",new_size);
				throw new Exception("Could not allocate array");
			}

			//TYPE* hash_temp = new TYPE[new_size];
			// Copy from the old one
			for(unsigned long i=0; i<_max_size && i<new_size; i++)
				temp_vec[i] = _vector[i];
				
			// Delete the old one
			if(_vector)
				//delete[] _vector;
				free(_vector);

			// Swap the two
			_vector = temp_vec;
			_max_size = new_size;
		}

		unsigned long max_size() {
			return _max_size;
		}

	private:

		/**
		 * The resize method will resize it based on pre-defined conditions
		 */
		inline void resize() {
			// If the max size is below 2G, we'll double it
			if(_max_size < ONE_G) {
				resize((unsigned long)(_max_size * GROW_SIZE));
			}
			// Otherwise, we'll only increment by 1M
			else {
				resize(_max_size+ONE_G);
			}
		}

		void Init(unsigned long size) {
			_cur_pos = 0;
			_max_size = size;
			if(_vector)
				free(_vector);

			_vector = (TYPE*)malloc(sizeof(TYPE) * size);
			//_vector = new TYPE[size];
		}
	
		TYPE* _vector;
		unsigned long _cur_pos;
		unsigned long _max_size;

		bool _long;

		static const unsigned long DEFAULT_SIZE=1024;

		static const unsigned int ONE_K=1024;
		static const unsigned int ONE_M=ONE_K*1024;
		static const unsigned long ONE_G=ONE_M*1024;

		static const unsigned long MAX_DOUBLE_SIZE=1024*1024*1024;
		static const unsigned long MAX_DOUBLE_INC=1024*1024*1024;
		static const unsigned int MAX_LONG=~0;
};


#endif //VECTOR_H
