#include "BSScoredSeq.h"

string BSScoredSeq::print_full(double denom, Genome &gen, Read &read) const {
	if(!positions.size())	//if there were no matches, just return.
		return "";

	ostringstream temp;
	
	double total_score = log_align_score / denom;
	
	//we only need to compute this alignment string once because, while there may be many locations for each
	//sequence, we're ensuring that the sequences are the exact same.

	bin_seq bs;
	unsigned int seq_size = sequence.size();

	string read_consensus = GetConsensus(read);

	string gen_string = gen.GetString((*positions.begin()).first, seq_size);
	string aligned = bs.get_align_score_w_traceback(read, read_consensus, gen_string).first;

	//temp << align_score << "\t" << log_align_score << "\t" << positions.size() << "\t" << total_score << "\t" << aligned << "\t";
	temp << align_score << "\t" << positions.size() << "\t" << total_score << "\t" << aligned;
		
	set<pair<unsigned long,bool> >::iterator sit;	
	for(sit = positions.begin(); sit != positions.end(); sit++)
		temp << "\t" << gen.GetPos(*sit);
		
	return temp.str();
}

void BSScoredSeq::score(double denom, Genome &gen, unsigned int len, Read &read, pthread_mutex_t &lock) {

	if(!positions.size())	//if there were no matches, just return.
		return;

	double total_score = log_align_score / denom;
	//set<pair<unsigned int,bool> >::iterator sit;

	//we only need to compute this alignment string once because, while there may be many locations for each
	//sequence, we're ensuring that the sequences are the exact same.

	bin_seq bs;	// global one
	unsigned int seq_size = sequence.size();

	string read_consensus = GetConsensus(read);

	string gen_string = gen.GetString((*positions.begin()).first, seq_size);
	string aligned = bs.get_align_score_w_traceback(read, read_consensus, gen_string).first;

	set<pair<unsigned long,bool> >::iterator sit;

	pthread_mutex_lock(&lock);
	for(sit=positions.begin(); sit!=positions.end(); sit++) {
		for(unsigned int i=0; i<aligned.size(); i++) {
			// We only need to mark these down if there's a 'c' in the genome
			if(gen_string[i] == 'c') {
				gen.AddGenScore((*sit).first+i,total_score,'c');
				
				gen.AddSeqScore((*sit).first+i,total_score,aligned[i]);
				
			}
		}
	}
	pthread_mutex_unlock(&lock);
}

inline string BSScoredSeq::GetConsensus(const Read &read) const {
	string seq = "";

	for(unsigned int i=0; i<read.length; i++) {
		seq += max_char(read.pwm[i]);
	}

	return seq;
}

/*
 * Because this is used to produce the consensus sequence, we need ambiguity characters...
 */
//inline char max_char(vector<double> &chr) {
inline char BSScoredSeq::max_char(const float* chr) const {
	if((chr[0] == chr[1]) && (chr[0] == chr[2]) && (chr[0] == chr[3]))
		return 'n';
	if(chr[0] >= chr[1]) {
		if(chr[0] >= chr[2]) {
			if(chr[0] >= chr[3])
				return 'a';
			else
				return 't';
			}
		else {
			if(chr[2] >= chr[3])
				return 'g';
			else
				return 't';
		}
	}
	else {
		if(chr[1] >= chr[2]) {
			if(chr[1] >= chr[3])
				return 'c';
			else
				return 't';
		}
		else {
			if(chr[2] >= chr[3])
				return 'g';
			else return 't';
		}
	}

	return 'n';
}

