#ifdef	MPI_RUN
#include <mpi.h>
#endif

#ifdef OMP_RUN
#include <omp.h>
#endif

#include <iostream>
#include <unistd.h>
#include <stdio.h>
#include <signal.h>
#include <vector>
#include <string>
#include <sstream>
#include <cstring>
#include <ctime>
#include <sys/time.h>
#include <sys/resource.h>
#include <cmath>
#include <set>
#include <map>
#include <pthread.h>
#include <algorithm>
#include <exception>
#include "const_include.h"
#include "GenomeMem.h"
#include "bin_seq.h"
#include "SeqManager.h"
#include "ScoredSeq.h"
#include "BSScoredSeq.h"
#include "NormalScoredSeq.h"
#include "SNPScoredSeq.h"
#include "SequenceOperations.h"
//#include "HMMScoredSeq.h"
#include "Exception.h"

unsigned int num_nw[8] = {0,0,0,0,0,0,0,0};


typedef set<ScoredSeq*,ScoredSeq > seq_set;
typedef map<string,ScoredSeq*> seq_map;


unsigned int gBISULFITE = 0;
unsigned int gATOG = 0; 
unsigned int gSNP = 0;
unsigned char gINT2BASE[16];
unsigned char g_bs_CONVERSION[256];
unsigned char g_gen_CONVERSION[256];

int gVERBOSE=0;
int gMER_SIZE=0;
unsigned int gMAX_MATCHES = 100;
unsigned int gREAD_BUFFER = 1024 * 5;
unsigned int gBUFFER_SIZE = 100000;
unsigned int gHASH_SIZE = 1048576;	//gHASH_SIZE is set as 4^10
//unsigned int gMAX_HASH_SIZE = 1000;
unsigned int gMAX_HASH_SIZE = 0;
float gALIGN_SCORE = 10;

float gGAP = -2;
float gMATCH = 1;
float gMISMATCH = -1;

int gMAX_GAP = 3;
const char* seq_file = 0;
const char* genome_file = 0;
const char* output_file = "gnumap_out";
ofstream of;
const char* pos_matrix = NULL;
char* g_adaptor = NULL;
int gPRINT_FULL=0;
float gALIGN_SCORES[256][4];
bool perc = false;
int gSEQ_LENGTH=-1;

bool fast = false;	// Do we perform a fast alignment or a more thorough one?

float gMIN_ADAPTOR_DIFF = 0.75;	//used to define the minimum difference from the given g_adaptor sequence to the consensus
unsigned int gMIN_CHOPPED_BASES = 4;	//if there's less than 4 bases that match the g_adaptor, we're not going to chop them.
int gGEN_SKIP = 0;	// The number of bases we skip as we hash the genome
unsigned int gGEN_SIZE = 8;	// The number of bases each GEN segment contains (per each unsigned int)
unsigned int gJUMP_SIZE = 0;	// The number of bases we jump in a read (per hash)

int gSGREX = false;

bool gILLUMINA = false;

//I'm not sure what BIN_SIZE means...It's not used anymore--only appearance is in Genome.cpp:365
//int BIN_SIZE = 10;
int gBIN_SIZE = 1;
 
// gSAVE and gREAD are global variables used to indicate whether the hashed genome should be
// read and saved or created from fasta files.
bool gSAVE=false;
bool gREAD=false;
char* gSAVE_FN;
char* gREAD_FN;

int bad_files =0;

void printPWM(Read* read);

#define READ_TOO_SHORT	-1
#define READ_TOO_POOR	-2
#define READ_TOO_MANY	999999
#define SAME_DIFF		0.0001

/*****************************************************
 * These next few are all for multi-threading
 *****************************************************/
unsigned int gNUM_THREADS = 1;
volatile bool setup_complete = false;	// global signal for complete setup
volatile bool finished = false;			// global signal for finishing everything

pthread_attr_t attr;
pthread_mutex_t lock;
pthread_mutex_t nwlock;
pthread_mutex_t cond_lock;
pthread_mutex_t comm_barrier_lock;
pthread_cond_t comm_barrier_cond = PTHREAD_COND_INITIALIZER;
volatile unsigned int cond_count = 0;
volatile unsigned int cond_thread_num;
//pthread_mutex_t read_lock;

// The writer thread
pthread_t writer;
pthread_mutex_t write_lock;

void* parallel_thread_run(void* t_opts);
struct thread_rets* omp_thread_run(struct thread_opts*);
void* mpi_thread_run(void* t_opts);
void getMoreReads(unsigned int &begin, unsigned int &end);
const unsigned int nReads = 1024*10;
unsigned int nReadPos;
unsigned int nReadPerProc;
Read* gReadArray[nReads];
seq_map gReadLocs[nReads];

// This is for making MPI work
bool gMPI_LARGEMEM = false;

double gReadDenominator[nReads];
double gTopReadScore[nReads];
vector<TopReadOutput> gTopReadOutput;

int iproc=0,nproc=1;

#ifdef MPI_RUN
MPI::Status status;
#endif

Genome gGen;
SeqManager* gSM;

/*****************************************************
 * These next functions are used in parsing command-
 * line arguments
 *****************************************************/
int ParseCmdLine(int argc, char* argv[]);
void GetParseError(ostream &os, char* argv[]);
int set_arg_ext(const char* param, const char* assign, int &count);
int set_arg(const char* param, const char* assign, int &count);
/*struct CMDLine {
	int errno;
	int errpos;
} this_cmd;*/
CMDLine this_cmd;
/*****************************************************/

void usage(int has_error, char* errmessage) {
	if(has_error) cout << endl << errmessage << endl ;

	cout << "Usage: gnumap [options] <file_to_parse>\n"
		 << "  -g, --genome=STRING          Genome .fa file(s)\n"
		 << "  -o, --output=STRING          Output file\n"
		 << "  -a, --align_score=DOUBLE     Limit for sequence alignment (default: 10)\n"
		 << "  -p, --percent=NONE           Use percentage when determining alignment score\n"
		 << "  -v, --verbose=INT            Verbose (default=0)\n"
		 << "  -c, --num_proc=INT           Number of processors on which to run\n"
		 << "  -m, --mer_size=INT           Mer size (default=10)\n"
		 << "  -B, --buffer=INT             Buffer size\n"
		 << "  -T, --max_match=INT          Maximum number of matches for a given\n"
		 << "                               sequence (default: 100)\n"
		 << "  -h, --max_hash=INT           Maximum values to store in the hash.\n"
		 << "                               (default: none )\n"
		 << "  -S, --subst_file=STRING      Position-Weight Matrix file for DNA\n"
		 << "                               Substitutions\n"
		 << "  -G, --gap_penalty=DOUBLE     Gap Penalty for Substitution Matrix\n"
		 << "  -M, --max_gap=INT            Maximum Number of Gaps to use in Alignment\n"
		 << "  -A, --adaptor=STRING         Adaptor sequence to remove from sequences\n"
		 << "  -0, --print_full             Print locations for the entire sequence, not\n"
		 << "                               just for the beginning.\n"
		 << "  -b, --bs_seq                 Flag to turn on the C to T conversion, used in\n"
		 << "                               bisulfite sequence analysis\n"
		 << "  -d, --a_to_g                 Flag that allows for A to G conversion\n"
		 << "  --fast                       Perform a fast alignment (at some reduction\n"
		 << "                               of accuracy)\n"
		 << "  -s, --gen_skip=INT           Number of bases to skip when the genome is aligned\n"
		 << "                               (default: none)\n"
		 << "  --bin_size                   The resolution for GNUMAP (default: 8)\n"
		 << "  -j, --jump                   The number of bases to jump in the sequence hashing\n"
		 << "                               (default: mer_size)\n"
		 << "  --snp                        Turn on SNP mapping (will output a .sgrex file)\n"
		 << "  --illumina                   Defines the fastq file as Illumina file (otherwise\n"
		 << "                               does nothing)\n"
		 << "\n"
		 << "For MPI usage:\n"
		 << "  --MPI_largemem               If the run requires a large amount of memory, this\n"
		 << "                               flag will spread it accross several nodes.\n"
		 << "                               (default:  not included)\n"
		 << "\n"
		 << "For reading and writing to a binary file:\n"
		 << "  --save=FILENAME              Save the genome out to a file\n"
		 << "  --read=FILENAME              Read the genome in from a file\n"
		 << endl
		 << "Help options:\n"
		 << "  -?, --help                   Show this help message\n";
	exit(1);
}

/* Return the current time in seconds, using a float precision number. */
double When() {
	struct timeval tp;
	gettimeofday(&tp, NULL);
	return ((double) tp.tv_sec + (double) tp.tv_usec * 1e-6);
}

/*
struct GenomeLocation {
	GenomeLocation() {
		amount = 0.0;
		packed_char = 0;
	}

	GenomeLocation(GEN_TYPE pc, float amt)
		: amount(amt), packed_char(pc) {}

	float amount;
	GEN_TYPE packed_char;
};
*/

/************************************
 * MPI Structures and DataTypes 	*
 ************************************/
 #ifdef MPI_RUN
 /**
  * This is an MPI function for performing a reduce with user-defined data types.
  *
  * @param in One of the vectors of length /len/
  * @param inout The other vector to sum, which will contain the final result as well.
  * 		Also length /len/.
  * @param len The length of /in/ and /inout/
  * @param dptr The MPI Datatype of this function (not used?)
  */
 void sumGen( GenomeLocation *in, GenomeLocation *inout, int *len, MPI_Datatype *dptr ) {
 	int i;
 	
 	for(i=0; i<*len; i++) {
 		inout->amount = in->amount+inout->amount;
		
 		++in; ++inout;
 	}
 }
 
 #endif

/**
 * We have a loss of precision with some log additions. So we'll just add in the log space
 * and go from there.
 */
/*
inline float LogSum(float a, float b) {
	if(a > b && (a != 0)
		return a + log(1+exp(b-a));
	else
		return b + log(1 + exp(a-b));
}
*/
/**
 * This is the same as the above function, only it requires the input to be the powers
 * of e.  This way, we don't need to take the log of anything.
 */
/*inline float ExpLogSum(float expA, float expB) {
	
}
*/

/**
 * Because this is used to produce the consensus sequence, we need ambiguity characters...
 */
//inline char max_char(vector<double> &chr) {
inline char max_char(float* chr) {
	if((chr[0] == chr[1]) && (chr[0] == chr[2]) && (chr[0] == chr[3]))
		return 'n';
	if(chr[0] >= chr[1]) {
		if(chr[0] >= chr[2]) {
			if(chr[0] >= chr[3])
				return 'a';
			else
				return 't';
			}
		else {
			if(chr[2] >= chr[3])
				return 'g';
			else
				return 't';
		}
	}
	else {
		if(chr[1] >= chr[2]) {
			if(chr[1] >= chr[3])
				return 'c';
			else
				return 't';
		}
		else {
			if(chr[2] >= chr[3])
				return 'g';
			else return 't';
		}
	}

	return 'n';
}
 
//inline string GetConsensus(vector<vector<double> > &read) {
inline string GetConsensus(Read &read) {
	string seq = "";

	for(unsigned int i=0; i<read.length; i++) {
		seq += max_char(read.pwm[i]);
	}

	return seq;
}

//I don't think this function is even used...
//vector<vector<double> > subvector(vector<vector<double> > &vec, int start, int length) {
float** subvector(float** &vec, int start, int length) {
	//vector<vector<double> > to_return;
	float** to_return = new float*[length];

	for(int i=start,j=0; i<start+length; i++,j++) {
		to_return[j] = new float[4];
		for(int k=0; k<4; k++) {
			to_return[j][k] = vec[i][k];
		}
		delete[] vec[i];
	}
	
	delete[] vec;	//we want to free the memory for vec because we're creating a new
					//one, right??
	
	return to_return;
}

struct thread_opts {
	thread_opts(Genome &g, SeqManager &s, int ti) :
		gen(g), sm(s), thread_id(ti) {}
	
	Genome& gen;
	SeqManager &sm;
	//vector<vector<double> >* seqs;
	unsigned int num_seqs;
	int thread_id;
};

struct thread_rets {
	ostringstream* of;
	unsigned int good_seqs;
	unsigned int bad_seqs;
};


void clearMapAt(unsigned int index) {
	if(gReadLocs[index].empty()) {
		return;
	}
		
	seq_map::iterator sit;
	
	for(sit = gReadLocs[index].begin(); sit != gReadLocs[index].end(); ++sit) { 
		if((*sit).second)
			delete (*sit).second;
	}
	gReadLocs[index].clear();
}

bool align_sequence(Genome &gen, seq_map &unique, const Read &search, const string &consensus, 
				double min_align_score, double &denominator, double &top_align_score,
				bool upstrand) {

	bin_seq bs;
	
	unsigned int i,j,last_hash=search.length-gMER_SIZE;

	for(i=0; i< last_hash; i+=gJUMP_SIZE) {
		HashLocation* hashes = NULL;
		pair<int, unsigned long> p_hash;
		string consensus_piece;

		// If we've run accross a hash location that is clear (possibly because we've
		// cleared it in a previous step of the algorithm), we want to get another hash
		// that's right next to it instead of jumping over everything.
		for(j=0; j+i<last_hash; j++) {
			consensus_piece = consensus.substr(i+j,gMER_SIZE);
			p_hash = bs.get_hash(consensus_piece);
			hashes = gen.GetMatches(p_hash.second);
			
			// p_hash.first would only be zero if there were n's in the sequence.
			// Just step to a location that doesn't (?) have any n's.
			if(!p_hash.first)  {
#ifdef DEBUG
				if(gVERBOSE > 2)
					cout << "Invalid hash sequence at step " << j+i << endl;
#endif
				continue;
			}

			if(hashes->size)	// we've found a valid location
				break;
#ifdef DEBUG
			if(gVERBOSE > 2)
				cout << "At location " << j+i << ", no valid hashes found." << endl;
#endif
		}
		if(!hashes->size) {	// There are no valid locations for this specific hash
#ifdef DEBUG
			if(gVERBOSE > 2)
				cout << "No valid hashes for this sequence." << endl;
#endif
			break;	// Go to the next location
		}


#ifdef DEBUG
		cout << "Found " << hashes->size << " hashes for this string." << endl;
#endif
		
		for(unsigned int vit = 0; vit<hashes->size; ++vit) {
			//Match the sequence to a sequence that has two characters before and after.
			//string to_match = gen.GetString((*vit)-i-2, search.size()+4);
			
			unsigned long beginning = (hashes->hash_arr[vit]<=i) ? 0 : (hashes->hash_arr[vit]-i);
			
			// Check to see if we've already checked this location
			//if(gen.getSetThreadID(beginning,pthread_self()))
				//continue;
			
			string to_match = gen.GetString(beginning, search.length);
			
			// QUINN: Build up the string array
			
			if(to_match.size() == 0)	//means it's on a chromosome boundary--not valid sequence
				continue;

			double align_score = 0;
			char aligned[to_match.length()+1];
			strcpy(aligned,to_match.c_str());

			//if(gBISULFITE)
				//align_score = bs.get_align_score_w_traceback(search,to_match,aligned);
			//else
			
			// Here's a tester thing...
#ifdef OMP_RUN
			num_nw[omp_get_thread_num()]++;
#endif
				align_score = bs.get_align_score(search,to_match,i,i+gMER_SIZE-1);
				//align_score = bs.get_align_score(search,to_match);

#ifdef DEBUG
			if(gVERBOSE > 1) 
				cerr << "matching: " << to_match
					 << "  with  : " << consensus << "\tat pos   " << beginning 
					 							  << "\tand step " << i
												  << "\tand alignment " << align_score 
												  << "\tand min " << min_align_score << endl;
#endif
					 			
			if(align_score > top_align_score)
				top_align_score = align_score;
				
			if(align_score >= min_align_score) {
				//if(align_score >= (this_align_score + 4.6))
					//this_align_score = align_score-2;
				
				ScoredSeq* temp;
				
				if(gSNP)
					temp = 
						new SNPScoredSeq(to_match, align_score, beginning, upstrand);
						//new HMMScoredSeq(to_match,align_score, beginning, upstrand);
				else if(gBISULFITE)
					temp =
						new BSScoredSeq(to_match,align_score, beginning, upstrand);
				else
					temp =
						new NormalScoredSeq(to_match,align_score, beginning, upstrand);
						
				if(!upstrand)
					to_match = reverse_comp(to_match);
#ifdef DEBUG
				fprintf(stderr,"[%s%c] just added %lu at jump %u with size %u\n",search.name,upstrand ? '+' : '-',beginning,i,hashes->size);
#endif

				pair<seq_map::iterator,bool> it_bool = unique.insert(pair<string,ScoredSeq*>(to_match,temp));

				if(!it_bool.second) {	//If it wasn't inserted into the set--because 
										//	it had already been inserted...
					
					//int c = unique.count(*it_bool.first);
					//cout << "count: " << c << endl;
					//if(c > 1)
						//cout << "count: " << c << " " << (*it_bool.first)->print() << endl;
					delete temp;
					if( (*it_bool.first).second->add_spot(beginning,upstrand) )
						denominator += exp(align_score);
				}
				else
					denominator += exp(align_score);
				
			}
		}
		
		if(unique.size() > gMAX_MATCHES) {
			return false;
		}
	
	}
	
	return true;
}

//string print_top_matches(Genome &gen, vector<vector<double> > &search, string &consensus) {
//string print_top_matches(Genome &gen, Read &search, string &consensus, 
//			unsigned int &num_not_matched, unsigned int &num_matched) {
void set_top_matches(Genome &gen, unsigned int rIndex, string &consensus, 
			unsigned int &num_not_matched, unsigned int &num_matched) {

	bin_seq bs;
	double min_align_score = 0;
	double top_align_score = 0;
	double denominator = 0;
	

	Read* search = gReadArray[rIndex];
	seq_map* unique = &gReadLocs[rIndex];

	// Too short to align
	if(search->length < (unsigned int)gMER_SIZE) { 
		// Here's where we'll need to do something funky in the output
		num_not_matched++;
		gReadDenominator[rIndex] = 0;
		gTopReadScore[rIndex] = READ_TOO_SHORT;
		clearMapAt(rIndex);
		return;
	}


	// We need to do the length-1 because it will go to the base at that position.
	double max_align_score = 
		bs.get_align_score(*search,consensus,0,search->length-1);

	
#ifdef DEBUG
	if(gVERBOSE > 2) {
		cout << "\nNext Read:" << endl;
		cout << consensus << "\t" << max_align_score << "\t";
	}
#endif

	// Too low of quality to align
	if(max_align_score < 0) { 
		// Here's where we'll need to do something funky in the output

		num_not_matched++;
		gReadDenominator[rIndex] = 0;
		gTopReadScore[rIndex] = READ_TOO_POOR;
		clearMapAt(rIndex);
		return;
	}

	if(perc)
		min_align_score = gALIGN_SCORE * max_align_score;
	else
		min_align_score = gALIGN_SCORE;
		

	bool aligned = align_sequence(gen, *unique, *search, consensus, 
								min_align_score, denominator, top_align_score,
								true);
	if(!aligned) {
		clearMapAt(rIndex);
		
		num_matched++;
		
		gReadDenominator[rIndex] = 0;
		gTopReadScore[rIndex] = READ_TOO_MANY;
		return;
	}


	/************************************************************/
	/* We need to do the reverse of this as well...				*/
	/************************************************************/
	string rev_consensus = reverse_comp(consensus);	
	// Copy it so we don't need to reverse it again
	float** rev_pwm = reverse_comp_cpy(search->pwm,search->length);
	Read rc(rev_pwm,search->length);
	rc.name = search->name;


	//printPWM(&search);
#ifdef DEBUG
	if(gVERBOSE > 2) {
		cout << "\tPerforming reverse compliment alignment with sequence " << consensus << endl;
	}
#endif
	

	aligned = align_sequence(gen, *unique, rc, rev_consensus, 
							min_align_score, denominator, top_align_score,
							false);

	// Clean up the copy
	for(unsigned int i=0; i<rc.length; i++) {
		delete[] rev_pwm[i];
	}
	delete[] rev_pwm;

#ifdef DEBUG
	if(gVERBOSE > 2) {
		cout << "\tFound " << gReadLocs[rIndex].size() << " matches for this sequence\n";
	}
#endif
	if(!aligned) {
		// Too many sequences
		clearMapAt(rIndex);
		num_matched++;
		
		gReadDenominator[rIndex] = 0;
		gTopReadScore[rIndex] = READ_TOO_MANY;
		
		return;
	}

	if(gReadLocs[rIndex].size() == 0) {
		num_not_matched++;
		gReadDenominator[rIndex] = 0;
		gTopReadScore[rIndex] = 0;
		return;
	}

	//fprintf(stderr,"Index %d has %u elements\n",rIndex,gReadLocs[rIndex].size());
	num_matched++;
	gReadDenominator[rIndex] = denominator;
	//fprintf(stderr,"Index %d has denominator of %f\n",rIndex,denominator);
	gTopReadScore[rIndex] = top_align_score;
	return;
}


void create_match_output(Genome &gen, unsigned int rIndex, string &consensus) {

	if(gReadLocs[rIndex].size() == 0)
		return;
		
	//fprintf(stderr,"[%s] has %d matches\n",gReadArray[rIndex]->name,gReadLocs[rIndex].size());
	
	//to_return << endl;
	//to_return << "Denominator: " << denominator << endl;
	//seq_set::iterator sit;
	seq_map::iterator sit;

	ScoredSeq *max = new NormalScoredSeq;	// But empty...

	unsigned int len = gPRINT_FULL ? consensus.length() : 1;
	
	for(sit = gReadLocs[rIndex].begin(); sit != gReadLocs[rIndex].end(); ++sit) {

		(*sit).second->score(gReadDenominator[rIndex],gGen,len,*gReadArray[rIndex],lock);

		// Free up the record
		if( (*(*sit).second).is_greater(*max)) {
//		if( (*(*sit).second).is_greater(*max, gReadDenominator[rIndex])) {
			//fprintf(stderr,"cur(%f) is greater than max(%f)\n",(*(*sit).second).get_score(),max->get_score());
			ScoredSeq *temp = max;
			max = (*sit).second;
			delete temp;
		}
		else 
			delete (*sit).second;
	}	
	//fprintf(stderr,"MAX IS %f and size is %d and DENOM is %f\n",max->get_score(), gReadLocs[rIndex].size(), gReadDenominator[rIndex]);
	
	// We just want to clear any spurious elements out of it
	gReadLocs[rIndex].clear();
	
	// Only do this if I have one of the top reads...
	// we've got rounding error here (apparently)...
	if(max->get_score() > gTopReadScore[rIndex]-SAME_DIFF 
			&& gTopReadScore[rIndex] != READ_TOO_SHORT
			&& gTopReadScore[rIndex] != READ_TOO_POOR 
			&& gTopReadScore[rIndex] != READ_TOO_MANY) {
			
		vector<TopReadOutput> out_vec = 
				max->get_SAM(gReadDenominator[rIndex],*gReadArray[rIndex],consensus,gen,rIndex);
	
		pthread_mutex_lock(&write_lock);
		vector<TopReadOutput>::iterator vit = out_vec.begin();
		for(; vit != out_vec.end(); vit++) {
			//fprintf(stderr,"Pushing back...\n");
			gTopReadOutput.push_back(*vit);
		}
		pthread_mutex_unlock(&write_lock);
	}
	//else
	//	fprintf(stderr,"Failed here: %f vs %f\n",max->get_score(),gTopReadScore[rIndex]);
	
	if(max->get_score() > gTopReadScore[rIndex]) {
		fprintf(stderr,"GOTTA PROBLEM: %f vs %f...\n\n",max->get_score(),gTopReadScore[rIndex]);
	}
	
	delete max;
	
}

/**
 * This function will print the corresponding strings for all locations at this hash
 */
string print(Genome &gen, HashLocation* pos, long begin, int size) {
	string to_return = "";
	
	//cout << "matches: " << pos.size() << endl;
	for(unsigned long i=0; i<pos->size; i++) {
		to_return += gen.GetString(pos->hash_arr[i] - begin, size) + "\t";
	}
	
	return to_return;
}


/* GetPWM is used for getting a PWM matrix representing the match (m) and mismatch (mm)
 * penalties.
 * The input is a file, which should be a 4x4 matrix of the format:
 * 				 A   C   G   T
 * 			A    m  mm  mm  mm
 *			C   mm   m  mm  mm
 * 			G   mm  mm   m  mm
 * 			T   mm  mm  mm   m
 * For ease in reading, the letters defining the rows and columns are disregarded.
 *
 *
 * Note:  GetPWM will modify the global gALIGN_SCORES variable.
 */
void getPWM(const char* fn) {
	float temp_ALIGN_SCORES[5][4];

	int THIS_BUFFER = 100;
	
	ifstream in;
	in.open(fn);
	char temp_chr[THIS_BUFFER];
	bool contains_labels=false;
	float a,c,g,t;
	char label[THIS_BUFFER];
	char alabel[THIS_BUFFER];
	char clabel[THIS_BUFFER];
	int count=0;
	int num_lines=4;

	// Get all the lines of input
	while(in.getline(temp_chr,THIS_BUFFER) && count < num_lines) {
		if(sscanf(temp_chr," %s %s %s %s",alabel,clabel,label,label) == 4 
				&& (tolower(*alabel) == 'a')
				&& (tolower(*clabel) == 'c') ) {
			if(gVERBOSE > 1)
				printf("Matched here: %s\n",temp_chr);
			in.getline(temp_chr,THIS_BUFFER);
			contains_labels=true;
		}
		if(gVERBOSE > 1)
			cout << "Line: " << temp_chr << endl;

		if(contains_labels) {
			if(sscanf(temp_chr,"%s %f %f %f %f",label,&a,&c,&g,&t) != 5) {
				char* error = new char[THIS_BUFFER];
				strcat(error,"Error in Score File: ");
				strcat(error,temp_chr);
				throw(error);
			}
		}
		else {
			if(sscanf(temp_chr,"%f %f %f %f",&a,&c,&g,&t) != 4) {
				char* error = new char[THIS_BUFFER];
				strcat(error,"Error in Score File: ");
				strcat(error,temp_chr);
				throw(error);
			}
		}
		
		temp_ALIGN_SCORES[count][0] = (float)a;
		temp_ALIGN_SCORES[count][1] = (float)c;
		temp_ALIGN_SCORES[count][2] = (float)g;
		temp_ALIGN_SCORES[count][3] = (float)t;

		count++;
	}
	if(count < num_lines) {
		throw new Exception("Error in Score File:  Not enough lines");
	}

	cout << "Count: " << count << "\tNum lines: " << num_lines << endl;

#ifdef DEBUG
	//for testing purposes...
	if(gVERBOSE > 1) {
		cout << "Matrix: " << endl;
		cout << "\t0\t1\t2\t3" << endl;
		for(int i=0; i<5; i++) {
			cout << i << "\t";
			for(int j=0; j<4; j++) 
				cout << temp_ALIGN_SCORES[i][j] << "\t";
			cout << endl;
		}
		cout << endl;
		throw new Exception("Passed.");
	}
#endif

	for(unsigned int i=0; i<4; i++) {
		gALIGN_SCORES[(unsigned int)'a'][i] = temp_ALIGN_SCORES[0][i];
		gALIGN_SCORES[(unsigned int)'c'][i] = temp_ALIGN_SCORES[1][i];
		gALIGN_SCORES[(unsigned int)'g'][i] = temp_ALIGN_SCORES[2][i];
		gALIGN_SCORES[(unsigned int)'t'][i] = temp_ALIGN_SCORES[3][i];
	}
}

void printPWM(Read* read) {
	printf("Name: %s\n",read->name);
	printf("Length: %d\n",read->length);
	for(unsigned int i=0; i<read->length; i++) {
		printf("%f %f %f %f\n",read->pwm[i][0],read->pwm[i][1],read->pwm[i][2],read->pwm[i][3]);
	}
}

void sig_handler(int signum) {
	fprintf(stderr,"Just caught signal %d\n",signum);

	// Print the backtrace
	void* bt[10];
	size_t size = backtrace(bt,10);
	backtrace_symbols_fd(bt, size, 2);

	exit(0);
}

int main(int argc, char* argv[]) {

#ifdef MPI_RUN
	MPI::Init(argc, argv);
	//MPI_Comm_size(MPI_COMM_WORLD, &nproc);
	nproc = MPI::COMM_WORLD.Get_size();
	//MPI_Comm_rank(MPI_COMM_WORLD, &iproc);
	iproc = MPI::COMM_WORLD.Get_rank();
	fprintf(stderr,"iproc:%d nproc:%d\n",iproc,nproc);
#endif

	double time_prog_start = When();

	// We want to install a sig handler because this is throwing a bus error somewhere...
	struct sigaction bus_action;
	bus_action.sa_handler = sig_handler;
	sigaction(SIGBUS, &bus_action, NULL);

	//Print command line args
	cerr << "This is GNUMAP, Version 2.0.0, for public and private use." << endl;

	memset(g_gen_CONVERSION,4,256);
	g_gen_CONVERSION[(int)'a'] = g_gen_CONVERSION[(int)'A'] = 0;
	g_gen_CONVERSION[(int)'c'] = g_gen_CONVERSION[(int)'C'] = 1;
	g_gen_CONVERSION[(int)'g'] = g_gen_CONVERSION[(int)'G'] = 2;
	g_gen_CONVERSION[(int)'t'] = g_gen_CONVERSION[(int)'T'] = 3;
	g_gen_CONVERSION[(int)'n'] = g_gen_CONVERSION[(int)'N'] = 4;
	g_gen_CONVERSION[(int)'\n'] = g_gen_CONVERSION[(int)'\r'] = 5;
	// Strange white-space characters
	g_gen_CONVERSION[10] = g_gen_CONVERSION[11] = g_gen_CONVERSION[12] = g_gen_CONVERSION[13] = 5;
	g_gen_CONVERSION[(int)'\0'] = 6;
	g_gen_CONVERSION[(int)'>'] = 7;

	memset(g_bs_CONVERSION,4,256);
	g_bs_CONVERSION[(unsigned int)'a'] = g_bs_CONVERSION[(unsigned int)'A'] = 0;
	g_bs_CONVERSION[(unsigned int)'c'] = g_bs_CONVERSION[(unsigned int)'C'] = 1;
	g_bs_CONVERSION[(unsigned int)'g'] = g_bs_CONVERSION[(unsigned int)'G'] = 2;
	g_bs_CONVERSION[(unsigned int)'t'] = g_bs_CONVERSION[(unsigned int)'T'] = 3;
	g_bs_CONVERSION[(unsigned int)'n'] = g_bs_CONVERSION[(unsigned int)'N'] = 4;
	g_bs_CONVERSION[(unsigned int)'\n'] = g_bs_CONVERSION[(unsigned int)'\r'] = 5;
	// Strange white-space characters
	g_bs_CONVERSION[10] = g_bs_CONVERSION[11] = g_bs_CONVERSION[12] = g_bs_CONVERSION[13] = 5;
	g_bs_CONVERSION[(unsigned int)'\0'] = 6;
	g_bs_CONVERSION[(unsigned int)'>'] = 7;

	memset(gINT2BASE,'?',16);
	gINT2BASE[0] = 'a';
	gINT2BASE[1] = 'c';
	gINT2BASE[2] = 'g';
	gINT2BASE[3] = 't';
	gINT2BASE[4] = 'n';


#ifdef DEBUG
	printf("int\tbs\tgen\n");
	for(int i=0; i<256; i++) {
		printf("%d\t%c\t%d\t%d\n",i,(char)i,g_bs_CONVERSION[i],g_gen_CONVERSION[i]);
	}
	printf("int2base\n");
	for(int i=0; i<16; i++) {
		printf("%d\t%c\n",i,gINT2BASE[i]);
	}
#endif

	cout << endl << "Command Line Arguments:  ";
	for(int i=0; i<argc; i++)
		cout << argv[i] << ' ';
	cout << endl;
	
	gVERBOSE = 0;

	if(argc == 1) //means only the bin/gnumap parameter
		usage(0,"");

	if(argc < 2) {
		usage(1, "Need at least a genome and a file to map.");
		exit(1);
	}

	int rc = ParseCmdLine(argc,argv);
	// an error occurred during option processing
	if(rc != 0) {
		GetParseError(cerr, argv);
		return -1;
	}


	//set up alignment matrix

	for(unsigned int i=0; i<256; i++)
		for(unsigned int j=0; j<4; j++)
			gALIGN_SCORES[i][j] = gGAP;
	gALIGN_SCORES[(int)'a'][0] = gMATCH;
	gALIGN_SCORES[(int)'a'][1] = gMISMATCH;
	gALIGN_SCORES[(int)'a'][2] = gMISMATCH;
	gALIGN_SCORES[(int)'a'][3] = gMISMATCH;

	gALIGN_SCORES[(int)'c'][1] = gMATCH;
	gALIGN_SCORES[(int)'c'][0] = gMISMATCH;
	gALIGN_SCORES[(int)'c'][2] = gMISMATCH;
	gALIGN_SCORES[(int)'c'][3] = gMISMATCH;

	gALIGN_SCORES[(int)'g'][2] = gMATCH;
	gALIGN_SCORES[(int)'g'][0] = gMISMATCH;
	gALIGN_SCORES[(int)'g'][1] = gMISMATCH;
	gALIGN_SCORES[(int)'g'][3] = gMISMATCH;

	gALIGN_SCORES[(int)'t'][3] = gMATCH;
	gALIGN_SCORES[(int)'t'][0] = gMISMATCH;
	gALIGN_SCORES[(int)'t'][1] = gMISMATCH;
	gALIGN_SCORES[(int)'t'][2] = gMISMATCH;

/*
	for(int i=0; i<4; i++) {
		for(int j=0; j<4; j++) {
			if(i == j)
				gALIGN_SCORES[i][j] = MATCH;
			else
				gALIGN_SCORES[i][j] = MISMATCH;
		}
		gALIGN_SCORES[4][i] = gGAP;
	}
*/

	/* Manage errors that might occur */	
	// If they don't include a fasta file for sequences
	if( (seq_file == NULL) )
		usage(1, "Specify a single file e.g., sequences.fa\n");
	if(!genome_file)
		usage(1, "Specify a genome to map to with the -g flag\n");

	if(pos_matrix != NULL) {
		try {
			getPWM(pos_matrix);
		}
		catch(const char* err) {
			cerr << "ERROR: \n\t" << err << endl;
			return -1;
		}
		catch(Exception *e) {
			cerr << "Error: \n\t" << e->GetMessage() << endl;
			delete e;
			return -1;
		}
	}
	
	ostringstream params;
	params << "# Parameters: " << endl;
	params << "#\tVerbose: " << gVERBOSE << endl;
	if(gREAD)
		params << "#\tGenome file to read from: " << gREAD_FN << endl;
	else 
		params << "#\tGenome file(s): " << genome_file << endl;
	if(gSAVE)
		params << "#\tGenome file to save to: " << gSAVE_FN << endl;
	params << "#\tOutput file: " << output_file << endl;
	params << "#\tSequence file: " << seq_file << endl;
	if(!perc)
		params << "#\tAlign score: " << gALIGN_SCORE << endl;
	else
		params << "#\tAlign percentage: " << gALIGN_SCORE*100 << "%" << endl;
	params << "#\tNumber of threads: " << gNUM_THREADS << endl;
	if(fast) {
		if(!gMER_SIZE)	//if the user didn't specify one
			gMER_SIZE = 14;
		params << "#\tUsing FAST alignment mode" << endl;
	}
	if(gGEN_SKIP != 0) {
		if(gGEN_SKIP < 0)
			usage(1, "-s: Invalid Genome Skip size\n");
		params << "#\tSkipping " << gGEN_SKIP << " bases during hash step" << endl;
	}
	gGEN_SKIP++;	// We're mod'ing in the function, so we want to mod by GEN_SKIP+1
	if(gMAX_HASH_SIZE > 0)
		params << "#\tLargest Hash Size: " << gMAX_HASH_SIZE << endl;
	if(!gMER_SIZE)	//if the user didn't specify one
		gMER_SIZE = 10;
	params << "#\tMer size: " << gMER_SIZE << endl;
	if(gJUMP_SIZE) {
		if(gJUMP_SIZE < 1)
			usage(1, "--jump: Invalid jump size\n");
	}
	else {
		if(fast)
			gJUMP_SIZE = gMER_SIZE;
		else
			gJUMP_SIZE = gMER_SIZE/2;
	}
	params << "#\tUsing jump size of " << gJUMP_SIZE << endl;
	if(pos_matrix != NULL)
		params << "#\tUsing User-Defined Alignment Scores" << endl;
	else 
		params << "#\tUsing Default Alignment Scores" << endl;
	params << "#\tGap score: " << gGAP << endl;
	params << "#\tMaximum Gaps: " << gMAX_GAP << endl;
	if(g_adaptor) {
		params << "#\tUsing Adaptor Sequence: " << g_adaptor << endl;
	}
	if(gBISULFITE) {
		gPRINT_FULL = true;
		params << "#\tUsing BISULFITE conversion" << endl;
		g_bs_CONVERSION[(int)'c'] = g_bs_CONVERSION[(int)'C'] = 3;	//change it so every 'c' will be *hashed* as a 't'
		//gALIGN_SCORES[3][1] = MATCH;
		gALIGN_SCORES[(int)'c'][3] = gMATCH;
	}
	if(gATOG) {
		gPRINT_FULL = true;
		params << "#\tUsing A to G conversion" << endl;
		g_bs_CONVERSION[(int)'a'] = g_bs_CONVERSION[(int)'A'] = 2;	//change it so every 'a' will be *hashed* as a 'g'
		//gALIGN_SCORES[2][0] = MATCH;
		gALIGN_SCORES[(int)'a'][2] = gMATCH;
	}
	if(gGEN_SIZE != 8) {
		if(gGEN_SIZE > 8)
			usage(1,"Invalid bin size (must be 8 or less)\n");
		params << "#\tUsing irregular bin size of " << gGEN_SIZE << endl;
	}

	//determine the size of the hash
	gHASH_SIZE = (unsigned int) pow(4.0,gMER_SIZE);

#ifdef DEBUG
	printf("Alignment scores:\n");
	printf("\tA\tC\tG\tT\n");
	printf("A\t%f\t%f\t%f\t%f\n",gALIGN_SCORES[(int)'a'][0],gALIGN_SCORES[(int)'a'][1],
			gALIGN_SCORES[(int)'a'][2],gALIGN_SCORES[(int)'a'][3]);
	printf("C\t%f\t%f\t%f\t%f\n",gALIGN_SCORES[(int)'c'][0],gALIGN_SCORES[(int)'c'][1],
			gALIGN_SCORES[(int)'c'][2],gALIGN_SCORES[(int)'c'][3]);
	printf("G\t%f\t%f\t%f\t%f\n",gALIGN_SCORES[(int)'g'][0],gALIGN_SCORES[(int)'g'][1],
			gALIGN_SCORES[(int)'g'][2],gALIGN_SCORES[(int)'g'][3]);
	printf("T\t%f\t%f\t%f\t%f\n",gALIGN_SCORES[(int)'t'][0],gALIGN_SCORES[(int)'t'][1],
			gALIGN_SCORES[(int)'t'][2],gALIGN_SCORES[(int)'t'][3]);
	printf("N\t%f\t%f\t%f\t%f\n",gALIGN_SCORES[(int)'n'][0],gALIGN_SCORES[(int)'n'][1],
			gALIGN_SCORES[(int)'n'][2],gALIGN_SCORES[(int)'n'][3]);
#endif

	if(gVERBOSE) {
		cerr << params.str() << endl;
	}
	
	// This works for MPI and for everything else
	if(nproc > 1) {
		char* output_ext = new char[strlen(output_file)+4];
		sprintf(output_ext,"%s_%d",output_file,iproc);
		output_file = output_ext;
	}
	// If we're not doing the large memory version
	// or if we are and this is the head node
	if(!gMPI_LARGEMEM || iproc == 0) {
		of.open(output_file);
		// Check to make sure the output file/directory exists
		if(of.bad() || of.fail()) {
			cerr << "ERROR:\n\t";
			perror("Error in reading output file");
			return -1;
		}

		// only write the params if you're the first node		
		if(iproc == 0)
			of << params.str();
	}

	if(gVERBOSE)
		cerr << "Hashing the genome." << endl;
	// Use the global genome: gGen
	//Genome gen; 
	unsigned long my_start, my_end;
	
	try {

		unsigned long gen_size=0;
		
		if(gMPI_LARGEMEM && iproc == 0) {
			Genome notherGen;
			// we don't care about these parameters.  Just read the whole thing
			notherGen.use(genome_file);
			gen_size = notherGen.count();
	//		fprintf(stderr,"Size of genome here: %lu\n",gen_size);
		}
		
		my_start = 0;
		my_end = gen_size;

#ifdef MPI_RUN
		// tell each node what part they need to get
		//MPI_Broadcast((char*)&my_start,sizeof(my_start),MPI_CHAR,
		MPI::COMM_WORLD.Bcast(&gen_size, 1, MPI_UNSIGNED_LONG, 0);

		// If we want to have large memory MPI, we should have declared the flag.
		if(gMPI_LARGEMEM) {
			my_start = gen_size/nproc * iproc;
			my_end = my_start + gen_size/nproc;
		}
		//return 0;
		
#endif
		//fprintf(stderr,"[%d/%d] gen_size=%lu, my_start=%lu, my_end=%lu\n",iproc,nproc,gen_size,my_start,my_end);

		gGen.use(genome_file, my_start, my_end);
		gGen.LoadGenome();

	}
	catch(const char* err) {
		cerr << "ERROR: \n\t" << err << endl;
		return -1;
	}
	catch(bad_alloc&) {
		perror("ERROR HERE");
		cerr << "ERROR: \n\tUnable to allocate enough memory.\n\t(Pick a smaller mer size?)" << endl;
		return -1;
	}
	catch(exception &e) {
		cerr << "ERROR: \n\t" << e.what() << endl;
		return -1;
	}
	catch(Exception *e) {
		fprintf(stderr,"ERROR: \n\t%s\n",e->GetMessage());
	}
	catch(...) {
		cerr << "ERROR: \n\tUnknown error." << endl;
		return -1;
	}

	if(gVERBOSE > 1) 
		cerr << "\nTime to hash: " << (1.0*clock())/CLOCKS_PER_SEC << endl;			
	
	if(iproc == 0)
		// Output SAM by default
		of << "# <QNAME>\t<FLAG>\t<RNAME>\t<POS>\t<MAPQ>\t<CIGAR>\t<MRNM>\t<MPOS>\t<ISIZE>\t<SEQ>\t<QUAL>\n";

	if(gMPI_LARGEMEM) {
		gSM = new SeqManager(gReadArray, seq_file, nReads, gNUM_THREADS);	
		//gSM = &sm;
	}
	else {
		// Pass the SeqManager the number of nodes in the whole machine as well as what node we are
		gSM = new SeqManager(gReadArray, seq_file, nReads, gNUM_THREADS, nproc, iproc);	
	}
	

	/************************************
	 * Do the threading					*
	 ************************************/
	
	pthread_t pthread_hand[gNUM_THREADS];
	void *pthread_ret[gNUM_THREADS];
	pthread_mutex_init(&lock,NULL);
	pthread_mutex_init(&nwlock,NULL);
	pthread_mutex_init(&write_lock,NULL);
	pthread_mutex_init(&cond_lock,NULL);
	pthread_mutex_init(&comm_barrier_lock,NULL);
	
	
	//Here is where we set the threading information.
	pthread_setconcurrency(gNUM_THREADS);
	pthread_attr_init(&attr);
	pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);

	//An array of pointers so we can delete them when we're done.
	thread_opts* t_o_ptr[gNUM_THREADS];
	
#ifdef OMP_RUN
	cond_thread_num = 1;
	omp_set_num_threads(gNUM_THREADS);
	thread_opts* t_o = new thread_opts(gGen,*gSM,0);
	thread_rets* ret = omp_thread_run(t_o);

	unsigned int seqs_matched = 0, seqs_not_matched = 0;
	seqs_matched = ret->good_seqs;
	seqs_not_matched = ret->bad_seqs;

#else
	for(unsigned int i=0; i<gNUM_THREADS; i++) {

		//The parameters to pass to the thread
		thread_opts* t_o = new thread_opts(gGen,*gSM,i);
		t_o_ptr[i] = t_o;

#ifdef MPI_RUN
		if(gMPI_LARGEMEM) 
			pthread_create(&pthread_hand[i], NULL, mpi_thread_run, (void*)t_o);
		else
			pthread_create(&pthread_hand[i], NULL, parallel_thread_run, (void*)t_o);

#else
		pthread_create(&pthread_hand[i], NULL, parallel_thread_run, (void*)t_o);
#endif
	}



	/************************************************************************/
	/* Here is where we'll begin the process of matching and writing to the	*/
	/* output file.															*/
	/************************************************************************/	
	unsigned int seqs_matched = 0;
	unsigned int seqs_not_matched = 0;
	cond_thread_num = gNUM_THREADS;
	setup_complete = true;

	
	
	/********************************************************/
	/* wait for all the threads to join back together... 	*/
	/********************************************************/
	for (unsigned int i=0; i<gNUM_THREADS; i++) {
		pthread_join(pthread_hand[i], &pthread_ret[i]);
	}
	
	/********************************************************/
	/* clean up the memory leaks stuff					 	*/
	/********************************************************/
	for(unsigned int i=0; i<gNUM_THREADS; i++) {
		//delete[] seq_ptr[i];
		delete t_o_ptr[i];
	}
	
	for(unsigned int i=0; i<gNUM_THREADS; i++) {
		seqs_matched += ((thread_rets*)pthread_ret[i])->good_seqs;
		seqs_not_matched += ((thread_rets*)pthread_ret[i])->bad_seqs;

		delete (thread_rets*)pthread_ret[i];
	}
#endif
	
	if(gVERBOSE) 
		cerr << endl;
		
	
	if(gVERBOSE)
		cerr << "\n[" << iproc << "/-] Printing output." << endl;

#ifdef MPI_RUN
	// If it's not the largemem (where we each print out our genome), we need to reduce
	if(!gMPI_LARGEMEM) {
			
		fprintf(stderr,"[%d/-] Sending genomes...",iproc);
		float* gen_ptr = gGen.GetGenomeAmtPtr();
		unsigned int gen_size = (unsigned int)gGen.size()/gGEN_SIZE;
	
		float* sendGenome = (float*)malloc(sizeof(float)*gen_size);	
		if(!sendGenome)
			fprintf(stderr,"ERROR!!! Could not create genome of size %lu bytes!!!\n",sizeof(float)*gen_size);

		
		memcpy(sendGenome,gen_ptr,(size_t)(gen_size*sizeof(float)));
		MPI::COMM_WORLD.Reduce(sendGenome, gen_ptr, gen_size, MPI::FLOAT, MPI_SUM, 0);
		if(gSNP) {
			if(gVERBOSE > 1) fprintf(stderr,"A...");
			// Do the read A's now
			gen_ptr = gGen.GetGenomeAPtr();
			memcpy(sendGenome,gen_ptr,gen_size*sizeof(float));
			MPI::COMM_WORLD.Reduce(sendGenome, gen_ptr, gen_size, MPI::FLOAT, MPI_SUM, 0);
			if(gVERBOSE > 1) fprintf(stderr,"C...");
			// C's
			gen_ptr = gGen.GetGenomeCPtr();
			memcpy(sendGenome,gen_ptr,gen_size*sizeof(float));
			MPI::COMM_WORLD.Reduce(sendGenome, gen_ptr, gen_size, MPI::FLOAT, MPI_SUM, 0);
			if(gVERBOSE > 1) fprintf(stderr,"G...");
			// G's
			gen_ptr = gGen.GetGenomeGPtr();
			memcpy(sendGenome,gen_ptr,gen_size*sizeof(float));
			MPI::COMM_WORLD.Reduce(sendGenome, gen_ptr, gen_size, MPI::FLOAT, MPI_SUM, 0);
			if(gVERBOSE > 1) fprintf(stderr,"T...");
			// T's
			gen_ptr = gGen.GetGenomeTPtr();
			memcpy(sendGenome,gen_ptr,gen_size*sizeof(float));
			MPI::COMM_WORLD.Reduce(sendGenome, gen_ptr, gen_size, MPI::FLOAT, MPI_SUM, 0);
			if(gVERBOSE > 1) fprintf(stderr,"N...");
			// N's
			gen_ptr = gGen.GetGenomeNPtr();
			memcpy(sendGenome,gen_ptr,gen_size*sizeof(float));
			MPI::COMM_WORLD.Reduce(sendGenome, gen_ptr, gen_size, MPI::FLOAT, MPI_SUM, 0);
		}
		
		fprintf(stderr,"\nFinished!  Printing final .sgr/.sgrex file\n");
		// Then print out our new genome
		if(iproc == 0)
			gGen.PrintFinal(output_file);
	}
	else
		// Even if we're doing MPI, have everyone print to their (separate) output files
		gGen.PrintFinal(output_file);
		
#else
	gGen.PrintFinal(output_file);
#endif
/*
	int msg;
	if(iproc == 0) {
		//Print to the .sgr file
		fprintf(stderr,"Printing from %lu to %lu\n",my_start, my_end);
		gGen.PrintFinal(output_file);
#ifdef MPI_RUN
		MPI::COMM_WORLD.Send(&msg, 1, MPI::INT, (iproc+1)%nproc, 0);
		fprintf(stderr,"[%d/%d] Just sent out a bit to %d\n",iproc,nproc,(iproc+1)%nproc);
#endif
	}

#ifdef MPI_RUN	
	// Now, in order, we need to print out the rest of the genomes
	for(int i=1; i<nproc; i++) {
		if(iproc == i) {
			// Receive,
			fprintf(stderr,"[%d/%d] waiting for something from %d\n",iproc,nproc,i-1);
			MPI::COMM_WORLD.Recv(&msg, 1, MPI::INT, i-1, 0, status);
			fprintf(stderr,"[%d/%d] Just recvd a bit\n",iproc,nproc);

			fprintf(stderr,"Printing from %lu to %lu\n",my_start, my_end);
			
			gGen.PrintFinal(output_file);
			// Then print your data
			//gGen.AppendFinal(output_file);
			// Then send to the next node
			MPI::COMM_WORLD.Send ( &msg, 1, MPI::INT, (i+1)%nproc, 0);
		}
	}
	if(iproc == 0) {
		fprintf(stderr,"[%d/%d] waiting for something from %d\n",iproc,nproc,nproc-1);
		MPI::COMM_WORLD.Recv(&msg, 1, MPI::INT, nproc-1, 0, status);
	}
*/

#ifdef MPI_RUN
	unsigned int bseqs_matched = seqs_matched;
	unsigned int bseqs_not_matched = seqs_not_matched;
	MPI::COMM_WORLD.Reduce(&bseqs_matched, &seqs_matched, 1, MPI_UNSIGNED, MPI_SUM, 0);
	MPI::COMM_WORLD.Reduce(&bseqs_not_matched, &seqs_not_matched, 1, MPI_UNSIGNED, MPI_SUM, 0);
	
#endif

	if(iproc == 0) {
	
		double time_prog_end = When();
		
		ostringstream stats;
		stats 	<< "\n#Finished.\n";
		stats	<< "#\tTotal Time: " << time_prog_end-time_prog_start << " seconds.\n";
		stats	<< "#\tFound " << seqs_matched+seqs_not_matched << " sequences.\n";
		stats	<< "#\tSequences matched: " << seqs_matched << "\n";
		stats	<< "#\tSequences not matched: " << seqs_not_matched << "\n";
		stats	<< "#\tOutput written to " << output_file << "\n";
		
		cerr << stats.str();
		of << stats.str();		
	}

	of.close();

	if(!gMPI_LARGEMEM)
		delete gSM;
		
#ifdef MPI_RUN
	MPI::Finalize();
#endif

	return 0;
}


void comm_cond_wait() {
	// We only want to wait in here if we need to communicate anything.  Otherwise, just continue
	if(!gMPI_LARGEMEM)
		return;
		
#ifdef MPI_RUN
	pthread_mutex_lock(&cond_lock);
	if(++cond_count < cond_thread_num) {
		//fprintf(stderr,"Thread %d/%d waiting...\n",cond_count,cond_thread_num);
		pthread_cond_wait(&comm_barrier_cond, &cond_lock);
		pthread_mutex_unlock(&cond_lock);
		
	}
	else {
		//fprintf(stderr,"Thread %d/%d made it in!\n",cond_count,cond_thread_num);
		// Do your stuff

		// MPI communicate everything you need to
		
		// Do a reduce-all with a SUM on the denominators
		When();
		When();
		double start_time = When();
		double sendDenominator[nReads];
		memcpy(sendDenominator,gReadDenominator,nReads*sizeof(double));

		MPI::COMM_WORLD.Allreduce(&sendDenominator, &gReadDenominator, nReads,
					MPI::DOUBLE, MPI_SUM);
		double end_time = When();
		fprintf(stderr,"[%d] Total time for Denoms is %f\n",iproc,end_time-start_time);
		
		start_time = When();
		// Do a reduce-all with a MAX on the top reads
		double sendReads[nReads];
		memcpy(sendReads, gTopReadScore, nReads*sizeof(double));

		MPI::COMM_WORLD.Allreduce(&sendReads, &gTopReadScore, nReads,
					MPI::DOUBLE, MPI_MAX);
		end_time = When();
		fprintf(stderr,"[%d] Total time for TopReads is %f\n",iproc,end_time-start_time);

		cond_count = 0;
		pthread_mutex_unlock(&cond_lock);
		pthread_cond_broadcast(&comm_barrier_cond);
	}
#endif
}

/*
struct TopReadOutput {
	char[100] READ_NAME;
	char[100] CHR_NAME;
	unsigned long CHR_POS;
	bool strand;
	int MAPQ;
	char[100] CIGAR
	unsigned int readIndex;
};
*/


void write_cond_wait() {
	pthread_mutex_lock(&cond_lock);
	if(++cond_count < cond_thread_num) {
		//fprintf(stderr,"Thread %d/%d waiting...\n",cond_count,cond_thread_num);
		pthread_cond_wait(&comm_barrier_cond, &cond_lock);
		pthread_mutex_unlock(&cond_lock);
	}
	else {
		//fprintf(stderr,"Thread %d/%d made it in!\n",cond_count,cond_thread_num);

		if(gMPI_LARGEMEM) {

#ifdef MPI_RUN
			if(iproc == 0) {
				//fprintf(stderr,"\nSize of vector: %u\n",gTopReadOutput.size());
				unsigned int total_outputed = 0;
				
				// Do your own reads
				vector<TopReadOutput>::iterator vit;
				for(vit = gTopReadOutput.begin(); vit != gTopReadOutput.end(); vit++) {
					total_outputed++;
					string consensus = GetConsensus(*(gReadArray[(*vit).readIndex]));
					string QUAL = str2qual(*(gReadArray[(*vit).readIndex]));
					
					of << (*vit).READ_NAME << "\t";
					if((*vit).strand)
						of << 0x0010 << "\t";
					else
						of << 0x0000 << "\t";
					of << (*vit).CHR_NAME << "\t" << (*vit).CHR_POS << "\t";
					of << (*vit).MAPQ << "\t";
					of << (*vit).CIGAR << "\t";
					of << "=\t0\t0\t";
					of << consensus << "\t";
					of << QUAL.c_str() << "\n";
				}
				// Clear the output vector
				gTopReadOutput.clear();
				
				
				// The vector we'll read to
				vector<TopReadOutput> msg;
				
				// Receive from each node, writing to output
				for(int i=1; i<nproc; i++) {
	
					int msg_size;
					MPI::COMM_WORLD.Recv(&msg_size, 1, MPI::INT, i, 0, status);
					// Reserve space in the vector
					//msg.reserve(msg_size);
					
					TopReadOutput msg[msg_size];
					MPI::COMM_WORLD.Recv((unsigned char*)msg, msg_size * sizeof(TopReadOutput), 
							MPI::CHAR, i, 0, status );
	
					//fprintf(stderr,"\nSize of vector: %u\n",msg_size);
					
					// Now process the vector
					for(int j=0; j<msg_size; j++) {
						total_outputed++;
						string consensus = GetConsensus(*(gReadArray[(msg[j]).readIndex]));
						string QUAL = str2qual(*(gReadArray[(msg[j]).readIndex]));
						
						of << (msg[j]).READ_NAME << "\t";
						if((msg[j]).strand)
							of << 0x0010 << "\t";
						else
							of << 0x0000 << "\t";
						of << (msg[j]).CHR_NAME << "\t" << (msg[j]).CHR_POS << "\t";
						of << (msg[j]).MAPQ << "\t";
						of << (msg[j]).CIGAR << "\t";
						of << "=\t0\t0\t";
						of << consensus << "\t";
						of << QUAL.c_str() << "\n";
					}
				}
				
				// If there's any others we need to write to a file, write them as well.
				// If they don't match, we'll just ignore them...
				for(unsigned int i=0; i<nReads; i++) {
					if(gReadArray[i] && 
						(gTopReadScore[i] == READ_TOO_SHORT || gTopReadScore[i] == READ_TOO_POOR)) {
						total_outputed++;
						string consensus = GetConsensus(*(gReadArray[i]));
						string QUAL = str2qual(*(gReadArray[i]));
						
						of << gReadArray[i]->name << "\t";
						of << 0x0200 << "\t*\t0\t0\t*\t=\t0\t0\t";
						of << consensus << "\t";
						of << QUAL.c_str() << "\n";
					}
				}
				
				//fprintf(stderr,"Just wrote out %u reads\n",total_outputed);
			}
			else {
				//fprintf(stderr,"This is not the place to write...\n");
				// Send it to the central node to write to a file
			
				// Send the number
				int msg_size = gTopReadOutput.size();
				MPI::COMM_WORLD.Send ( &msg_size, 1, MPI::INT, 0, 0);
			
				// Send the data (cast as a character)
				MPI::COMM_WORLD.Send ( (unsigned char*)(&gTopReadOutput.front()), 
						msg_size * sizeof(TopReadOutput), 
						MPI::CHAR, 0, 0 );
						
				gTopReadOutput.clear();
			}
		}
		else {
			// Write it all to the file 
			vector<TopReadOutput>::iterator vit;
			for(vit = gTopReadOutput.begin(); vit != gTopReadOutput.end(); vit++) {
				string consensus = GetConsensus(*(gReadArray[(*vit).readIndex]));
				string QUAL = str2qual(*(gReadArray[(*vit).readIndex]));
				
				of << (*vit).READ_NAME << "\t";
				if((*vit).strand)
					of << 0x0010 << "\t";
				else
					of << 0x0000 << "\t";
				of << (*vit).CHR_NAME << "\t" << (*vit).CHR_POS << "\t";
				of << (*vit).MAPQ << "\t";
				of << (*vit).CIGAR << "\t";
				of << "=\t0\t0\t";
				of << consensus << "\t";
				of << QUAL.c_str() << "\n";
			}
			// Clear the output vector
			gTopReadOutput.clear();
			
			// If there's any others we need to write to a file, write them as well.
			// If they don't match, we'll just ignore them...
			for(unsigned int i=0; i<nReads; i++) {
				if(gReadArray[i] && 
					(gTopReadScore[i] == READ_TOO_SHORT || gTopReadScore[i] == READ_TOO_POOR)) {
					string consensus = GetConsensus(*(gReadArray[i]));
					string QUAL = str2qual(*(gReadArray[i]));
					
					of << gReadArray[i]->name << "\t";
					of << 0x0200 << "\t*\t0\t0\t*\t=\t0\t0\t";
					of << consensus << "\t";
					of << QUAL.c_str() << "\n";
				}
			}
#endif
		}
		
		cond_count = 0;
		pthread_mutex_unlock(&cond_lock);
		pthread_cond_broadcast(&comm_barrier_cond);
	}
}

#define OLD
void single_write_cond_wait(int thread_no) {
	
#ifdef OLD

	pthread_mutex_lock(&write_lock);
	if(++cond_count < cond_thread_num) {
		//fprintf(stderr,"[%d] Thread %d/%d waiting...\n",thread_no,cond_count,cond_thread_num);
		pthread_cond_wait(&comm_barrier_cond, &write_lock);
	}	
	else {
#else
	pthread_mutex_lock(&write_lock);
	if(iproc == 0) {
#endif
	
		//fprintf(stderr,"[%d] Thread %d made it in!\n",thread_no,cond_count);
		
		// Write it all to the file 
		vector<TopReadOutput>::iterator vit;
		for(vit = gTopReadOutput.begin(); vit != gTopReadOutput.end(); vit++) {
			string consensus = GetConsensus(*(gReadArray[(*vit).readIndex]));
			string QUAL = str2qual(*(gReadArray[(*vit).readIndex]));
			
			of << (*vit).READ_NAME << "\t";
			if((*vit).strand)
				of << 0x0010 << "\t";
			else
				of << 0x0000 << "\t";
			of << (*vit).CHR_NAME << "\t" << (*vit).CHR_POS << "\t";
			of << (*vit).MAPQ << "\t";
			of << (*vit).CIGAR << "\t";
			of << "=\t0\t0\t";
			of << consensus << "\t";
			of << QUAL.c_str() << "\n";
		}
		// Clear the output vector
		gTopReadOutput.clear();
		
#ifdef OLD
		cond_count = 0;
		pthread_cond_broadcast(&comm_barrier_cond);
#endif
	}
	

	// Everyone takes care of their unwritten threads
	unsigned int nPerProc = nReads/gNUM_THREADS;
	//fprintf(stderr,"[%d/%d] CREATING OUTPUT from %d to %d\n",iproc,thread_no,nPerProc*thread_no,nPerProc*(thread_no+1));
	
	// If there's any others we need to write to a file, write them as well.
	for(unsigned int i=nPerProc*thread_no; i<nPerProc*(thread_no+1); i++) {
		if(gReadArray[i] && 
				(gTopReadScore[i] == READ_TOO_SHORT || gTopReadScore[i] == READ_TOO_POOR)) {
			string consensus = GetConsensus(*(gReadArray[i]));
			string QUAL = str2qual(*(gReadArray[i]));
			
			of << gReadArray[i]->name << "\t";
			of << 0x0200 << "\t*\t0\t0\t*\t=\t0\t0\t";
			of << consensus << "\t";
			of << QUAL.c_str() << "\n";
		}
	}	
	
	pthread_mutex_unlock(&write_lock);
}


// We just want to make sure we're clean before we delete all the reads...
void clean_cond_wait(bool my_finished) {
	pthread_mutex_lock(&cond_lock);
	finished += my_finished;

	if(++cond_count < cond_thread_num) {
		pthread_cond_wait(&comm_barrier_cond, &cond_lock);
		pthread_mutex_unlock(&cond_lock);
		
	}
	else {
		//fprintf(stderr,"Exiting the output thingy\n");
		// Now that we're done, reset the global SeqManager's edit point
		gSM->resetCounter();
		
		cond_count = 0;
		pthread_cond_broadcast(&comm_barrier_cond);
		pthread_mutex_unlock(&cond_lock);
	}
}

/**
 * The thread runner that chooses MPI
 */
void* parallel_thread_run(void* t_opts) {

	//cout << "New Thread" << endl;
	unsigned int thread_id = ((thread_opts*)t_opts)->thread_id;

//	unsigned int numRPT = 128;
	unsigned int numRPT = nReads/gNUM_THREADS;
	
	unsigned int read_begin = numRPT*thread_id;
	unsigned int read_end = read_begin + numRPT;

	unsigned int good_seqs = 0;
	unsigned int bad_seqs = 0;

	unsigned int seqs_processed = 0;
	
	//SeqManager *sm = &((thread_opts*)t_opts)->sm;
	SeqManager *sm = gSM;

	bool my_finished = false;

	
	// Fill the read buffer
	while(!finished) {
		//fprintf(stdout,"[%d/%d] Thread continuing....\n",iproc,thread_id);

		//fprintf(stderr,"[%d] Getting more reads from %u to %u\n",thread_id,read_begin,read_end);
		my_finished = !sm->getMoreReads(read_begin,read_end,false);
		unsigned int num_seqs = sm->getCurrNumReads()/gNUM_THREADS;
		unsigned int progress = num_seqs/100;
		if(progress < 1)
			progress = 1;


		for(unsigned int k=read_begin; k < read_end; k++, seqs_processed++) {
			//vector<vector<double> > temp_vect = ((thread_opts*)t_opts)->seqs[k];
			Read* temp_read = gReadArray[k];
			if(!temp_read) {
				//cout << "Found end" << endl;
				//break;
				//printf("Thread %d: Found null Read at pos %u (out of %u)\n",
				//	((thread_opts*)t_opts)->thread_id,k,read_buffer);
				break;
			}
			
			/*
			if(gVERBOSE) {
				if( (seqs_processed % progress ) == 0) {
					fprintf(stderr,"\r[%d/%d] %d%% complete",iproc,thread_id,(int)(((double)seqs_processed/num_seqs)*100));
					//cerr << "\r["<<thread_id<<"] " << (int)(((double)seqs_processed/num_seqs)*100) << "% complete"; cerr.flush();
				}
			}
			*/
			string consensus = GetConsensus(*temp_read);

			set_top_matches(((thread_opts*)t_opts)->gen, k, consensus, bad_seqs, good_seqs);
			create_match_output(((thread_opts*)t_opts)->gen, k, consensus);


		}		 
			
		//write_cond_wait();
		single_write_cond_wait(thread_id);
		
		for(unsigned int k=read_begin; k<read_end; k++) {
			
			Read* temp_read = gReadArray[k];
			if(!temp_read)
				continue;
			
			//done with the read.  Delete it.
			for(unsigned int i=0; i<temp_read->length; i++) {
				delete[] temp_read->pwm[i];
			}
			delete[] temp_read->pwm;
			if(temp_read->name)
				delete[] temp_read->name;

			delete temp_read;
			gReadArray[k] = 0;
		}

		//fprintf(stdout,"[%d/%d] Before Clean cond wait and cond_thread_num=%d\n",iproc,thread_id,cond_thread_num);
		clean_cond_wait(my_finished);
		//fprintf(stdout,"[%d/%d] After Clean cond wait and cond_thread_num=%d and my_finished=%d, finished=%d\n",iproc,thread_id,cond_thread_num,my_finished,finished);
	}
		
	//fprintf(stderr,"[%d/%d] Finished after processing %d/%u reads with %u good and %u bad\n",iproc,thread_id,seqs_processed,sm->getTotalSeqCount(),good_seqs,bad_seqs);
	
	//delete[] ((thread_opts*)t_opts)->seqs;

	struct thread_rets* ret = new thread_rets;
	ret->good_seqs = good_seqs;
	ret->bad_seqs = bad_seqs;
	
	return (void*)ret;
}

/**
 * This is just a block of code, not really a function call.  We just want to put everything in here...
 */
#ifdef OMP_RUN
thread_rets* omp_thread_run(thread_opts* t_opts) {
	fprintf(stderr,"CALLING OMP_THREAD_RUN\n");

	unsigned int i,j, nReadsRead, good_seqs=0, bad_seqs=0;
	double total_time=0, comm_time=0;

#pragma omp parallel shared(t_opts,gReadArray,gReadDenominator,gTopReadScore)
{
	while(!finished) {

		fprintf(stderr,"[%d/%d] Num nw's: %u\n",iproc,omp_get_thread_num(),num_nw[omp_get_thread_num()]);
		num_nw[omp_get_thread_num()] = 0;
#pragma omp single
{
		t_opts->sm.fillReadArray(gReadArray,nReads,nReadsRead);
		fprintf(stderr,"[%d/-] Num Reads to read: %u and found %u.  Finished is %d\n",iproc,nReads,nReadsRead,finished);

}

		// Use omp to parallelize this code
#pragma omp for private(i) schedule(dynamic,128) reduction(+:good_seqs) reduction(+:bad_seqs)
		for(i=0; i<nReadsRead; i++) {
			//fprintf(stderr,"[%d:%d] Setting match output for %d\n",iproc,omp_get_thread_num(),i);
			string consensus = GetConsensus(*(gReadArray[i]));
			set_top_matches(t_opts->gen, i, consensus, bad_seqs, good_seqs);
		}

// wait for all the threads to sync
#pragma omp single
{
		double start_comm = When();
		comm_cond_wait();
		comm_time += When()-start_comm;
		fprintf(stderr,"[%d] Made it past the comm_cond_wait\n",iproc);
}

#pragma omp for private(i) schedule(dynamic,128)
		for(i=0; i<nReadsRead; i++) {
			//fprintf(stderr,"[%d:%d] Creating match output for %d\n",iproc,omp_get_thread_num(),i);
			string consensus = GetConsensus(*(gReadArray[i]));
			create_match_output(t_opts->gen, i, consensus);
		}

#pragma omp single
{
		double start = When();
		write_cond_wait();
		total_time += When()-start;
		fprintf(stderr,"[%d] Made it past the write_cond_wait\n",iproc);
}

#pragma omp for private(i,j) schedule(static)
		for(i=0; i<nReadsRead; i++) {
			if(!gReadArray[i])
				continue;

			for(j=0; j<gReadArray[i]->length; j++)
				delete[] gReadArray[i]->pwm[j];

			delete[] gReadArray[i]->pwm;
			if(gReadArray[i]->name)
				delete[] gReadArray[i]->name;

			delete gReadArray[i];
			gReadArray[i] = 0;
			gReadDenominator[i] = 0;
			gTopReadScore[i] = 0;
		}

#pragma omp single
{
		gSM->printProgress();
		fprintf(stderr,"[%d/-] Seconds used in write_comm: %f\tcomm_cond: %f\n",iproc,total_time,comm_time);
		finished = gSM->isFinished();
}


	}
}
	
	fprintf(stderr,"[%d/-] Total time used in write_comm: %f\tcomm_cond: %f\n",iproc,total_time,comm_time);
	
	struct thread_rets* ret = new thread_rets;
	ret->good_seqs = good_seqs;
	ret->bad_seqs = bad_seqs;
	return ret;
}

#endif

/**
 * The function each of the threads will run
 */
void* mpi_thread_run(void* t_opts) {
	fprintf(stderr,"CALLING MPI_THREAD_RUN\n");

	while(!setup_complete);	// Busy-Wait for the setup to complete
	
	unsigned int i,j,begin=0,end=0;
	unsigned int bad_seqs=0, good_seqs=0;
	
	bool my_finished = false;
	
	// main loop body
	while(!finished) {
		// Get your portion of the reads here
		// It's locking, so we don't need to worry about multi-entrances
		//my_finished = !((thread_opts*)t_opts)->sm.getMoreReads(begin, end, true);
		my_finished = !gSM->getMoreReads(begin, end, true);

/*
		if(finished) {
			pthread_mutex_lock(&cond_lock);
			// Decrement the number of threads that will participate in the waits
			cond_thread_num--;
			pthread_mutex_unlock(&cond_lock);
			break;
		}
*/			
		//fprintf(stderr,"Thread %u has seqs %u-%u\n",((thread_opts*)t_opts)->thread_id,begin,end);
		
		// Do the first part of the alignment
		unsigned int count = 0;
		for(i=begin; i<end; i++, count++) {
			string consensus = GetConsensus(*(gReadArray[i]));
			set_top_matches(((thread_opts*)t_opts)->gen, i, consensus, bad_seqs, good_seqs);

			// For some reason, the percentage complete doesn't work...
			//if(gVERBOSE)
			//	fprintf(stderr,"\r%d%% complete",(int)(((double)i)/gSM->getNumSeqsPProc())*100);

		}

		//fprintf(stderr,"Thread %d about to enter cond_wait\n",((thread_opts*)t_opts)->thread_id);
		// Wait for each thread to finish
		comm_cond_wait();
		//fprintf(stderr,"Thread %d just exited cond_wait\n",((thread_opts*)t_opts)->thread_id);

		for(i=begin; i<end; i++) {
				
			string consensus = GetConsensus(*(gReadArray[i]));
			create_match_output(((thread_opts*)t_opts)->gen, i, consensus);
		}

		//fprintf(stderr,"Thread %d about to enter write cond_wait\n",((thread_opts*)t_opts)->thread_id);
		// Wait for each thread to finish, then write them.
		write_cond_wait();
		
		// Delete the reads you've used
		for(i=begin; i<end; i++) {
			if(!gReadArray[i])
				continue;
			
			for(j=0; j<gReadArray[i]->length; j++) {
				delete[] gReadArray[i]->pwm[j];
			}
			delete[] gReadArray[i]->pwm;
			if(gReadArray[i]->name)
				delete[] gReadArray[i]->name;

			delete gReadArray[i];
			gReadArray[i] = 0;
			gReadDenominator[i] = 0;
			gTopReadScore[i] = 0;
			
			//clearMapAt(i);
		}
		
		// Even if we're not using MPI, we should wait here.
		clean_cond_wait(my_finished);
		//fprintf(stderr,"Thread %d finished, starting over\n",((thread_opts*)t_opts)->thread_id);
	}
	
	struct thread_rets* ret = new thread_rets;
	ret->good_seqs = good_seqs;
	ret->bad_seqs = bad_seqs;
	
	return (void*)ret;
}





int ParseCmdLine(int argc, char* argv[]) {
	
	bool found_other = false;

	for(int i=1; i<argc; i++) {
		if(*argv[i] == '-') { //check for normal
			int set_ret;

			if(*((argv[i])+1) == '-')  { //check for extended
				//printf("Found extended\n");
				set_ret = set_arg_ext(argv[i],argv[i+1],i);
			}
			else {
				set_ret = set_arg(argv[i],argv[i+1],i);
				i++;	//increment past the next expression
			}
			
			if(set_ret != 0) {
				this_cmd.errnum = set_ret;
				this_cmd.errpos = i;
				return -1;
			}

		}
		else {
			//if we've already found the the other cmd arg, it doesn't work. 
			if(found_other) {
				this_cmd.errnum = INVALID_ARG;
				this_cmd.errpos = i;
				return INVALID_ARG;
			}
			else {	//set the other arg to this pointer
				seq_file = argv[i];
			}
		}
	}

	return 0;

}

void GetParseError(ostream &os, char* argv[]) {
	if(this_cmd.errnum == PARSE_ERROR) {
		os << "Irregular Parameter in: " << argv[this_cmd.errpos] << endl;
	}
	else if(this_cmd.errnum == NO_MATCHING_ARG) {
		os << "No matching arg in: " << argv[this_cmd.errpos] << endl;
	}
	else if(this_cmd.errnum == INVALID_ARG) {
		os << "Please specify only a single sequence file, or multiple files in \"quotations\" ("
		   << argv[this_cmd.errpos] << ")" << endl;
	}
	else {
		os << "Unknown error " << this_cmd.errnum << " in: " << argv[this_cmd.errpos] << endl;
	}

}


int set_arg(const char* param, const char* assign, int &count) {
	double temp_dbl;

	switch(*(param+1)) {	//switch on the first character
		case 'g':
			genome_file = assign;
			break;
		case 'l':
			if(sscanf(assign,"%d",&gSEQ_LENGTH) < 1)
				return PARSE_ERROR;
			break;
		case 'o':
			output_file = assign;
			break;
		case 'a':
			//if(sscanf(assign,"%lf",&ALIGN_SCORE) < 1)
			if(sscanf(assign,"%lf",&temp_dbl)< 1)
				return PARSE_ERROR;
			//	return (PARSE_ERROR * (int)(*param));
			gALIGN_SCORE = (float)temp_dbl;
			break;
		case 'p':
			count--;
			perc = true;
			break;
		case 'v':
			if(sscanf(assign,"%d",&gVERBOSE) < 1)
				return PARSE_ERROR;
			break;
		case 'c':
			if(sscanf(assign,"%d",&gNUM_THREADS) < 1)
				return PARSE_ERROR;
			break;
		case 'm': 
			if(sscanf(assign,"%d",&gMER_SIZE) < 1)
				return PARSE_ERROR;
			break;
		case 'B':
			if(sscanf(assign,"%d",&gBUFFER_SIZE) < 1)
				return PARSE_ERROR;
			break;
		case 'T':
			if(sscanf(assign,"%d",&gMAX_MATCHES) < 1)
				return PARSE_ERROR;
			break;
		case 'h':
			if(sscanf(assign,"%d",&gMAX_HASH_SIZE) < 1)
				return PARSE_ERROR;
			break;
		case 'S':
			pos_matrix = assign;
			break;
		case 'G':
			//if(sscanf(assign,"%lf",&gGAP) < 1)
			if(sscanf(assign,"%lf",&temp_dbl) < 1)
				return PARSE_ERROR;
			gGAP = (float)temp_dbl;
			break;
		case 'M':
			if(sscanf(assign,"%d",&gMAX_GAP) < 1)
				return PARSE_ERROR;
			break;
		case 'A':
			g_adaptor = new char[strlen(assign)+1];
			unsigned int i;
			for(i=0; i<strlen(assign); i++)
				g_adaptor[i] = tolower((char)assign[i]);
			// Null-terminate the string
			g_adaptor[i] = '\0';
			//g_adaptor = assign;
			break;
		case '0':
			count--;
			gPRINT_FULL = 1;
			break;
		case 'b':
			count--;
			gBISULFITE = 1;
			break;
		case 'd':
			count--;
			gATOG = 1;
			break;
		case 's':
			if(sscanf(assign,"%d",&gGEN_SKIP) < 1)
				return PARSE_ERROR;
			break;
		case 'j':
			if(sscanf(assign,"%d",&gJUMP_SIZE) < 1)
				return PARSE_ERROR;
			break;
		case '?':
			usage(0,"");
		default:
			return PARSE_ERROR;
	}

	return 0;

}

enum {
	PAR_ORIG=256,
	PAR_GENOME,
	PAR_LENGTH,
	PAR_OUTPUT,
	PAR_ALIGN_SCORE,
	PAR_PERCENT,
	PAR_VERBOSE,
	PAR_NUM_PROC,
	PAR_MER_SIZE,
	PAR_BUFFER,
	PAR_MAX_MATCH,
	PAR_MAX_HASH,
	PAR_SUBST_FILE,
	PAR_GAP_PENALTY,
	PAR_MAX_GAP,
	PAR_ADAPTOR,
	PAR_PRINT_FULL,
	PAR_BS_SEQ,
	PAR_A_TO_G,
	PAR_FAST,
	PAR_GEN_SKIP,
	PAR_READ,
	PAR_SAVE,
	PAR_GEN_SIZE,
	PAR_JUMP,
	PAR_SNP,
	PAR_ILL,
	PAR_LARGEMEM
};

int set_arg_ext(const char* param, const char* assign, int &count) {
	double temp_dbl;
	int which = 0;
	int adjust=0;

	if(strncmp(param+2,"genome=",7) == 0) {	//check to see if it's the genome
		//which = 'g';
		which = PAR_GENOME;
		adjust = 8;	//the size of "genome" plus 2
	}
	else if(strncmp(param+2,"length=",7) == 0) {
		//which = 'l';
		which = PAR_LENGTH;
		adjust = 8;
	}
	else if(strncmp(param+2,"output=",7) == 0) {
		//which = 'o';
		which = PAR_OUTPUT;
		adjust = 8;
	}
	else if(strncmp(param+2,"align_score=",12) == 0) {
		//which = 'a';
		which = PAR_ALIGN_SCORE;
		adjust = 13;
	}
	else if(strncmp(param+2,"percent",7) == 0) {
		//which = 'p';
		which = PAR_PERCENT;
		adjust = 9;
	}
	else if(strncmp(param+2,"verbose=",8) == 0) {
		//which = 'v';
		which = PAR_VERBOSE;
		adjust = 9;
	}
	else if(strncmp(param+2,"num_proc=",9) == 0) {
		//which = 'c';
		which = PAR_NUM_PROC;
		adjust = 10;
	}
	else if(strncmp(param+2,"mer_size=",9) == 0) {
		//which = 'm';
		which = PAR_MER_SIZE;
		adjust = 10;
	}
	else if(strncmp(param+2,"buffer=",7) == 0) {
		//which = 'B';
		which = PAR_BUFFER;
		adjust = 8;
	}
	else if(strncmp(param+2,"max_match=",9) == 0) {
		//which = 'T';
		which = PAR_MAX_MATCH;
		adjust = 10;
	}
	else if(strncmp(param+2,"max_hash=",10) == 0) {
		//which = 'h';
		which = PAR_MAX_HASH;
		adjust = 11;
	}
	else if(strncmp(param+2,"subst_file=",9) == 0) {
		//which = 'S';
		which = PAR_SUBST_FILE;
		adjust = 10;
	}
	else if(strncmp(param+2,"gap_penalty=",12) == 0) {
		//which = 'G';
		which = PAR_GAP_PENALTY;
		adjust = 13;
	}
	else if(strncmp(param+2,"max_gap=",8) == 0) {
		//which = 'M';
		which = PAR_MAX_GAP;
		adjust = 9;
	}
	else if(strncmp(param+2,"adaptor=",8) == 0) {
		//which = 'A';
		which = PAR_ADAPTOR;
		adjust = 9;
	}
	else if(strncmp(param+2,"print_full",10) == 0) {
		//which = '0';
		which = PAR_PRINT_FULL;
		adjust = 12;
	}
	else if(strncmp(param+2,"bs_seq",6) == 0) {
		//which = 'b';
		which = PAR_BS_SEQ;
		adjust = 8;
	}
	else if(strncmp(param+2,"a_to_g",6) == 0) {
		//which = 'd';
		which = PAR_A_TO_G;
		adjust = 8;
	}
	else if(strncmp(param+2,"fast",4) == 0) {
		//which = 'f';
		which = PAR_FAST;
		adjust = 6;
	}
	else if(strncmp(param+2,"gen_skip=",9) == 0) {
		//which = 's';
		which = PAR_GEN_SKIP;
		adjust = 10;
	}
	else if(strncmp(param+2,"read=",5) == 0) {
		//which = 1;
		which = PAR_READ;
		adjust = 6;
		gREAD=true;
	}
	else if(strncmp(param+2,"save=",5) == 0) {
		//which = 2;
		which = PAR_SAVE;
		adjust = 6;
		gSAVE=true;
	}
	else if(strncmp(param+2,"bin_size=",9) == 0) {
		//which = 3;
		which = PAR_GEN_SIZE;
		adjust = 10;
	}
	else if(strncmp(param+2,"jump",4) == 0) {
		which = PAR_JUMP;
		adjust = 6;
	}
	else if(strncmp(param+2,"snp",3) == 0) {
		which = PAR_SNP;
		adjust = 5;
	}
	else if(strncmp(param+2,"illumina",8) == 0) {
		which = PAR_ILL;
		adjust = 10;
	}
	else if(strncmp(param+2,"MPI_largemem",12) == 0) {
		which = PAR_LARGEMEM;
		adjust = 14;
	}
	else if(strcmp(param+2,"help") == 0) {
		usage(0,""); //usage will exit immediately
	}
	else 
		return NO_MATCHING_ARG;

	param += ++adjust;

	switch(which) {	//switch on the first character
		//case 'g':
		case PAR_GENOME:
			genome_file = param;
			break;
		//case 'l':
		case PAR_LENGTH:
			if(sscanf(param,"%d",&gSEQ_LENGTH) < 1)
				return PARSE_ERROR;
			break;
		//case 'o':
		case PAR_OUTPUT:
			output_file = param;
			break;
		//case 'a':
		case PAR_ALIGN_SCORE:
			//if(sscanf(param,"%lf",&ALIGN_SCORE) < 1)
			if(sscanf(param,"%lf",&temp_dbl) < 1)
				return PARSE_ERROR;
			//	return (PARSE_ERROR * (int)(*param));
			gALIGN_SCORE = (float)temp_dbl;
			break;
		//case 'p':
		case PAR_PERCENT:
			//count--;
			perc = true;
			break;
		//case 'v':
		case PAR_VERBOSE:
			if(sscanf(param,"%d",&gVERBOSE) < 1)
				return PARSE_ERROR;
			break;
		//case 'c':
		case PAR_NUM_PROC:
			if(sscanf(param,"%d",&gNUM_THREADS) < 1)
				return PARSE_ERROR;
			break;
		//case 'm': 
		case PAR_MER_SIZE:
			if(sscanf(param,"%d",&gMER_SIZE) < 1)
				return PARSE_ERROR;
			break;
		//case 'B':
		case PAR_BUFFER:
			if(sscanf(param,"%d",&gBUFFER_SIZE) < 1)
				return PARSE_ERROR;
			break;
		//case 'T':
		case PAR_MAX_MATCH:
			if(sscanf(param,"%d",&gMAX_MATCHES) < 1)
				return PARSE_ERROR;
			break;
		//case 'h':
		case PAR_MAX_HASH:
			if(sscanf(param,"%d",&gMAX_HASH_SIZE) < 1)
				return PARSE_ERROR;
			break;
		//case 'S':
		case PAR_SUBST_FILE:
			pos_matrix = param;
			break;
		//case 'G':
		case PAR_GAP_PENALTY:
			//if(sscanf(param,"%lf",&gGAP) < 1)
			if(sscanf(param,"%lf",&temp_dbl) < 1)
				return PARSE_ERROR;
			gGAP = (float)temp_dbl;
			break;
		//case 'M':
		case PAR_MAX_GAP:
			if(sscanf(param,"%d",&gMAX_GAP) < 1)
				return PARSE_ERROR;
			break;
		//case 'A':
		case PAR_ADAPTOR:
			g_adaptor = new char[strlen(assign)];
			for(unsigned int i=0; i<strlen(assign); i++)
				g_adaptor[i] = tolower((char)param[i]);
			//g_adaptor = (char*)param;
			break;
		//case '0':
		case PAR_PRINT_FULL:
			//count--;	//don't decrement in the extended version
			gPRINT_FULL = 1;
			break;
		//case 'b':
		case PAR_BS_SEQ:
			//count--;
			gBISULFITE = 1;
			break;
		//case 'd':
		case PAR_A_TO_G:
			//count--;
			gATOG = 1;
			break;
		//case 'f':
		case PAR_FAST:
			//count--;
			fast = true;
			break;
		//case 's':
		case PAR_GEN_SKIP:
			if(sscanf(param,"%d",&gGEN_SKIP) < 1)
				return PARSE_ERROR;
			break;
		//case 1:	//read
		case PAR_READ:
			gREAD_FN = (char*)param;
			if(!gREAD_FN)
				return PARSE_ERROR;
			break;
		//case 2:	//save
		case PAR_SAVE:
			gSAVE_FN = (char*)param;
			if(!gSAVE_FN)
				return PARSE_ERROR;
			break;
		case PAR_GEN_SIZE:
			if(sscanf(param,"%ud",&gGEN_SIZE) < 1)
				return PARSE_ERROR;
			break;
		case PAR_JUMP: 
			if(sscanf(param,"%ud",&gJUMP_SIZE) < 1)
				return PARSE_ERROR;
			break;
		case PAR_SNP:
			gSNP = true;
			gGEN_SIZE = 1;
			gPRINT_FULL = 1;
			break;
		case PAR_ILL:
			gILLUMINA = true;
			break;
		case PAR_LARGEMEM:
			gMPI_LARGEMEM = true;
			break;
		default:
			return NO_MATCHING_ARG;
	}

	return 0;

}
