#include "GenomeMem.h"

/*
 * Constructor.  Creates new Genome with filename.
 * fn - the filename of the file to be parsed into the genome.
 */
Genome::Genome(const char* fn) :
	offset(0), bin_offset(0), packed_genome(NULL), amount_genome(NULL) {

	Init(fn);
	reader = new Reader();

}

Genome::Genome() :
	offset(0), bin_offset(0), reader(NULL), packed_genome(NULL), amount_genome(NULL) {

}

Genome::~Genome() {
	delete reader;
	if(packed_genome)
		free(packed_genome);
	if(amount_genome)
		free(amount_genome);

	if(gSNP)
        for(unsigned int i=0; i<NUM_CHAR_VALS; i++)
            delete[] reads[i];
    if(gBISULFITE) {
        for(unsigned int i=0; i<NUM_CHAR_VALS; i++)
            delete[] reads[i];

        delete[] gens[1];
    }
    if(gATOG) {
        for(unsigned int i=0; i<NUM_CHAR_VALS; i++)
            delete[] reads[i];

        delete[] gens[0];
    }
} 

void Genome::Init(const char* fn) {

	// Don't need to init names if we're going to do so in the read file
	if(!gREAD) 	{
		if(names.size()) //if we've already set up the "names" vector, just return.
			return;

		char files[strlen(fn)];
		strcpy(files,fn);
		
		const char* delims = ", \"\n";
		char* token = strtok(files,delims);
		if(gVERBOSE > 1)
			cout << "Found:" << endl;

		while(token != NULL) {
			if(gVERBOSE)
				cout << "\t" << token << endl;
			string temp_str = token;
			
			// Get rid of spaces that make it impossible to handle
			ReplaceSpaceWithUnderscore(temp_str);
			
			names.push_back(pair<string,unsigned long>(temp_str,0));
			
			token = strtok(NULL,delims);
		}

		//Push a dummy string on the end so we can know when to end...
		names.push_back(pair<string,unsigned long>("end",0));
	}

	//clear the *_spots files
	for(int i=0; i<NUM_CHAR_VALS; i++) {
		reads[i] = 0;
		gens[i] = 0;
	}

}

void Genome::use(const char* fn) {
	use(fn,0,0);
}

void Genome::use(const char* fn, unsigned long s, unsigned long e) {
	my_start = s;
	my_end = e;
	if(my_start >= my_end)
		// read everything
		read_all = true;
	else
		read_all = false;
	
	//fprintf(stderr,"Reading Genome from %lu to %lu, read_all=%d\n",s,e,read_all);
	
	if(reader != NULL) {
		cout << "non-null reader" << endl;
		delete reader;
	}
	Init(fn);
	reader = new Reader();
}

void Genome::print_to_file(hash_map<unsigned int, HashLocation> & gh, char* ofn) { 
	printf("Printing to file: %s\n",ofn);
	int SEQ_LENGTH = gMER_SIZE; //This is an archaic #def'd element that needed to be removed.
	ofstream out;
	out.open(ofn,ios_base::out);
	
	bin_seq bs;
	
	unsigned int i;
	
	for(i=0; i<gHASH_SIZE; ++i) {
		if(!gh[i].size)
			continue;
		out << i << "\t" << gh[i].size;
		unsigned int j;
		out << '\t' << bs.hash2str(i,gMER_SIZE);
		for(j=0; j<gh[i].size; j++)
			out << '\t' << gh[i].hash_arr[j] << "[" << GetPos(gh[i].hash_arr[j],true) << "] :" 
				<< GetString(gh[i].hash_arr[j],SEQ_LENGTH);
		out << endl;
	}
	
	string gen_str = GetString(0,genome_size);
	out << gen_str << endl;
	
	for(unsigned int i=0; i<names.size(); i++) {
		out << names[i].first << "\t" << names[i].second << endl;
	}
	
	out.close();
}

unsigned int Genome::readGen(char* fn) {
	FILE* readF;

	readF = fopen(fn,"r");
	
	if(!readF) {
		printf("Trying to open: >%s<\n",fn);
		perror("File wouldn't open");
		throw "Could not open Genome read file";
	}

	unsigned long temp_gen_hash_size;
	//fscanf(readF,"%ld\n",&gen_hash_size);
	fread(&temp_gen_hash_size, sizeof(temp_gen_hash_size), 1, readF);
	gHASH_SIZE = temp_gen_hash_size;
	// fprintf(stderr,"gen_hash_size: %d\n",gHASH_SIZE);
	gen_hash.set_size(gHASH_SIZE);

	// Read in the HASH_SIZE
	fread(&gMER_SIZE, sizeof(gMER_SIZE), 1, readF);
	// gMER_SIZE = 10;
	fprintf(stderr,"\tMer Size: %d\n",gMER_SIZE);
	
	// Read in the hash
	while(true) {
		unsigned int index,hash_size;
		//fscanf(readF,"%d %d\n",&index,&hash_size);
		fread(&index, sizeof(index), 1, readF);
		fread(&hash_size, sizeof(hash_size), 1, readF);
		//fprintf(stderr,"%d %d\n",index,hash_size);
		if(!hash_size)	//This is a marker for the end of the read.
			break;

		gen_hash[index].size = hash_size;
		// malloc a new array for this location
		unsigned long* temp_arr = new unsigned long[hash_size];

		for(unsigned int j=0; j<hash_size; j++) {
			unsigned long hash_spot;
			fread(&hash_spot,sizeof(unsigned long),1,readF);
			temp_arr[j] = hash_spot;
			//gen_hash[index].push_back(hash_spot);
		}
		gen_hash[index].hash_arr = temp_arr;
		if(hash_size != gen_hash[index].size)
			fprintf(stderr,"****ERROR: At location %d, %ld should be %d\n",index,gen_hash[index].size,hash_size);
	}

	// fprintf(stderr,"Finished reading in hash.  Reading in the genome now\n");
		
	// Read in the genome.
	unsigned long temp_genome_size = 0;

	GEN_TYPE former_gen_size = gGEN_SIZE;
	// The gGEN_SIZE is the number of chars stored in each GenomeLocation
	fread(&gGEN_SIZE, sizeof(gGEN_SIZE), 1, readF);
	// The size of the genome
	fread(&temp_genome_size, sizeof(temp_genome_size), 1, readF);
	if(former_gen_size != gGEN_SIZE)
		fprintf(stderr,"\tGEN_SIZE: %d\n",gGEN_SIZE);

	//genome = new GenomeLocation[temp_genome_size];
	//genome = (GenomeLocation*)malloc(sizeof(GenomeLocation) * temp_genome_size);
	packed_genome = (GEN_TYPE*)malloc(sizeof(GEN_TYPE) * temp_genome_size);
	amount_genome = (float*)malloc(sizeof(float) * temp_genome_size);

	for(unsigned long i=0; i<temp_genome_size; i++) {
		GEN_TYPE genome_char = 0;
		fread(&(genome_char),sizeof(GEN_TYPE),1,readF);
		//genome[i].packed_char = genome_char;
		//genome[i].amount = 0;
		packed_genome[i] = genome_char;
		amount_genome[i] = 0;
	}

	genome_size = temp_genome_size;

	int num_chrs = 0;
	fscanf(readF,"\n%d\n",&num_chrs);
	//fprintf(stderr,"\n%d\n",num_chrs);
	
	/* Read the chromosome names from a file */
	for(int i=0; i<num_chrs; i++) {
		char name[255];
		unsigned long chr_pos;
		fscanf(readF,"%s %ld\n",name,&chr_pos);
		//fprintf(stderr,"Name: %s Size: %ld\n",name,chr_pos);
		names.push_back(pair<string,unsigned long>(name,chr_pos));
	}

#ifdef DEBUG	
	for(unsigned int i=0; i<names.size(); i++) {
		fprintf(stderr,"names[%u] (%lu): %s\n",i,names[i].second,names[i].first.c_str());
	}
#endif

	fclose(readF);

	return 0;
}


void Genome::LoadGenome() {
	if(gREAD) {
		fprintf(stderr,"\tReading from file: %s\n",gREAD_FN);// read the file in from memory
		readGen(gREAD_FN);
		fprintf(stderr,"\tJust finished reading file %s.\n",gREAD_FN);
	}
	else
		hash_and_store();


	fix_hash();
	//print_to_file(gen_hash,"printed_hash");

#ifdef DEBUG
		for(unsigned int i=0; i<names.size(); i++) {
			printf("Name[%d]: %s\tBegin: %ld\n",i,names[i].first.c_str(),names[i].second);
		}
		cout.flush();
#endif

	//if we are doing the bisulfite sequencing, we want to record what nucleotide
	//comes at each position
	if(gSNP) {
		printf("Finding SNPs\n");
		for(unsigned int i=0; i<NUM_CHAR_VALS; i++) {
			reads[i] = new float[genome_size];
			memset(reads[i],0,genome_size*sizeof(float));
		}
	
	}
	if(gBISULFITE) {
		printf("Using bisulfite conversion techniques\n");
		for(unsigned int i=0; i<NUM_CHAR_VALS; i++) {
			reads[i] = new float[genome_size];
			memset(reads[i],0,genome_size*sizeof(float));
		}
		
		gens[1] = new float[genome_size]; //Represents a genomic 'c'
		memset(gens[1],0,genome_size*4);
		//gens[3] = new float[genome_vec.size()];
	}
	if(gATOG) {
		for(unsigned int i=0; i<NUM_CHAR_VALS; i++) {
			reads[i] = new float[genome_size];
			memset(reads[i],0,genome_size*4);
		}
		
		gens[0] = new float[genome_size];	// Represents a genomic 'a'
		memset(gens[0],0,genome_size*4);
		//gens[2] = new float[genome_vec.size()];
	}

	if(gSAVE) {
		fprintf(stderr,"\tSaving to file: %s\n",gSAVE_FN);
		saveGen(gSAVE_FN);
	}
}

void Genome::hash_and_store() {
	
	// Records what position we want to start saving at
	bool start_yet = my_start == 0 ? true : false;
	
	// Set some values that are necessary for the 'store()' function
	offset = 0;
	bin_offset = 0;

	hash_map<unsigned int, vector<unsigned long> > * gen_hash_vec
		= new hash_map<unsigned int, vector<unsigned long> >;
	//gvector<GenomeLocation> * genome_vec
	//	= new gvector<GenomeLocation>;
	gvector<GEN_TYPE> *genome_vec
		= new gvector<GEN_TYPE>;

	Reader reader_vec;
	unsigned char gen_piece[gBUFFER_SIZE+1];

	printf("Hashing Genome...\n");
	gen_hash_vec->set_size(gHASH_SIZE);

	
	bin_seq bs;	

	unsigned long counter = 0;
	unsigned long int chars_stored = 0;
	unsigned long int chr_begin_pos = 0;
	unsigned int j;

	for(j=0; j<names.size()-1; j++) {
		reader_vec.use(names[j].first);
		string chr_name = reader_vec.GetName();
		//fprintf(stderr,"Name here is: %s\n",chr_name.c_str());
		ReplaceSpaceWithUnderscore(chr_name);
		
		//unsigned long chr_begin_pos = counter;
		chr_begin_pos = chars_stored;
						
		unsigned long file_size = reader_vec.GetFileSize();
		//cout << "File Size: " << file_size << endl;
		bool keep_reading = true;
		unsigned int i;

		while(keep_reading) { 
	
			keep_reading = reader_vec.read(gen_piece);
	
			//If we're not at the end of the file, then we can throw it the whole chunk,
			//without the extra gMER_SIZE bases at the end.
			//When we come to the end of the file, we can give it everything.
			//We'll also return the number of characters stored.
			//printf("Buffer size is: %u and mer size is: %u\n",gBUFFER_SIZE,gMER_SIZE);
			if(chars_stored != counter) {
				fprintf(stderr,"NE: chars_stored: %lu counter: %lu\n",chars_stored,counter);
				chars_stored = counter;
			}
			else {
				;//fprintf(stdout,"EQ: chars_stored: %u counter: %u\n",chars_stored,counter);
			}
			
			if(keep_reading) {
				try {
					// We'll have it increment our counter for us
					store(genome_vec,gen_piece,gBUFFER_SIZE-gMER_SIZE,chars_stored);
				}
				catch(bad_alloc& e) {
					printf("Caught bad alloc around store!\n");
					throw e;
				}
				//printf("keeping reading\n");
			}
			else {
				try {
					store(genome_vec,gen_piece,gBUFFER_SIZE,chars_stored);
				}
				catch(bad_alloc& e) {
					printf("Caught bad alloc around store!\n");
					throw e;
				}
			}
			
	 
			for(i=0; i < gBUFFER_SIZE-gMER_SIZE; i++) {
					
				pair<int ,unsigned int> hash = bs.get_hash(i,gen_piece,gBUFFER_SIZE-gMER_SIZE);

				if(hash.first >= 0) {	//if it's zero, that's a result of an N character.
					if(hash.first != 0)	{	//don't want to store N's.
						if(counter % gGEN_SKIP == 0) { //This allows us a smaller hash size in total.
							try {
								if(start_yet)
									(*gen_hash_vec)[hash.second].push_back(counter);
							}
							catch(bad_alloc& e) {
								printf("Caught bad alloc around push_back!\n");
								throw e;
							}
							//printf("Hash of %u (%d) is %s\n",hash.second,hash.first,bs.hash2str(hash.second,gMER_SIZE).c_str());
						}
						
						// print output
						if( gVERBOSE && ((counter % (file_size/50) ) == 0)) {
							cout << ".";
							cout.flush();
						}

					}
				}
						
				//if hash.first is -2, it means it's a new 'chromosome'.
				if(hash.first == -2) {
					//printf("New Genome at counter: %lu and i: %u\n",counter,i);
					
					//Find the name of this genome.
					//at position i, we should have the '>'
					char gen_name[MAX_NAME_LEN+5];	//the name can only be 100 chars long
					unsigned int gen_pos=0; 
					unsigned int name_pos = i;
					
					//we want to read the name of the next chromosome that's been found
					while(gen_piece[++i] != '\n' && i < gBUFFER_SIZE) { 
						if(gen_pos < MAX_NAME_LEN)
							gen_name[gen_pos] = gen_piece[i];
						gen_pos++;
					}
					//Three things could happen here: 
					// 1. The end of the name (\n) extends beyone the end of the buffer
					//		(it didn't finish reading the name)
					// 2. The name extends into the last MER_SIZE bases, but doesn't 
					//		continue past the end of the buffer
					
					//*** CASE 1 ***//
					if(i >= gBUFFER_SIZE) {	//it didn't finish reading the name
						//fprintf(stderr,"Special Case #1(%u)\n",i);
						//fprintf(stdout,"Special Case #1\n");
						char* line = reader_vec.ReadLine();
						string temp = line;
						delete[] line;
						
						// also need to store any potentially missed characters
						// We want to include the name so we can find it...
						// But if the name portion is short, we want to at least the last
						// gMER_SIZE bases
						if(name_pos + gMER_SIZE < gBUFFER_SIZE)  {
							// Don't need to store or increment the counter
						}
						else {
							// We need to store the last characters before the name
							//chars_stored += store(genome_vec,gen_piece+gBUFFER_SIZE-gMER_SIZE,gMER_SIZE);
							//unsigned int temp_num = store(genome_vec,gen_piece+gBUFFER_SIZE-gMER_SIZE,gMER_SIZE);
							// Get from where the name began to at least MER_SIZE chars
							//unsigned int temp_num = store(genome_vec,gen_piece+name_pos,gMER_SIZE);
							unsigned int to_read = gBUFFER_SIZE - gMER_SIZE;
							store(genome_vec,gen_piece+to_read,gMER_SIZE, chars_stored);

						}
					
						// Set up the 'chromosome' name
						unsigned int chr_name_pos = 0;
						while(gen_pos < MAX_NAME_LEN && chr_name_pos < temp.size())	//don't overflow the gen_name
							gen_name[gen_pos++] = temp[chr_name_pos++];
						gen_name[gen_pos] = '\0';
					}
					
					//*** CASE 2 ***//
					// Here, there's not enough characters on the end of the sequence to read
					// a new hash, so it throws off the numbering and reading(?).
					// By default, i < gBUFFER_SIZE
					else if(i + gMER_SIZE >= gBUFFER_SIZE) {
						// Shift the reader's offset by the number of bases remaining in the sequence
						// so it will read them again.
						reader_vec.ShiftOffset(gMER_SIZE - (gBUFFER_SIZE - i));
						
						//fprintf(stdout,"MAde it to special case #3: %lu (%u)--%u %s\n",
						//	chars_stored,i,gBUFFER_SIZE-i-1,gen_piece+gBUFFER_SIZE-(gBUFFER_SIZE-i-1));
						//fprintf(stdout,"Incrementing offset by %d and hash.second is %d\n",gMER_SIZE - (gBUFFER_SIZE - i),hash.second);
					}
					else {
						;//printf("Nothing special :( i: %u vs OTHER: %u diff: %u\n",i,gBUFFER_SIZE-gMER_SIZE,gBUFFER_SIZE-gMER_SIZE-i);
					}

					gen_name[gen_pos > MAX_NAME_LEN ? MAX_NAME_LEN : gen_pos] = '\0'; 
					
					string str_gen_name = gen_name;
					ReplaceSpaceWithUnderscore(str_gen_name);
					if(gVERBOSE > 1) 
						printf("Found new genome at pos %u: %s\n",chr_begin_pos,gen_name);

					vector<pair<string,unsigned long> >::iterator vit = names.begin();
					//Find the position of this genome name
					for(unsigned int name_pos=0; name_pos < j; name_pos++)
						vit++;
					vit++;	// the vector inserts to the one "previous", but we want to insert
							// after the current name.

					//printf("Inserting: %s, %d at %d before %s, %d ",str_gen_name.c_str(),0,j,(*vit).first.c_str(),(*vit).second);
					//printf("and after %s, %d",(*(vit-1)).first.c_str(),(*(vit-1)).second);
					names.insert(vit,pair<string,unsigned long>(str_gen_name,0));

					//take care of the former name.
					//1. set the name in the genome to have the correct position
					//2. set the chr_begin_pos var to be the number of chars we've stored so far.
					names[j] = pair<string,unsigned long>(chr_name,chr_begin_pos);
					chr_name = str_gen_name;
				
					counter += hash.second;	//increment the counter the number of bases we've skipped
					
					chr_begin_pos = counter;
					j++;
					
					// Check to see if we've come across the starting point
					if(counter >= my_start)
						start_yet = true;
					// We want to return here, right?  Because we don't care about the rest of the genome...
					if(counter > my_end && !read_all) {
						//fprintf(stderr,"Exiting HERE!\n");
						keep_reading = false;
						// Increment j because that's what will happen at the end of the loop
						j++;
						break;
					}
				}
				

				//if hash.first is -1, it means it's found the end-of-line char.
				else if(hash.first == -1 || i >= gBUFFER_SIZE-gMER_SIZE) {
					//if(hash.first == -1)
						//printf("Found the eof char\n");
					break;
				}
				else {	// can't be -2 or -1, so it must be nonnegative.
					counter++;

					// Check to see if we've come across the starting point
					if(counter >= my_start)
						start_yet = true;
					if(counter > my_end && !read_all) {
						//fprintf(stderr,"Exiting HERE2! (at %lu vs set:%lu)\n",counter,my_end);
						keep_reading = false;
						break;
					}
					//printf("<%u>",counter);
				}
			}
		}
		if(gVERBOSE)
			cout << endl;

		//fprintf(stderr,"Making name at %d (pos=%lu) to %s\n",j,chr_begin_pos,chr_name.c_str());
		names[j] = pair<string,unsigned long>(chr_name,chr_begin_pos);
		
		//there are cases where the end of the line throws off our numbering
		//We'll fix it here.
		unsigned long int end = i;
		int count=0;
		for(unsigned int p=end; gen_piece[p] != '\0'; p++) {
			//if(base2int(gen_piece[p]) <= 4) {
			if(g_gen_CONVERSION[(int)gen_piece[p]] <= 4) {
				counter++;
				count++;
			}
		}

		// Check to see if we've come across the starting point
		if(counter >= my_start)
			start_yet = true;
		if(counter > my_end && !read_all) {
			//fprintf(stderr,"Exiting HERE3!\n");
			keep_reading = false;
			j++;
			break;
		}

	}

	chr_begin_pos = chars_stored;

	fprintf(stderr,"Making name at %d (pos=%lu) to %s\n",j,chr_begin_pos,"end");
	names[j] = pair<string,unsigned long>("end",chr_begin_pos);

	fprintf(stderr,"Size of genome: %lu\n",chr_begin_pos);

		
#ifdef DEBUG
		for(unsigned int i=0; i<names.size(); i++) {
			printf("Name[%d]: %s\tBegin: %lu\n",i,names[i].first.c_str(),names[i].second);
		}
		fflush(stdout);
		printf("Last Hash number: %lu\n",chars_stored);
#endif

	// Need to convert it to the real, vector genome.
	convertToVector(gen_hash_vec, genome_vec);

	// fprintf(stderr,"End of hash_and_store\n");
}

void Genome::ReplaceSpaceWithUnderscore(string &str) {
	for(unsigned int i=0; i<str.length(); i++) {
		if(str[i] == ' ' || str[i] == 10 || str[i] == 11 || str[i] == 12 || str[i] == 13)
			str[i] = '_';
	}

}

//void Genome::convertToVector(hash_map<unsigned int, vector<unsigned long> > *gh, 
//							 gvector<GenomeLocation> *g) {
void Genome::convertToVector(hash_map<unsigned int, vector<unsigned long> > *gh, 
							 gvector<GEN_TYPE> *pg) {

	fprintf(stderr,"Converting to Vector...\n");
	fprintf(stderr,"\tTrying to create hash of size %lu\n",gh->size());
	
	unsigned long count = 0;
	unsigned long max_size = 0;
	unsigned long min_size = 999999999;
	

	// Set up the hashed genome
	gen_hash.set_size(gh->size());
	for(unsigned int i=0; i<gh->size(); i++) {
		gen_hash[i].size = (*gh)[i].size();
	
		// These next three lines are for debug purposes
		if(gen_hash[i].size > max_size)
			max_size = gen_hash[i].size;
		count += gen_hash[i].size;
		if(gen_hash[i].size < min_size && gen_hash[i].size != 0)
			min_size = gen_hash[i].size;
		
		gen_hash[i].hash_arr = new unsigned long[(*gh)[i].size()];
		for(unsigned int j=0; j<(*gh)[i].size(); j++) {
			gen_hash[i].hash_arr[j] = (*gh)[i][j];
		}

		// Save space by clearing this location.
		//(*gh)[i].erase((*gh)[i].begin,(*gh)[i].end());
	}
	fprintf(stderr,"\tFinished create hash.\n");
	fprintf(stderr,"\tStats: Longest hash is %lu, shortest is %lu, and average is %f\n",max_size,min_size,((double)count)/gh->size());

	// Free this space created by the genome vector
	delete gh;

	//fprintf(stderr,"Just deleted gh\n");

	// Set up the genome
	//if(genome)
	//	free(genome);
	if(packed_genome)
		free(packed_genome);
	if(amount_genome)
		free(amount_genome);
	
	//fprintf(stderr,"Just freed the genome, trying to malloc %lu\n",sizeof(GenomeLocation) * g->size());
	//genome = new GenomeLocation[g->size()];
	//genome = (GenomeLocation*)malloc(sizeof(GenomeLocation) * g->size());
	packed_genome = (GEN_TYPE*)malloc(sizeof(GEN_TYPE) * pg->size());
	amount_genome = (float*)malloc(sizeof(float) * pg->size());
	memset(amount_genome, 0, sizeof(float)*pg->size());


	//fprintf(stderr,"Creating new genome with location of %lu * %u = %lu\n",g->size(),gGEN_SIZE,g->size() * gGEN_SIZE);
	fprintf(stderr,"\tCreating new genome with a size of %lu\n",pg->size() * gGEN_SIZE);
	
	//if(!genome)
	if(!packed_genome || !amount_genome)
		throw(new Exception("Could not malloc genome"));
	
	//int percent = g->size()/100;

	for(unsigned long i=0; i<pg->size(); i++) {
		//if(i % percent == 0 && i != 0)
			//fprintf(stderr,"\r%d%% complete",g->size()/i);

		//genome[i].packed_char = (*g)[i].packed_char;
		//genome[i].amount = 0;
		packed_genome[i] = (*pg)[i];
	}
	genome_size = pg->size();
	
	// Clear up the memory we've used
	pg->clear();
	delete pg;
	
	fprintf(stderr,"Finished Vector Conversion\n");
}



#define READ_EXTRA	1000

/*
 * store() will store the genome into a smaller array of BIT64_t portions.
 * Each character is represented by 4 bits (because we require the 4 nucleotides
 * and an additional N character), and can be accessed directly.  
 *
 * One of the problems we have now is the last few characters in gen.  They will
 * appear the second time at the beginning of the genome yet again, throwing off the 
 * numbering system. 
 *
 * @par gen - the piece of the genome to store.
 * @par num - the number of characters to store.
 * @par num_bases - the base counter.  And tells when to start recording
 */
//void Genome::store(gvector<GenomeLocation> *g, unsigned char* gen, 
//						unsigned long int num, unsigned long &num_bases) {
void Genome::store(gvector<GEN_TYPE> *pg, unsigned char* gen, 
						unsigned long int num, unsigned long &num_bases) {

	// DONE!!
	if(num_bases > my_end+READ_EXTRA && !read_all)
		return;
		
	bool start_yet = num_bases >= my_start ? true : false;

	//fprintf(stderr,"READ_ALL: %d start_yet: %d\n",read_all,start_yet);

	unsigned long gen_pos=0;
//	unsigned long int num_bases=0;
	
	while(gen_pos<num) {
		GEN_TYPE temp = 0;
		
		if(bin_offset != 0) {
			if( (start_yet && num_bases <= my_end+READ_EXTRA) || read_all) {
				//fprintf(stderr,"Is this the error?  Offset is %d and my_start is %ul\n",offset,my_start);
				//temp = (*g)[offset-my_start/gGEN_SIZE].packed_char;
				//temp = (*g)[offset].packed_char;
				temp = (*pg)[offset];
				pg->pop_back();
			}
		}
			
		for(; (bin_offset<gGEN_SIZE) && gen_pos<num; gen_pos++) {	//can fit 16 chars into a BIT64_t 
			//int int_to_add = base2int(gen[gen_pos]);
			int int_to_add = g_gen_CONVERSION[(int)gen[gen_pos]];
			
			
			//If it's a valid char.
			if(int_to_add < 5) {
				num_bases++;
				if(num_bases > my_start)
					start_yet = true;
				if(num_bases > my_end+READ_EXTRA && !read_all) {
					// remember this location, then return
					//g->push_back(GenomeLocation(temp,0));
					pg->push_back(temp);
					return;	// DONE!
				}

				if(start_yet) {
					temp <<= 4;
					temp |= (int_to_add & 0xf);
					//fprintf(stderr,"Adding %c: %x(%d), now %0.8x\n",gen[gen_pos],int_to_add & 0xf,int_to_add,temp);
					++bin_offset;

				}
									
			}
			//If we come accross the flag for the end of the stream--should never happen...
			else if(int_to_add == 6) { 
				gen_pos = num;
				//cout << "Found end." << endl;
				break;
			}
			else if(int_to_add == 7) {	//come accross a new chromosome name.
				string temp_str = (char*)(gen+gen_pos);
				//while(gen[gen_pos] != '\n' && gen[gen_pos] != '\0' && gen_pos<num )
				while( gen[gen_pos] != '\n' && gen[gen_pos] != '\0' )
					gen_pos++;	//increment the counter until we've found the end of the name
			}
			else ;//do nothing--it's an invalid character and not an end-of-line
		}
				
		if(bin_offset != 0)	{ //if we've made any changes to this value...
			try {
				//g->push_back(pair<GEN_TYPE,float>(temp,0));
				if(read_all || (start_yet && num_bases <= my_end+READ_EXTRA)) {
					//fprintf(stderr,"Pushing %0.8x onto genome\n",temp);
					//g->push_back(GenomeLocation(temp,0));
					pg->push_back(temp);
				}
			}
			catch(bad_alloc& e) {
				printf("Found ERROR!!!:  %lu elements out of max of %lu\n",pg->size(),pg->max_size());
				//printf("Total space is %lu\n",g->size()*sizeof(GenomeLocation));
				printf("Total space is %lu\n",pg->size()*sizeof(GEN_TYPE));
				throw e;
			}
		}
		else ;	//no changes.  Coming out of the previous for-loop
		
		
		//means we've filled it fully--so increment the offset.
		if(bin_offset == gGEN_SIZE) {
			if(start_yet) {
				offset++;
				bin_offset = 0;
			}
		}


	}
	
	//fprintf(stdout,"Just read %ld bases out of %lu assigned\n",num_bases, num);
	return;
}



/*
 * This function will just count the genome--it won't store anything.
 */
unsigned long Genome::count() {
	
	// Set some values that are necessary for the 'store()' function
	offset = 0;
	bin_offset = 0;

	Reader reader_vec;
	unsigned char gen_piece[gBUFFER_SIZE+1];

	if(gVERBOSE)
		printf("Counting Genome to Reserve Size...\n");

	
	bin_seq bs;	

	unsigned long counter = 0;
	unsigned long int chr_begin_pos = 0;
	unsigned int j;

	for(j=0; j<names.size()-1; j++) {
		reader_vec.use(names[j].first);
		string chr_name = reader_vec.GetName();
		ReplaceSpaceWithUnderscore(chr_name);
		
		//unsigned long chr_begin_pos = counter;
		chr_begin_pos = counter;
						
		bool keep_reading = true;
		unsigned int i;

		while(keep_reading) { 
	
			keep_reading = reader_vec.read(gen_piece);
	
			for(i=0; i < gBUFFER_SIZE-gMER_SIZE; i++) {
					
				pair<int ,unsigned int> hash = bs.get_hash(i,gen_piece,gBUFFER_SIZE-gMER_SIZE);
						
				//if hash.first is -2, it means it's a new 'chromosome'.
				if(hash.first == -2) {
					//printf("New Genome at counter: %lu and i: %u\n",counter,i);
					
					//Find the name of this genome.
					//at position i, we should have the '>'
					char gen_name[MAX_NAME_LEN+5];	//the name can only be 100 chars long
					unsigned int gen_pos=0; 
					
					//we want to read the name of the next chromosome that's been found
					while(gen_piece[++i] != '\n' && i < gBUFFER_SIZE) { 
						if(gen_pos < MAX_NAME_LEN)
							gen_name[gen_pos] = gen_piece[i];
						gen_pos++;
					}
					//Three things could happen here: 
					// 1. The end of the name (\n) extends beyone the end of the buffer
					//		(it didn't finish reading the name)
					// 2. The name extends into the last MER_SIZE bases, but doesn't 
					//		continue past the end of the buffer
					
					//*** CASE 1 ***//
					if(i >= gBUFFER_SIZE) {	//it didn't finish reading the name
						//fprintf(stderr,"Special Case #1(%u)\n",i);
						//fprintf(stdout,"Special Case #1\n");
						char* line = reader_vec.ReadLine();
						string temp = line;
						delete[] line;
						
						// Set up the 'chromosome' name
						unsigned int chr_name_pos = 0;
						while(gen_pos < MAX_NAME_LEN && chr_name_pos < temp.size())	//don't overflow the gen_name
							gen_name[gen_pos++] = temp[chr_name_pos++];
						gen_name[gen_pos] = '\0';
					}
					
					//*** CASE 2 ***//
					// Here, there's not enough characters on the end of the sequence to read
					// a new hash, so it throws off the numbering and reading(?).
					// By default, i < gBUFFER_SIZE
					else if(i + gMER_SIZE >= gBUFFER_SIZE) {
						// Shift the reader's offset by the number of bases remaining in the sequence
						// so it will read them again.
						reader_vec.ShiftOffset(gMER_SIZE - (gBUFFER_SIZE - i));
						
					}

					gen_name[gen_pos > MAX_NAME_LEN ? MAX_NAME_LEN : gen_pos] = '\0'; 
					
					string str_gen_name = gen_name;
					ReplaceSpaceWithUnderscore(str_gen_name);

					vector<pair<string,unsigned long> >::iterator vit = names.begin();
					//Find the position of this genome name
					for(unsigned int name_pos=0; name_pos < j; name_pos++)
						vit++;
					vit++;	// the vector inserts to the one "previous", but we want to insert
							// after the current name.

					names.insert(vit,pair<string,unsigned long>(str_gen_name,0));

					//take care of the former name.
					//1. set the name in the genome to have the correct position
					//2. set the chr_begin_pos var to be the number of chars we've stored so far.
					names[j] = pair<string,unsigned long>(chr_name,chr_begin_pos);
					chr_name = str_gen_name;
				
					counter += hash.second;	//increment the counter the number of bases we've skipped

					chr_begin_pos = counter;
					j++;
				}
				

				//if hash.first is -1, it means it's found the end-of-line char.
				else if(hash.first == -1 || i >= gBUFFER_SIZE-gMER_SIZE) {
					break;
				}
				else {	// can't be -2 or -1, so it must be nonnegative.
					counter++;
				}
			}
		}

		names[j] = pair<string,unsigned long>(chr_name,chr_begin_pos);
		
		//there are cases where the end of the line throws off our numbering
		//We'll fix it here.
		unsigned long int end = i;
		int count=0;
		for(unsigned int p=end; gen_piece[p] != '\0'; p++) {
			//if(base2int(gen_piece[p]) <= 4) {
			if(g_gen_CONVERSION[(int)gen_piece[p]] <= 4) {
				counter++;
				count++;
			}
		}
	}
	names[j] = pair<string,unsigned long>("end",counter);
	
//	fprintf(stderr,"Size of genome: %lu vs %lu\n",counter,chr_begin_pos);

		
#ifdef DEBUG 
		for(unsigned int i=0; i<names.size(); i++) {
			printf("Name[%d]: %s\tBegin: %lu\n",i,names[i].first.c_str(),names[i].second);
		}
		fflush(stdout);
		printf("Last Hash number: %lu\n",counter);
#endif

	if(gVERBOSE)
		printf("...Finished!\n\n");
	
	return counter;

}



unsigned int Genome::saveGen(char* fn) {
	FILE* saveF;

	saveF = fopen(fn,"w");
	if(!saveF)
		throw "Could not open Genome save file";

	/* Write the hash out to a file */
	//fprintf(saveF,"%ld\n",gen_hash.size());
	unsigned long tempsize = gen_hash.size();
	fwrite(&tempsize, sizeof(unsigned long), 1, saveF);
	fwrite(&gMER_SIZE, sizeof(int), 1, saveF);
	//fprintf(stderr,"%ld\n",gen_hash.size());
	for(unsigned int i=0; i<gen_hash.size(); i++) {
		if(gen_hash[i].size) {
			unsigned int hash_size = gen_hash[i].size;
			//fprintf(saveF,"%d %d\n",i,hash_size);
			fwrite(&i, sizeof(unsigned int), 1, saveF);
			fwrite(&hash_size, sizeof(unsigned int), 1, saveF);
			//fprintf(stderr,"%d %d\n",i,hash_size);
			for(unsigned int j=0; j<hash_size; j++) {
				fwrite(&(gen_hash[i].hash_arr[j]),sizeof(unsigned long),1,saveF);
			}
		}
	}
	// Write out an end marker
	//fprintf(saveF,"0 0\n");
	unsigned int tempmarker = 0;
	fwrite(&tempmarker, sizeof(unsigned int), 1, saveF);
	fwrite(&tempmarker, sizeof(unsigned int), 1, saveF);

	/* Write the genome out to a file. */
	//fprintf(saveF,"%d %ld\n",gGEN_SIZE,genome.size());
	unsigned long tempgenome_size = genome_size;
	fwrite(&gGEN_SIZE, sizeof(gGEN_SIZE), 1, saveF);
	fwrite(&tempgenome_size, sizeof(tempgenome_size), 1, saveF);
	//fprintf(stderr,"%d %ld\n",gGEN_SIZE,genome.size());
	for(unsigned int i=0; i<genome_size; i++) {
		//fwrite(&(genome[i].packed_char),sizeof(GEN_TYPE),1,saveF);
		fwrite(&(packed_genome[i]),sizeof(GEN_TYPE),1,saveF);
	}

	fprintf(saveF,"\n%d\n",(int)names.size());
	//fprintf(stderr,"\n%d\n",(int)names.size());
	
	/* Write the chromosome names to a file */
	for(unsigned int i=0; i<names.size(); i++) {
		fprintf(saveF,"%s %ld\n",names[i].first.c_str(),names[i].second);
	}
	
	fclose(saveF);

	return 0;
}

void Genome::fix_hash() {
	//Means there is no user-defined max-hash.
	if(gMAX_HASH_SIZE <= 0)
		return;

	for(unsigned int i=0; i<gHASH_SIZE; i++)
		if(gen_hash[i].size > gMAX_HASH_SIZE) {
#ifdef DEBUG
			if(gVERBOSE > 1)
				printf("Clearing hash at pos %d with size %d\n",i,gen_hash[i].size);
#endif
			delete[] gen_hash[i].hash_arr;
			gen_hash[i].hash_arr = 0;
			gen_hash[i].size = 0;
		}
}

/*
 * Returns a string from begin to begin+size.
 * the first position in the genome is position 0.
 * @par begin - the beginning of the string to return
 * @par size - the size of the string to return.
 */
string Genome::GetString(const unsigned long begin, const unsigned int size) { 
#ifdef DEBUG
	fprintf(stderr,"Getting string at %lu: ",begin);
#endif
 
	string to_return = "";

	/***********************Check for Bounds******************************/
	//We don't want to get a string beyond the end of the chromosome...
	int chr_num=0;
	//											it's already been pre-incremented
	while(names[chr_num].second <= begin && ((unsigned int)chr_num <= names.size()) ) chr_num++;
	//while(names[++chr_num].second <= begin && ((unsigned int)chr_num <= names.size()) );
	
	//means it's not valid--will extend beyond the chromosome file
	if( (names[chr_num].second < begin+size) ) 
		//printf("Extending Beyond the Chromosome\n");
		return "";
	/*********************************************************************/

	unsigned long pos_in_gen = (begin-my_start)/gGEN_SIZE; 
	unsigned int temp_size = size;
	
	//fprintf(stderr,"getting %lu, max is %lu and asked for is %lu\n",pos_in_gen,genome_size,begin);
	assert(pos_in_gen <= genome_size);
	//GEN_TYPE t_temp = genome[pos_in_gen].packed_char;
	GEN_TYPE t_temp = packed_genome[pos_in_gen];

	//cerr << "\tbegin: " << begin << endl;
	//cerr << "\tmy_start: " << my_start << endl;
	//cerr << "\tbegin-my_start: " << begin-my_start << endl;
	//cerr << "\tpos_in_gen: " << pos_in_gen << endl;
	//cerr << "\tt_temp: " << t_temp << endl;
		
	//if(gVERBOSE > 3)
	//	cout << "getting position: " << t_temp << endl;
	
	unsigned int count=0;

	while(count < size) {
		//fprintf(stderr,"\tt_temp: %0.8x %0.8x\n",t_temp,genome[pos_in_gen+1].packed_char);

		string temp = "";
	
		if(count == 0) {	//get the last few characters off the BIT64_t
			unsigned long bit_offset = ((begin-my_start)%gGEN_SIZE);

			//for(unsigned long j=0; j<gGEN_SIZE-bit_offset; j++) {
			for(unsigned int j=gGEN_SIZE-bit_offset; j>0; j--) {
				//shift 0xf left 4 bases for each j.
				//then we don't need to reverse the string.
				//printf("Shifting by: %d = %lld\n",4*(j-1),(t_temp>>(4*(j-1))) & 0xf);
				//temp += int2base(t_temp & 0xf);
				//temp += int2base(0xf & (t_temp >> (4*(j-1))));
				temp += gINT2BASE[0xf & (t_temp >> (4*(j-1)))];
#ifdef DEBUG
                fprintf(stderr,"%d",0xf & (t_temp >> (4*(j-1))));
#endif
				//t_temp >>= 4;
				//fprintf(stderr,"Round %d, temp is %s\n",count,temp.c_str());
				count++;
			}
			//temp = reverse(temp);
			if(temp.size() > size)
				return temp.substr(0,size);
		}
		else {	//we now want the first offset characters off the BIT64_t.
			//for(unsigned int j=0; j<gGEN_SIZE; j++) {
			for(unsigned int j=gGEN_SIZE; j>0; j--) {
				
				//fprintf(stderr,"Shifting by: %d = %lld\n",4*(j-1),(t_temp>>(4*(j-1))) & 0xf);
				//temp += int2base(t_temp & 0xf);
				//temp += int2base((t_temp >> (4*(j-1))) & 0xf);
				temp += gINT2BASE[0xf & (t_temp >> (4*(j-1)))];
				//t_temp >>= 4;
#ifdef DEBUG
                fprintf(stderr,"%d",0xf & (t_temp >> (4*(j-1))));
#endif
			}
			//temp = reverse(temp);
			temp = temp.substr(0,temp_size);
			count += temp.size();
			//fprintf(stderr,"Round %d, temp is %s and return is %s\n",count,temp.c_str(),to_return.c_str());
		}
		to_return += temp;
		
		temp_size = size - count;
		
		++pos_in_gen;
		if((pos_in_gen)  >= genome_size) {
				t_temp = (GEN_TYPE)0x44444444;
		}
		else
			//t_temp = genome[pos_in_gen].packed_char;
			t_temp = packed_genome[pos_in_gen];
	}
#ifdef DEBUG
    fprintf(stderr,"\n");
#endif
	
	return to_return;
}

inline int Genome::base2int(char c) {
	switch(c) {
		case 'A' : return 0;
		case 'C' : return 1;
		case 'G' : return 2;
		case 'T' : return 3;
		case 'a' : return 0;
		case 'c' : return 1;
		case 'g' : return 2;
		case 't' : return 3; 
		case 'N' : return 4;
		case 'n' : return 4;
		case '\n' : return 5;
		case '\r' : return 5;
		case '\0' : return 6;	//6 is the flag for the end of the genome stream.
		case '>' : return 7;
		default: return 4;	//we want invalid characters to be represented by "N"
	}
	return 5;	//5 is an invalid character.
} 

inline char Genome::int2base(int i) {
	//printf("i: %d\n",i);
	switch(i) {
		case 0: return 'a';
		case 1: return 'c';
		case 2: return 'g';
		case 3: return 't';
		case 4: return 'n';
	}
	return '?';
}

inline string Genome::reverse(string &s) {
	string temp = "";
	int len = s.length();
	
	for(int i=1; i<=len; i++)
		temp += s[len-i];
	
	return temp;
}

HashLocation* Genome::GetMatches(unsigned int hash) {

	return &gen_hash[hash];
}

float Genome::GetScore(const unsigned long &pos) {

	unsigned long pos_temp = (pos-my_start)/gGEN_SIZE;
	//return genome[pos_temp].amount;
	return amount_genome[pos_temp];
}
void Genome::AddScore(const unsigned long &pos, const float &amt) {
//	if(amt > 1)
//		fprintf(stderr,"Genome Error!! %d\n",__LINE__);
		
	//float pos_temp = pos / BIN_SIZE;
	//pos_temp *= BIN_SIZE;
	unsigned long pos_temp = pos-my_start;
	//genome[pos_temp/gGEN_SIZE].amount+= amt;
	amount_genome[pos_temp/gGEN_SIZE] += amt;
}

void Genome::AddSeqScore(unsigned long pos, const float* amt, const float scale) {
	pos -= my_start;
	
	if(reads[A_POS])
		reads[A_POS][pos/gGEN_SIZE] += amt[A_POS] * scale;
	if(reads[C_POS])
		reads[C_POS][pos/gGEN_SIZE] += amt[C_POS] * scale;
	if(reads[G_POS])
		reads[G_POS][pos/gGEN_SIZE] += amt[G_POS] * scale;
	if(reads[T_POS])
		reads[T_POS][pos/gGEN_SIZE] += amt[T_POS] * scale;
	if(reads[N_POS])
		reads[N_POS][pos/gGEN_SIZE] += amt[N_POS] * scale;
}

void Genome::AddSeqScore(unsigned long pos, const float amt, char which) {
	pos -= my_start;
	
	switch(which) {
		case 'a':
			if(reads[0])	//if we've created one, then add to it.  Used for both gBISULFITE and gATOG
				reads[0][pos/gGEN_SIZE] += amt;
			break;
		case 'c':
			if(reads[1])
				reads[1][pos/gGEN_SIZE] += amt;
			break;
		case 'g':
			if(reads[2])
				reads[2][pos/gGEN_SIZE] += amt;
			break;
		case 't':
			if(reads[3])
				reads[3][pos/gGEN_SIZE] += amt;
			break;
		case 'n':
			if(reads[4])
				reads[4][pos/gGEN_SIZE] += amt;
			break;
		case '-':
			if(reads[5])
				reads[5][pos/gGEN_SIZE] += amt;
			break;
	}
}

void Genome::AddGenScore(unsigned long pos, const float amt, char which) {
	pos -= my_start;

	//genome[pos/gGEN_SIZE].second += amt;	//we only want to add it to the genome once.
										//we'll do it while we're adding to the genome sequence files.
							//actually, we're already adding the score of each sequence to the genome.
							//we're doing it a second time here.  bad news, bears.
	switch(which) {
		case 'a':
			if(gens[0])
				//gens[0][pos/gGEN_SIZE] += amt;
				gens[0][pos/gGEN_SIZE] = 1;
		break;
		case 'c':
			if(gens[1])
				//gens[1][pos/gGEN_SIZE] += amt;
				gens[1][pos/gGEN_SIZE] = 1;
		break;
		case 'g':
			if(gens[2])
				//gens[2][pos/gGEN_SIZE] += amt;
				gens[2][pos/gGEN_SIZE] = 1;
		break;
		case 't':
			if(gens[3])
				//gens[3][pos/gGEN_SIZE] += amt;
				gens[3][pos/gGEN_SIZE] = 1;
		break;
		case 'n':
			if(gens[4])
				//gens[3][pos/gGEN_SIZE] += amt;
				gens[4][pos/gGEN_SIZE] = 1;
		break;

		case '-':
			if(gens[5])
				//gens[3][pos/gGEN_SIZE] += amt;
				gens[5][pos/gGEN_SIZE] = 1;
		break;
	}
}

string Genome::GetPos(const pair<unsigned long,bool> p_pos) {
	return GetPos(p_pos.first,p_pos.second);
}
string Genome::GetPos(const unsigned long &pos, const bool upstrand) {
	int chr_num = 0;

	while(names[chr_num].second <= pos && ((unsigned int)chr_num < names.size()) )
		chr_num++;
		
	ostringstream to_return;
	
	chr_num--;
	to_return << names[chr_num].first << ":";
	if(upstrand)
		to_return << "+";
	else
		to_return << "-";
	to_return << (pos - names[chr_num].second);
	
	return to_return.str();
}

pair<string,unsigned long> Genome::GetPosPair(const unsigned long pos) {
	
	int chr_num = 0;

	while(names[chr_num].second <= pos && ((unsigned int)chr_num < names.size()) )
		chr_num++;
		
	chr_num--;

	string chr_name = names[chr_num].first;
	unsigned long chr_pos = (pos - names[chr_num].second);
	
	return pair<string,unsigned long>(chr_name,chr_pos);
}

/*
int Genome::getSetThreadID(long unsigned int pos, pthread_t myID) {

	// Already been set
	if(pos/gGEN_SIZE >= genome_size) 
		printf("pos: %lu\tmax: %lu\n",pos,genome_size);
		
	if(genome[pos/gGEN_SIZE].thread_id == myID)
		return -1;
	
	// Hasn't been set so set it.
	genome[pos/gGEN_SIZE].thread_id = myID;
	return 0;
}
*/

inline bool Genome::is_max_pos(char to_check, float a, float c, float g, float t, float n) {
	switch(to_check) {
		case 'a':
			return !(a >= c && a >= g && a >= t && a >= n);
		case 'c':
			return !(c >= a && c >= g && c >= t && c >= n);
		case 'g':
			return !(g >= c && g >= a && g >= t && g >= n);
		case 't':
			return !(t >= c && t >= g && t >= a && t >= n);
		case 'n':
			return !(n >= c && n >= g && n >= t && n >= a);
	}

	return false;
}

inline unsigned int Genome::max_pos(float array[], unsigned int array_size) {
	float* pos = max_element(array, array+array_size);
	unsigned int max_pos = pos - &array[0];
	return max_pos;
}

const unsigned int MIN_COVERAGE = 5;
const float MIN_RATIO = 1.5;
const float MAX_PVAL = 0.01;

/**
 * dchisq will give the chi-square distribution at x with n degrees of freedom
 * n must be more than zero
 */
inline double dchisq(float x, int n) {
	/*
	this is dchisq(x,n):
	f_n(x) = 1 / (2^(n/2) Gamma(n/2)) x^(n/2-1) e^(-x/2)
		OR
	1 / (pow(2,n/2) * gamma(n/2)) * pow(x,n/2-1) * exp(-x/2)
	*/
	
	return 1.0 / (pow(2.0,((double)n)/2.0) * tgamma(((double)n)/2))
			* pow((double)x,((double)n)/2-1) 
			* exp(-x/2);
}

inline float Genome::LRT(float chars[5], int &snp_pos) {
	snp_pos = max_pos(chars, 5);

	double sum = chars[0] + chars[1] + chars[2] + chars[3] + chars[4];

	double ratio = pow(.2,sum) 
					/ ( pow(chars[snp_pos]/sum,(double)chars[snp_pos]) 
					    * pow(1-chars[snp_pos]/sum,sum-chars[snp_pos]) );
	
	double pval = dchisq(-2*log(ratio),1);
	
	return pval;
	
}

inline float Genome::dipLRT(float chars[5], int &snp_pos) {
	snp_pos = max_pos(chars, 5);
	
	float sum = chars[0] + chars[1] + chars[2] + chars[3] + chars[4];

	float ratio = pow((float).2,sum) 
					/ (pow(chars[snp_pos]/sum,chars[snp_pos]) 
					  * pow(1-chars[snp_pos]/sum,sum-chars[snp_pos]));
	float pval1 = dchisq(-2*log(ratio),1);

	float second[5];
	for(unsigned int i=0; i<5; i++) {
		second[i] = chars[i];
	}
	second[snp_pos] = 0;
	int snp2 = max_pos(second, 5);
	
	float pval2;
	if(second[snp2] == 0)
		pval2 = -1;
	else {
		//rat = (.2)^n/((x[m]/n)^(x[m])*(x[m2]/n)^(x[m2])*(1-(x[m]+x[m2])/n)^(n-x[m]-x[m2]))
		ratio = pow((float).2,sum)
				/ ( pow(chars[snp_pos]/sum, chars[snp_pos])
				    * pow(chars[snp2]/sum, chars[snp2])
				    * pow(1 - (chars[snp_pos] + chars[snp2])/sum, 
				    	  sum - chars[snp_pos] - chars[snp2]));
		pval2 = dchisq(-2*log(ratio), 2);
	}
	
	return 0;
}

inline float Genome::is_snp(float chars[5], int &snp_pos) {

	return LRT(chars,snp_pos);	
}


void Genome::AppendFinal(const char* fn) {
	PrintFinal(fn, true);
}
void Genome::PrintFinal(const char* fn) {
	PrintFinal(fn, false);
}
void Genome::PrintFinal(const char* fn, bool append) {
	if(gSNP) {
		// We got a new update from Evan.  Instead of printing out a file for each, we want to print
		// out one file (dubbed the '.sgrex' file for 'sgr-extra') to be a tab-deliminated file
		// for the number of a's, c's, g's, t's, and gaps at each location in the genome.
		char sgrex_fn[strlen(fn)+10];
		strcpy(sgrex_fn,fn);
		strcat(sgrex_fn,".sgrex");
		printf("File is: %s\n",sgrex_fn);
		
		FILE* sgrex_file;
		if(append)
			sgrex_file = fopen(sgrex_fn,"a");
		else
			sgrex_file = fopen(sgrex_fn,"w");
			
		if(!sgrex_file)
			perror("Could not open .sgrex file");
		// output the Header
		//fprintf(sgrex_file,"chr\tpos\tsgr_val\tread_a\tread_c\tread_g\tread_t\tread_n\tread_gap\tsnp\n");
		fprintf(sgrex_file,"chr\tpos\tsgr_val\tread_a\tread_c\tread_g\tread_t\tread_n\tsnp\n");
		
		if(gVERBOSE > 1) 
			fprintf(stderr,"NAMES: %zu\n",names.size());

		for(unsigned int j=0,count=0; j<names.size()-1; j++) {
		//for(unsigned int j=0,count=0; names[j].first != "end"; j++) {
			if(gVERBOSE > 1) 
				fprintf(stderr,"\t[%d]: %s  %lu\n",j+1,names[j+1].first.c_str(),names[j+1].second);
			for(; count<names[j+1].second; count+=gGEN_SIZE) {
				if(count < my_start && !read_all)
					continue;
				if(count > my_end && !read_all)
					break;
					
				//if(genome[(count-my_start)/gGEN_SIZE].amount > (float)0.0) {	//if something matched here...
				if(amount_genome[(count-my_start)/gGEN_SIZE] > (float)0.0) {	//if something matched here...
				
									 //  chr pos  sgr
					fprintf(sgrex_file, "%s\t%ld\t%.5f", names[j].first.c_str(), count-names[j].second,
								//genome[(count-my_start)/gGEN_SIZE].amount);
								amount_genome[(count-my_start)/gGEN_SIZE]);

					// Print out each of the read positions
					for(unsigned int read_count=0; read_count<NUM_CHAR_VALS; read_count++) {
						fprintf(sgrex_file,"\t%.5f",reads[read_count][(count-my_start)/gGEN_SIZE]);
					}
					
					// Get the position here
					char at_pos = GetString(count,1)[0];
					// Determine if it's a snp
					float char_arr[5] = {	reads[0][(count-my_start)/gGEN_SIZE],
											reads[1][(count-my_start)/gGEN_SIZE],
											reads[2][(count-my_start)/gGEN_SIZE],
											reads[3][(count-my_start)/gGEN_SIZE],
											reads[4][(count-my_start)/gGEN_SIZE] };

					int snp_pos;
					float p_val = is_snp(char_arr, snp_pos);

					// If the max pos is not what's in the genome
					if(snp_pos != g_gen_CONVERSION[(int)at_pos]) {
					//if(snp_pos != gINT2BASE[(int)at_pos]) {
						if( p_val < 0.01) {
							fprintf(sgrex_file,"\tY:%c->%c p_val=%.5f", 
								at_pos, gINT2BASE[snp_pos], p_val);
						}
						else
							//fprintf(sgrex_file,"\tN:%c->%c p_val=%.6f", 
							//	at_pos, gINT2BASE[snp_pos], p_val);
							fprintf(sgrex_file,"\tN:%c->%c p_val=%.5f", 
								at_pos, gINT2BASE[snp_pos], p_val);
					}
					else
						fprintf(sgrex_file,"\tN:%c p_val=%.5f", at_pos, p_val);

					// Carriage return
					fprintf(sgrex_file,"\n");
				}
			}
		}
		
		// Close the file
		fclose(sgrex_file);
	}
	else if(gBISULFITE) {	// if any of the reads have been instantiated...
		// We got a new update from Evan.  Instead of printing out a file for each, we want to print
		// out one file (dubbed the '.sgrex' file for 'sgr-extra') to be a tab-deliminated file
		// for the number of a's, c's, g's, t's, and gaps at each location in the genome.
		char sgrex_fn[strlen(fn)+10];
		strcpy(sgrex_fn,fn);
		strcat(sgrex_fn,".sgrex");
		printf("File is: %s\n",sgrex_fn);

		FILE* sgrex_file;
		if(append)
			sgrex_file = fopen(sgrex_fn,"a");
		else
			sgrex_file = fopen(sgrex_fn,"w");
		if(!sgrex_file)
			perror("Could not open .sgrex file");
		// output the Header
		//fprintf(sgrex_file,"chr\tpos\tsgr_val\tread_a\tread_c\tread_g\tread_t\tread_n\tread_gap\n");
		fprintf(sgrex_file,"chr\tpos\tsgr_val\tread_a\tread_c\tread_g\tread_t\tread_n\\n");
		
		if(gVERBOSE > 1) fprintf(stderr,"NAMES: %ud\n",(unsigned int)names.size());

		for(unsigned int j=0,count=0; j<names.size()-1; j++) {
			if(gVERBOSE > 1) fprintf(stderr,"\t[%d]: %s\n",j,names[j].first.c_str());
			for(; count<names[j+1].second; count+=gGEN_SIZE) {
				if(count < my_start && !read_all)
					continue;
				if(count > my_end && !read_all)
					break;
					
				//if(genome[(count-my_start)/gGEN_SIZE].amount > (float)0.0) {	//if something matched here...
				if(amount_genome[(count-my_start)/gGEN_SIZE] > (float)0.0) {	//if something matched here...
									 //  chr pos  sgr
					fprintf(sgrex_file, "%s\t%ld\t%.5f", names[j].first.c_str(), count-names[j].second,
								//genome[(count-my_start)/gGEN_SIZE].amount);
								amount_genome[(count-my_start)/gGEN_SIZE]);

					// Print out each of the read positions
					for(unsigned int read_count=0; read_count<NUM_CHAR_VALS; read_count++) {
						fprintf(sgrex_file,"\t%.5f",reads[read_count][(count-my_start)/gGEN_SIZE]);
					}

					// Carriage return
					fprintf(sgrex_file,"\n");
				}
			}
		}
		
		// Close the file
		fclose(sgrex_file);
	}

	else {
		char* filename = new char[strlen(fn)+5];
		sprintf(filename,"%s.sgr",fn);
		
		FILE* sgr_file;
		if(append)
			sgr_file = fopen(filename,"a");
		else
			sgr_file = fopen(filename,"w");
		
		//fprintf(stderr,"Printing for %u genomes to %lu\n",names.size(),names[names.size()-1].second);
		for(unsigned int i=0,count=0; i<names.size()-1; i++) {
			for(; count<names[i+1].second; count+=gGEN_SIZE) {
				if(count < my_start && !read_all)
					continue;
				if(count > my_end && !read_all)
					break;

				//if(genome[(count-my_start)/gGEN_SIZE].amount> 0)
				if(amount_genome[(count-my_start)/gGEN_SIZE] > 0.00001) {
					//fprintf(stderr,"count=%u, start=%lu, end=%lu/%u=%lu, size=%lu, at: %f\n",count,my_start,my_end,gGEN_SIZE,my_end/gGEN_SIZE,genome_size,amount_genome[(count-my_start)/gGEN_SIZE]);
					
					fprintf(sgr_file, "%s\t%ld\t%.5f\n",names[i].first.c_str(),count-names[i].second,
						amount_genome[(count-my_start)/gGEN_SIZE]);
				}
			}
		}
		
		fclose(sgr_file);
		printf("Finished printing to %s\n",filename);
		delete[] filename;
	}


	
	
	/*
	for(unsigned i=0; i<NUM_CHAR_VALS; i++) {
		if(reads[i]) {
			char readfile[100+6];
			strcpy(readfile,fn);

			strcat(readfile,".");
			if(i == 0)
				strcat(readfile,"a");
			else if(i == 1)
				strcat(readfile,"c");
			else if(i == 2)
				strcat(readfile,"g");
			else if(i == 3)
				strcat(readfile,"t");
			else if(i == 4)
				strcat(readfile,"-");	//For the gap

			strcat(readfile,".sgr");
			out.open(readfile);
			for(unsigned int j=0,count=0; j<names.size()-1; j++) {
				for(; count<names[j+1].second; count+=gGEN_SIZE) {
					if(reads[i][count/gGEN_SIZE] > 0)
						out << names[j].first << "\t" << count - (names[j].second) << "\t" << reads[i][count/gGEN_SIZE] << endl;
					//if(a_spots[count/gGEN_SIZE] > 0)
						//out << names[i].first << "\t" << count - (names[i].second) << "\t" << a_spots[count/gGEN_SIZE] << endl;
				}

			}
			out.close();
		}

		if(gens[i]) {
			char genfile[100+10];
			strncpy(genfile,fn,100);
			strcat(genfile,".gen.");
			if(i == 0)
				strcat(genfile,"a");
			else if(i == 1)
				strcat(genfile,"c");
			else if(i == 2)
				strcat(genfile,"g");
			else if(i == 3)
				strcat(genfile,"t");
			else if(i == 4)
				strcat(genfile,"-");	//For the gap

			strcat(genfile,".sgr");
			out.open(genfile);
			for(unsigned int j=0,count=0; j<names.size()-1; j++) {
				for(; count<names[j+1].second; count+=gGEN_SIZE) {
					if(gens[i][count/gGEN_SIZE] > 0)
						out << names[j].first << "\t" << count - (names[j].second) << "\t" << gens[i][count/gGEN_SIZE] << endl;
					//if(a_spots[count/gGEN_SIZE] > 0)
						//out << names[i].first << "\t" << count - (names[i].second) << "\t" << a_spots[count/gGEN_SIZE] << endl;
				}

			}
			out.close();
		}
	}
	*/
}

unsigned long Genome::size() {
	return genome_size * gGEN_SIZE;
}

bool Genome::Test(ostream &os) {
	bool success = true;
	float gamma_val = 4/2.0;
	fprintf(stderr,"Gamma(%d/2=%f)=%f\n",4,gamma_val,tgamma(gamma_val));
	fprintf(stderr,"dchisq(%d,%d)=%f\n",4,1,dchisq(4,1));
	Genome gensnp;
	int i=0;
	float x[] = {0,0.000001, 0.008909,0, 0, 0};
	float pval = gensnp.is_snp(x,i);
	fprintf(stderr,"PVAL: %f at %d\n",pval,i);
	return success;
	 
	unsigned char* genome = new unsigned char[1000000];
	ifstream in("test/short.fa");
	string name;
	getline(in,name);
	in.read((char*)genome,1000000);
	
	//&genome.front() = "actgctgccggtN\ncccgtcggcaacgtgNNNNNgaa\n\0";
	
	try {
		os << "**TESTING HASH_AND_STORE**\n";

		unsigned int prev_MER_SIZE = gMER_SIZE;
		gMER_SIZE = 10;
		gHASH_SIZE = (unsigned int) pow(4.0,gMER_SIZE);
		Genome genYumei;
		char* fn = "test/yumei_test.fa";
		genYumei.use(fn);
		genYumei.LoadGenome();
		os << "Beginning string: " << genYumei.GetString(0,1000) << endl;
		os << endl;
		gMER_SIZE = prev_MER_SIZE;
		gHASH_SIZE = (unsigned int) pow(4.0,gMER_SIZE);

		Genome gen45;
		gREAD = true;
		gREAD_FN = "/data/public/genomes/human/gnumap/gnumap_human_s2_h100k_m11";
		gen45.use("/data/public/genomes/human/gnumap/gnumap_human_s2_h100k_m11");
		gen45.LoadGenome();
		os << "Getting string at the end of chr4: " << 1877852952-200 << " " << gen45.GetPos(1877852952-200,true) << endl;
		os << gen45.GetString(1877852952-200,200) << endl;
		os << "Getting string at the end of chrX: " << 2866253375-200 << " " << gen45.GetPos(2866253375-200,true) << endl;
		os << gen45.GetString(2866253375-200,200) << endl;
		gen45.AddScore(2866253375,4.76);
		os << "Getting value at 2866253375: " << gen45.GetScore(2866253375) << endl;
		os << "                    Should be 4.76" << endl;
		gREAD = false;
		
	}
	catch(const char* erro) {
		os << "\t********Error! Did you include the proper test files?" << endl;
	}
	catch (...) {
		os << "Unknown Error!" << endl;
	}

	delete[] genome;
	return success;
}
