#include "HMMScoredSeq.h"

void HMMScoredSeq::Init() {
		
		VF initDistrib (NumMatrixTypes);
		VF gapOpen (2*NumInsertStates);
		VF gapExtend (2*NumInsertStates);
		VVF emitPairs (256, VF (256, 1e-10));
		VF emitSingle (256, 1e-5);

  
  // read initial state distribution and transition parameters
    if (NumInsertStates == 1){
      for (int i = 0; i < NumMatrixTypes; i++) initDistrib[i] = initDistrib1Default[i];
      for (int i = 0; i < 2*NumInsertStates; i++) gapOpen[i] = gapOpen1Default[i];
      for (int i = 0; i < 2*NumInsertStates; i++) gapExtend[i] = gapExtend1Default[i];
    }
    else if (NumInsertStates == 2){
      for (int i = 0; i < NumMatrixTypes; i++) initDistrib[i] = initDistrib2Default[i];
      for (int i = 0; i < 2*NumInsertStates; i++) gapOpen[i] = gapOpen2Default[i];
      for (int i = 0; i < 2*NumInsertStates; i++) gapExtend[i] = gapExtend2Default[i];
    }
    else {
      cerr << "ERROR: No default initial distribution/parameter settings exist" << endl
           << "       for " << NumInsertStates << " pairs of insert states.  Use --paramfile." << endl;
      exit (1);
    }

    string alphabet = alphabetDefault;

    for (int i = 0; i < (int) alphabet.length(); i++){
      emitSingle[(unsigned char) tolower(alphabet[i])] = emitSingleDefault[i];
      emitSingle[(unsigned char) toupper(alphabet[i])] = emitSingleDefault[i];
      for (int j = 0; j <= i; j++){
    emitPairs[(unsigned char) tolower(alphabet[i])][(unsigned char) tolower(alphabet[j])] = emitPairsDefault[i][j];
    emitPairs[(unsigned char) tolower(alphabet[i])][(unsigned char) toupper(alphabet[j])] = emitPairsDefault[i][j];
    emitPairs[(unsigned char) toupper(alphabet[i])][(unsigned char) tolower(alphabet[j])] = emitPairsDefault[i][j];
    emitPairs[(unsigned char) toupper(alphabet[i])][(unsigned char) toupper(alphabet[j])] = emitPairsDefault[i][j];
    emitPairs[(unsigned char) tolower(alphabet[j])][(unsigned char) tolower(alphabet[i])] = emitPairsDefault[i][j];
    emitPairs[(unsigned char) tolower(alphabet[j])][(unsigned char) toupper(alphabet[i])] = emitPairsDefault[i][j];
    emitPairs[(unsigned char) toupper(alphabet[j])][(unsigned char) tolower(alphabet[i])] = emitPairsDefault[i][j];
    emitPairs[(unsigned char) toupper(alphabet[j])][(unsigned char) toupper(alphabet[i])] = emitPairsDefault[i][j];
      }
    }

    // build new model for aligning
    model = new ProbabilisticModel (initDistrib, gapOpen, gapExtend, emitPairs, emitSingle);
}

string HMMScoredSeq::print_full(double denom, Genome &gen, Read &read) const {

	if(!positions.size())	//if there were no matches, just return.
		return "";

	ostringstream temp;
	
	double total_score = log_align_score / denom;
	
	//we only need to compute this alignment string once because, while there may be many locations for each
	//sequence, we're ensuring that the sequences are the exact same.

	bin_seq bs;
	unsigned int seq_size = sequence.size();

	string read_consensus = GetConsensus(read);

	string gen_string = gen.GetString((*positions.begin()).first, seq_size);
	string aligned = bs.get_align_score_w_traceback(read, read_consensus, gen_string);

	//temp << align_score << "\t" << log_align_score << "\t" << positions.size() << "\t" << total_score << "\t" << aligned << "\t";
	temp << align_score << "\t" << positions.size() << "\t" << total_score << "\t" << aligned;
		
	set<pair<unsigned long,bool> >::iterator sit;	
	for(sit = positions.begin(); sit != positions.end(); sit++)
		temp << "\t" << gen.GetPos(*sit);
		
	return temp.str();
}

void HMMScoredSeq::score(double denom, Genome &gen, unsigned int len, Read &read, pthread_mutex_t &lock) {

	if(!positions.size())	//if there were no matches, just return.
		return;

	double total_score = log_align_score / denom;
	//set<pair<unsigned int,bool> >::iterator sit;

	//we only need to compute this alignment string once because, while there may be many locations for each
	//sequence, we're ensuring that the sequences are the exact same.

	bin_seq bs;	// global one
	unsigned int seq_size = sequence.size();

	string read_consensus = GetConsensus(read);

	string gen_string = gen.GetString((*positions.begin()).first, seq_size);
	string aligned = bs.get_align_score_w_traceback(read, read_consensus, gen_string);

	string name = read.name;
	Sequence seq1(read_consensus, name, read_consensus.length());
	string header_blank = "";
	Sequence seq2(gen_string, header_blank, gen_string.length());

	// compute forward and backward probabilities
	VF *forward = model->ComputeForwardMatrix (&seq1, &seq2); assert (forward);
	VF *backward = model->ComputeBackwardMatrix (&seq1, &seq2); assert (backward);

	// compute posterior probability matrix
	VF *posterior = model->ComputePosteriorMatrix (&seq1, &seq2, *forward, *backward); assert (posterior);
	pair<SafeVector<char> *, float> alignment = model->ComputeAlignment (seq1.GetLength(),
									seq2.GetLength(),
									*posterior);

	set<pair<unsigned long,bool> >::iterator sit;
	pthread_mutex_lock(&lock);
	// Iterate over all the positions (they'll have the same sequence)
	for(sit=positions.begin(); sit!=positions.end(); sit++) {
		for(unsigned int i=0; i<posterior->size(); i++) {
			// Add it to the total Genome score
			//gen.AddScore( (*sit).first+i,total_score );
			// Add it to the individual nucleotide scores
			//gen.AddSeqScore((*sit).first+i,total_score,aligned[i]);

		}
	}
	pthread_mutex_unlock(&lock);

	fprintf(stderr,"Posterior size: %d\n",posterior->size());
	fprintf(stderr,"%d and %d\n",seq1.GetLength(),seq2.GetLength());
	int totalLen = (seq1.GetLength()+1) * (seq2.GetLength()+1);
	int seq1Length = seq1.GetLength();
	int seq2Length = seq2.GetLength();
	fprintf(stderr,"This is after the supposed illegal instruction! (%d)\n",totalLen);

	int ij = 0;
	for(int i=0; i<=seq1Length; i++) {
		fprintf(stdout,"%d",i);
		for(int j=0; j<=seq2Length; j++) {
			float temp = (*posterior)[ij];
			if(temp <= 0)
				fprintf(stdout,"\t%d: ________",j);
			else 
				fprintf(stdout,"\t%d: %f",j, temp);
			ij += NumMatrixTypes;
		}
		fprintf(stdout,"\n");
	}

	delete forward;
	delete backward;
	delete posterior;
}
