#include "SNPScoredSeq.h"

string SNPScoredSeq::print_full(double denom, Genome &gen, Read &read) const {

	if(!positions.size())	//if there were no matches, just return.
		return "";

	ostringstream temp;
	
	double total_score = log_align_score / denom;
	
	//we only need to compute this alignment string once because, while there may be many locations for each
	//sequence, we're ensuring that the sequences are the exact same.

	bin_seq bs;
	unsigned int seq_size = sequence.size();

	string read_consensus = GetConsensus(read);

	string gen_string = gen.GetString((*positions.begin()).first, seq_size);

	// If it's on the bottom strand, we want the revcom of this
	if((*positions.begin()).second == false)
		gen_string = reverse_comp(gen_string);
	
	string aligned = bs.get_align_score_w_traceback(read, read_consensus, gen_string).first;

	//temp << align_score << "\t" << log_align_score << "\t" << positions.size() << "\t" << total_score << "\t" << aligned << "\t";
	temp << align_score << "\t" << positions.size() << "\t" << total_score << "\t" << aligned;
		
	set<pair<unsigned long,bool> >::iterator sit;	
	for(sit = positions.begin(); sit != positions.end(); sit++)
		temp << "\t" << gen.GetPos(*sit);
		
	return temp.str();
}

void SNPScoredSeq::score(double denom, Genome &gen, unsigned int len, Read &read, pthread_mutex_t &lock) {

	if(!positions.size())	//if there were no matches, just return.
		return;

	double total_score = log_align_score / denom;
	//set<pair<unsigned int,bool> >::iterator sit;

	//we only need to compute this alignment string once because, while there may be many locations for each
	//sequence, we're ensuring that the sequences are the exact same.

	bin_seq bs;	// global one
	unsigned int seq_size = sequence.size();

	string read_consensus = GetConsensus(read);

	string gen_string = gen.GetString((*positions.begin()).first, seq_size);

/*
	// If it's on the bottom strand, we want the revcom of this
	if((*positions.begin()).second == false) {
		printf("It's false: %s\n",read_consensus.c_str());
		read_consensus = reverse_comp(read_consensus);
		reverse_comp(read.pwm,read.length);

	}
*/
	if((*positions.begin()).second == false) {
		gen_string = reverse_comp(gen_string);
	}
	
#ifdef PHMM
	float** hmm_aligned = bs.pairHMM(read, read_consensus, gen_string);
	float** hmm_rev_aligned = reverse_comp_cpy_n(hmm_aligned, read.length);
	
	set<pair<unsigned long,bool> >::iterator sit;
	pthread_mutex_lock(&lock);
	for(sit=positions.begin(); sit!=positions.end(); sit++) {
		for(unsigned int i=0; i<gen_string.size(); i++) {
			gen.AddScore( (*sit).first+i,total_score );
			if((*sit).second)
				gen.AddSeqScore((*sit).first+i, hmm_aligned[i], total_score);
			else
				gen.AddSeqScore((*sit).first+i, hmm_rev_aligned[i], total_score);
		}
	}
	pthread_mutex_unlock(&lock);
#else
	string aligned = bs.get_align_score_w_traceback(read, read_consensus, gen_string).first;
	string rev_aligned = reverse_comp(aligned);
	
	set<pair<unsigned long,bool> >::iterator sit;
	pthread_mutex_lock(&lock);
	for(sit=positions.begin(); sit!=positions.end(); sit++) {
		for(unsigned int i=0; i<aligned.size(); i++) {
			// Add it to the total Genome score
			gen.AddScore( (*sit).first+i,total_score );
			// Add it to the individual nucleotide scores
			if((*sit).second)
				gen.AddSeqScore((*sit).first+i,total_score,aligned[i]);
			else
				gen.AddSeqScore((*sit).first+i,total_score,rev_aligned[i]);
	
		}
	}
	pthread_mutex_unlock(&lock);
#endif

	/*
		if((*positions.begin()).second == false) {
			// Change it back to how it was
			reverse_comp(read.pwm,read.length);
		}
	*/
			
}
