#include <iostream>
#include <unistd.h>
#include <stdio.h>
#include <vector>
//#include "gvector.h"
#include <string>
#include <sstream>
#include <cstring>
#include <ctime>
#include <cmath>
#include <set>
#include <map>
#include <pthread.h>
#include <algorithm>
#include <exception>
#include "const_include.h"
#include "GenomeMem.h"
#include "bin_seq.h"
#include "SeqReader.h"
#include "ScoredSeq.h"
#include "BSScoredSeq.h"
#include "NormalScoredSeq.h"
#include "SNPScoredSeq.h"
#include "SequenceOperations.h"
//#include "HMMScoredSeq.h"
#include "Exception.h"

typedef set<ScoredSeq*,ScoredSeq > seq_set;
typedef map<string,ScoredSeq* > seq_map;

unsigned int gBISULFITE = 0;
unsigned int gATOG = 0; 
unsigned int gSNP = 0;
unsigned char gINT2BASE[16];
unsigned char g_bs_CONVERSION[256];
unsigned char g_gen_CONVERSION[256];

int gVERBOSE=0;
int gMER_SIZE=0;
unsigned int gMAX_MATCHES = 100;
unsigned int gREAD_BUFFER = 1024 * 5;
unsigned int gBUFFER_SIZE = 100000;
unsigned int gHASH_SIZE = 1048576;	//gHASH_SIZE is set as 4^10
//unsigned int gMAX_HASH_SIZE = 1000;
unsigned int gMAX_HASH_SIZE = 0;
float gALIGN_SCORE = 10;

float gGAP = -2;
float gMATCH = 1;
float gMISMATCH = -1;

int gMAX_GAP = 3;
const char* seq_file = 0;
const char* genome_file = 0;
const char* output_file = "gnumap_out";
const char* pos_matrix = NULL;
char* g_adaptor = NULL;
int gPRINT_FULL=0;
float gALIGN_SCORES[256][4];
bool perc = false;
int gSEQ_LENGTH=-1;

bool fast = false;	// Do we perform a fast alignment or a more thorough one?

float gMIN_ADAPTOR_DIFF = 0.75;	//used to define the minimum difference from the given g_adaptor sequence to the consensus
unsigned int gMIN_CHOPPED_BASES = 4;	//if there's less than 4 bases that match the g_adaptor, we're not going to chop them.
int gGEN_SKIP = 0;	// The number of bases we skip as we hash the genome
unsigned int gGEN_SIZE = 8;	// The number of bases each GEN segment contains (per each unsigned int)
unsigned int gJUMP_SIZE = 0;	// The number of bases we jump in a read (per hash)

int gSGREX = false;
bool gSAM = false;

bool gILLUMINA = false;

//I'm not sure what BIN_SIZE means...It's not used anymore--only appearance is in Genome.cpp:365
//int BIN_SIZE = 10;
int gBIN_SIZE = 1;
 
int gNUM_THREADS = 1;

// gSAVE and gREAD are global variables used to indicate whether the hashed genome should be
// read and saved or created from fasta files.
bool gSAVE=false;
bool gREAD=false;
char* gSAVE_FN;
char* gREAD_FN;

int bad_files =0;

//static bin_seq bs;

pthread_attr_t attr;
pthread_mutex_t lock;

pthread_mutex_t read_lock;
// The writer thread
pthread_t writer;
pthread_mutex_t write_lock;

void* thread_print(void* t_opts);
void printPWM(Read* read);

/*****************************************************
 * These next functions are used in parsing command-
 * line arguments
 *****************************************************/
int ParseCmdLine(const int argc, const char* argv[]);
void GetParseError(ostream &os, const char* argv[]);
int set_arg_ext(const char* param, const char* assign, int &count);
int set_arg(const char* param, const char* assign, int &count);
/*struct CMDLine {
	int errno;
	int errpos;
} this_cmd;*/
CMDLine this_cmd;
/*****************************************************/

void usage(int has_error, char* errmessage) {
	if(has_error) cout << endl << errmessage << endl ;

	cout << "Usage: gnumap [options] <file_to_parse>\n"
		 << "  -g, --genome=STRING          Genome .fa file(s)\n"
		 << "  -o, --output=STRING          Output file\n"
		 << "  -a, --align_score=DOUBLE     Limit for sequence alignment (default: 10)\n"
		 << "  -p, --percent=NONE           Use percentage when determining alignment score\n"
		 << "  -v, --verbose=INT            Verbose (default=0)\n"
		 << "  -c, --num_proc=INT           Number of processors on which to run\n"
		 << "  -m, --mer_size=INT           Mer size (default=10)\n"
		 << "  -B, --buffer=INT             Buffer size\n"
		 << "  -T, --max_match=INT          Maximum number of matches for a given\n"
		 << "                               sequence (default: 100)\n"
		 << "  -h, --max_hash=INT           Maximum values to store in the hash.\n"
		 << "                               (default: none )\n"
		 << "  -S, --subst_file=STRING      Position-Weight Matrix file for DNA\n"
		 << "                               Substitutions\n"
		 << "  -G, --gap_penalty=DOUBLE     Gap Penalty for Substitution Matrix\n"
		 << "  -M, --max_gap=INT            Maximum Number of Gaps to use in Alignment\n"
		 << "  -A, --adaptor=STRING         Adaptor sequence to remove from sequences\n"
		 << "  -0, --print_full             Print locations for the entire sequence, not\n"
		 << "                               just for the beginning.\n"
		 << "  -b, --bs_seq                 Flag to turn on the C to T conversion, used in\n"
		 << "                               bisulfite sequence analysis\n"
		 << "  -d, --a_to_g                 Flag that allows for A to G conversion\n"
		 << "  --fast                       Perform a fast alignment (at some reduction\n"
		 << "                               of accuracy)\n"
		 << "  -s, --gen_skip=INT           Number of bases to skip when the genome is aligned\n"
		 << "                               (default: none)\n"
		 << "  --bin_size                   The resolution for GNUMAP (default: 8)\n"
		 << "  -j, --jump                   The number of bases to jump in the sequence hashing\n"
		 << "                               (default: mer_size)\n"
		 << "  --snp                        Turn on SNP mapping (will output a .sgrex file)\n"
		 << "  --illumina                   Defines the fastq file as Illumina file (otherwise\n"
		 << "                               does nothing)\n"
		 << "  --SAM                        Output is in SAM format (will also output the \n"
		 << "                               .sgr or .sgrex file)\n"
		 << "\n"
		 << "For reading and writing to a binary file:\n"
		 << "  --save=FILENAME              Save the genome out to a file\n"
		 << "  --read=FILENAME              Read the genome in from a file\n"
		 << endl
		 << "Help options:\n"
		 << "  -?, --help                   Show this help message\n";
	exit(1);
}

/*
 * Because this is used to produce the consensus sequence, we need ambiguity characters...
 */
//inline char max_char(vector<double> &chr) {
inline char max_char(float* chr) {
	if((chr[0] == chr[1]) && (chr[0] == chr[2]) && (chr[0] == chr[3]))
		return 'n';
	if(chr[0] >= chr[1]) {
		if(chr[0] >= chr[2]) {
			if(chr[0] >= chr[3])
				return 'a';
			else
				return 't';
			}
		else {
			if(chr[2] >= chr[3])
				return 'g';
			else
				return 't';
		}
	}
	else {
		if(chr[1] >= chr[2]) {
			if(chr[1] >= chr[3])
				return 'c';
			else
				return 't';
		}
		else {
			if(chr[2] >= chr[3])
				return 'g';
			else return 't';
		}
	}

	return 'n';
}
 
//inline string GetConsensus(vector<vector<double> > &read) {
inline string GetConsensus(Read &read) {
	string seq = "";

	for(unsigned int i=0; i<read.length; i++) {
		seq += max_char(read.pwm[i]);
	}

	return seq;
}

//I don't think this function is even used...
//vector<vector<double> > subvector(vector<vector<double> > &vec, int start, int length) {
float** subvector(float** &vec, int start, int length) {
	//vector<vector<double> > to_return;
	float** to_return = new float*[length];

	for(int i=start,j=0; i<start+length; i++,j++) {
		to_return[j] = new float[4];
		for(int k=0; k<4; k++) {
			to_return[j][k] = vec[i][k];
		}
		delete[] vec[i];
	}
	
	delete[] vec;	//we want to free the memory for vec because we're creating a new
					//one, right??
	
	return to_return;
}

struct thread_opts {
	//thread_opts(Genome &g, vector<vector<double> >* s, unsigned int n) :
	thread_opts(Genome &g, Read* s[], unsigned int n, int ti) :
		gen(g), sr(*(new SeqReader())), of(*(new ofstream)), 
		seqs(s), num_seqs(n), thread_id(ti) {}
	thread_opts(Genome &g, SeqReader &s, ofstream &stream, int ti) :
		gen(g), sr(s), of(stream), thread_id(ti) {}
	
	Genome& gen;
	SeqReader &sr;
	ofstream &of;
	//vector<vector<double> >* seqs;
	Read** seqs;
	unsigned int num_seqs;
	int thread_id;
};

struct thread_rets {
	ostringstream* of;
	unsigned int good_seqs;
	unsigned int bad_seqs;
};


/*
inline string reverse_comp(string &s) {
	string temp = "";
	int len = s.length();
	
	for(int i=1; i<=len; i++) {
		switch(s[len-i]) {
			case 'a':
				temp += 't';
				break;
			case 't':
				temp += 'a';
				break;
			case 'c':
				temp += 'g';
				break;
			case 'g':
				temp += 'c';
				break;
			default:
				temp += 'n';
				break;
		}
	}
	
	return temp;
} 

//we modify the former one
//inline vector<vector<double> > reverse_comp(vector<vector<double> > &v) {
inline void reverse_comp(float** pwm, unsigned int length) {
	float temp_pwm[length][4];


	//for(r_vit = v.rbegin(); r_vit != v.rend(); ++r_vit) {
	for(unsigned int i = 0; i < length; ++i) {
		//need to switch places with the characters: a-t, g-c
		//also need to rev-com it.
		temp_pwm[length-1-i][A_POS] = pwm[i][T_POS];
		temp_pwm[length-1-i][C_POS] = pwm[i][G_POS];
		temp_pwm[length-1-i][G_POS] = pwm[i][C_POS];
		temp_pwm[length-1-i][T_POS] = pwm[i][A_POS];
	}
		
	for(unsigned int i=0; i<length; i++) {
		for(unsigned int j=0; j<4; j++) {
			pwm[i][j] = temp_pwm[i][j];
		}
	}
	
	//cout << "temp's size: " << temp.size() << endl;
	//cout << "originals size: " << v.size() << endl; 
}
*/

//void clear(seq_set &to_clear) {
void clear(seq_map &to_clear) {
	//seq_set::iterator sit;
	seq_map::iterator sit;
	
	for(sit = to_clear.begin(); sit != to_clear.end(); ++sit) { 
		//delete(*sit);
		delete (*sit).second;
	}
}

bool align_sequence(Genome &gen, seq_map &unique, const Read &search, const string &consensus, 
				double min_align_score, double &denominator,
				bool upstrand) {
				
	// We don't want to cover every search string.
	// We'll cover each gMER_SIZE segment twice.
	//unsigned int adjust = JUMP_SIZE;
	//if(fast)
		//adjust = gMER_SIZE;

	bin_seq bs;
	
	unsigned int i,j,last_hash=search.length-gMER_SIZE;

	for(i=0; i< last_hash; i+=gJUMP_SIZE) {
		HashLocation* hashes = NULL;
		pair<int, unsigned long> p_hash;
		string consensus_piece;

		// If we've run accross a hash location that is clear (possibly because we've
		// cleared it in a previous step of the algorithm), we want to get another hash
		// that's right next to it instead of jumping over everything.
		for(j=0; j+i<last_hash; j++) {
			consensus_piece = consensus.substr(i+j,gMER_SIZE);
			p_hash = bs.get_hash(consensus_piece);
			hashes = gen.GetMatches(p_hash.second);
			
			// p_hash.first would only be zero if there were n's in the sequence.
			// Just step to a location that doesn't (?) have any n's.
			if(!p_hash.first)  {
#ifdef DEBUG
				if(gVERBOSE > 2)
					cout << "Invalid hash sequence at step " << j+i << endl;
#endif
				continue;
			}

			if(hashes->size)	// we've found a valid location
				break;
#ifdef DEBUG
			if(gVERBOSE > 2)
				cout << "At location " << j+i << ", no valid hashes found." << endl;
#endif
		}
		if(!hashes->size) {	// There are no valid locations for this specific hash
#ifdef DEBUG
			if(gVERBOSE > 2)
				cout << "No valid hashes for this sequence." << endl;
#endif
			break;	// Go to the next location
		}


#ifdef DEBUG
		cout << "Found " << hashes->size << " hashes for this string." << endl;
#endif
		
		for(unsigned int vit = 0; vit<hashes->size; ++vit) {
			//Match the sequence to a sequence that has two characters before and after.
			//string to_match = gen.GetString((*vit)-i-2, search.size()+4);
			
			unsigned long beginning = (hashes->hash_arr[vit]<=i) ? 0 : (hashes->hash_arr[vit]-i);
			
			// Check to see if we've already checked this location
			//if(gen.getSetThreadID(beginning,pthread_self()))
				//continue;
			
			string to_match = gen.GetString(beginning, search.length);
			
			// QUINN: Build up the string array
			
			if(to_match.size() == 0)	//means it's on a chromosome boundary--not valid sequence
				continue;

			double align_score = 0;
			char aligned[to_match.length()];
			strcpy(aligned,to_match.c_str());

			//if(gBISULFITE)
				//align_score = bs.get_align_score_w_traceback(search,to_match,aligned);
			//else
				align_score = bs.get_align_score(search,to_match,i,i+gMER_SIZE-1);
				//align_score = bs.get_align_score(search,to_match);

#ifdef DEBUG
			if(gVERBOSE > 1) 
				cout << "matching: " << to_match << endl
					 << "  with  : " << consensus << "\tat pos   " << beginning 
					 							  << "\tand step " << i
												  << "\tand alignment " << align_score << endl;
#endif
					 				
			if(align_score >= min_align_score) {
				//if(align_score >= (this_align_score + 4.6))
					//this_align_score = align_score-2;
				
				ScoredSeq* temp;
				
				if(gSNP)
					temp = 
						new SNPScoredSeq(to_match, align_score, beginning, upstrand);
						//new HMMScoredSeq(to_match,align_score, beginning, upstrand);
				else if(gBISULFITE)
					temp =
						new BSScoredSeq(to_match,align_score, beginning, upstrand);
				else
					temp =
						new NormalScoredSeq(to_match,align_score, beginning, upstrand);
						
				//fprintf(stderr,"%f\n",exp(align_score));
				if(!upstrand)
					to_match = reverse_comp(to_match);
				pair<seq_map::iterator,bool> it_bool = unique.insert(pair<string,ScoredSeq*>(to_match,temp));

				if(!it_bool.second) {	//If it wasn't inserted into the set--because 
										//	it had already been inserted...
					
					//int c = unique.count(*it_bool.first);
					//cout << "count: " << c << endl;
					//if(c > 1)
						//cout << "count: " << c << " " << (*it_bool.first)->print() << endl;
					delete temp;
					if( (*it_bool.first).second->add_spot(beginning,upstrand) )
						denominator += exp(align_score);
				}
				else
					denominator += exp(align_score);
				
			}
		}
		
		if(unique.size() > gMAX_MATCHES) {
			return false;
		}
	
	}
	
	return true;
}


//string print_top_matches(Genome &gen, vector<vector<double> > &search, string &consensus) {
string print_top_matches(Genome &gen, Read &search, string &consensus, 
			unsigned int &num_not_matched, unsigned int &num_matched) {
	
	bin_seq bs;
	double min_align_score = 0;
	ostringstream to_return;

	if(!gSAM)
		to_return << consensus << "\t";

	if(search.length < (unsigned int)gMER_SIZE) { 
		if(gSAM) {
			string QUAL = str2qual(search);
			to_return << search.name << "\t" << 0x0200 << "\t*\t0\t0\t*\t=\t0\t0\t" << consensus << "\t";
			to_return << QUAL.c_str() << "\n";
		}
		
		else
			to_return << "Too short to align\n";

		num_not_matched++;
		return to_return.str();
	}


	// We need to do the length-1 because it will go to the base at that position.
	double max_align_score = bs.get_align_score(search,consensus,0,search.length-1);
	//double max_align_score = bs.get_align_score(search,consensus);
	if(!gSAM)
		to_return << max_align_score << "\t"; 

	//bad idea:--for testing purposes only.
	//return to_return.str();
	//end of bad idea
	
#ifdef DEBUG
	if(gVERBOSE > 2) {
		cout << "\nNext Read:" << endl;
		cout << consensus << "\t" << max_align_score << "\n";
	}
#endif

	if(max_align_score < 0) { 
		num_not_matched++;
		if(gSAM) {
			string QUAL = bs.str2qual(search);
			to_return << search.name << "\t" << 0x0200 << "\t*\t0\t0\t*\t=\t0\t0\t" << consensus << "\t";
			to_return << QUAL.c_str() << "\n";
		}
		else
			to_return << "Not possible to obtain a positive alignment score\n";

		return to_return.str();
	}

	if(perc)
		min_align_score = gALIGN_SCORE * max_align_score;
	else
		min_align_score = gALIGN_SCORE;

	if(!gSAM)
		to_return << min_align_score << "\t";
	
	seq_map unique;  
	
	double denominator = 0;

	bool aligned = align_sequence(gen, unique, search, consensus, 
								min_align_score, denominator,
								true);
	if(!aligned) {
		clear(unique);
		num_matched++;
		if(!gSAM)
			to_return << gMAX_MATCHES << "+ (Too many matches to compute.)\n";
		return to_return.str();
	}

	/************************************************************/
	/* We need to do the reverse of this as well...				*/
	/************************************************************/

	// We're doing the reverse strand
	//consensus = reverse_comp(consensus);
	string rev_consensus = reverse_comp(consensus);
	//search.pwm = reverse_comp(search.pwm,search.length);
	//will modify the search.pwm reference
	//reverse_comp(search.pwm,search.length);
	float** rev_pwm = reverse_comp_cpy(search.pwm,search.length);
	Read rc(rev_pwm,search.length);


	//printPWM(&search);
#ifdef DEBUG
	if(gVERBOSE > 2) {
		cout << "\tPerforming reverse compliment alignment with sequence " << consensus << endl;
	}
#endif
	
	//aligned = align_sequence(gen, unique, search, consensus, 
	//						min_align_score, denominator,
	//						false);
	aligned = align_sequence(gen, unique, rc, rev_consensus, 
							min_align_score, denominator,
							false);

	// Clean up the copy
	for(unsigned int i=0; i<rc.length; i++) {
		delete[] rev_pwm[i];
	}
	delete[] rev_pwm;

	if(!aligned) {
		clear(unique);
		num_matched++;

		if(!gSAM)
			to_return << gMAX_MATCHES << "+ (Too many matches to compute.)\n";

		return to_return.str();
	}


//AFTER_REV:
	
	// Don't need to change back because we're creating a local reference above.
	// Change it back to the way it was before
	//consensus = reverse_comp(consensus);
	//will modify the search.pwm reference
	//reverse_comp(search.pwm,search.length);

	//to_return << endl;
	//to_return << "Denominator: " << denominator << endl;
	//seq_set::iterator sit;
	seq_map::iterator sit;

	ScoredSeq *max = new NormalScoredSeq;	// But empty...

	unsigned int len = gPRINT_FULL ? consensus.length() : 1;
	
	pthread_mutex_lock(&lock);
	for(sit = unique.begin(); sit != unique.end(); ++sit) {

		(*sit).second->score(denominator,gen,len,search);

		// Free up the record
		if( (*(*sit).second).is_greater(*max,denominator)) {
			ScoredSeq *temp = max;
			max = (*sit).second;
			delete temp;
		}
		else 
			delete (*sit).second;
	}
	pthread_mutex_unlock(&lock);
	
	if(!gSAM) {
		to_return << unique.size() << "\t";
		to_return << max->print_full(denominator,gen,search) << "\n";
	}
	else {
		to_return << max->print_SAM(denominator,search,consensus,gen);
	}

	delete max;
	
	//string str = to_return.str();
	//str = "\nHighest Score:\n";
	//string str = "(" << unique.size() << ")" << "\t";
	//str += max.print_full_w_score(denominator);

	if(unique.size() == 0)
		num_not_matched++;
	else
		num_matched++;

	return to_return.str();
	//return str;
}

/**
 * This function will print the corresponding strings for all locations at this hash
 */
string print(Genome &gen, HashLocation* pos, long begin, int size) {
	string to_return = "";
	
	//cout << "matches: " << pos.size() << endl;
	for(unsigned long i=0; i<pos->size; i++) {
		to_return += gen.GetString(pos->hash_arr[i] - begin, size) + "\t";
	}
	
	return to_return;
}


/* GetPWM is used for getting a PWM matrix representing the match (m) and mismatch (mm)
 * penalties.
 * The input is a file, which should be a 4x4 matrix of the format:
 * 				 A   C   G   T
 * 			A    m  mm  mm  mm
 *			C   mm   m  mm  mm
 * 			G   mm  mm   m  mm
 * 			T   mm  mm  mm   m
 * For ease in reading, the letters defining the rows and columns are disregarded.
 *
 *
 * Note:  GetPWM will modify the global gALIGN_SCORES variable.
 */
void getPWM(const char* fn) {
	float temp_ALIGN_SCORES[5][4];

	int THIS_BUFFER = 100;
	
	ifstream in;
	in.open(fn);
	char temp_chr[THIS_BUFFER];
	bool contains_labels=false;
	float a,c,g,t;
	char label[THIS_BUFFER];
	char alabel[THIS_BUFFER];
	char clabel[THIS_BUFFER];
	int count=0;
	int num_lines=4;

	// Get all the lines of input
	while(in.getline(temp_chr,THIS_BUFFER) && count < num_lines) {
		if(sscanf(temp_chr," %s %s %s %s",alabel,clabel,label,label) == 4 
				&& (tolower(*alabel) == 'a')
				&& (tolower(*clabel) == 'c') ) {
			if(gVERBOSE > 1)
				printf("Matched here: %s\n",temp_chr);
			in.getline(temp_chr,THIS_BUFFER);
			contains_labels=true;
		}
		if(gVERBOSE > 1)
			cout << "Line: " << temp_chr << endl;

		if(contains_labels) {
			if(sscanf(temp_chr,"%s %f %f %f %f",label,&a,&c,&g,&t) != 5) {
				char* error = new char[THIS_BUFFER];
				strcat(error,"Error in Score File: ");
				strcat(error,temp_chr);
				throw(error);
			}
		}
		else {
			if(sscanf(temp_chr,"%f %f %f %f",&a,&c,&g,&t) != 4) {
				char* error = new char[THIS_BUFFER];
				strcat(error,"Error in Score File: ");
				strcat(error,temp_chr);
				throw(error);
			}
		}
		
		temp_ALIGN_SCORES[count][0] = (float)a;
		temp_ALIGN_SCORES[count][1] = (float)c;
		temp_ALIGN_SCORES[count][2] = (float)g;
		temp_ALIGN_SCORES[count][3] = (float)t;

		count++;
	}
	if(count < num_lines) {
		throw new Exception("Error in Score File:  Not enough lines");
	}

	cout << "Count: " << count << "\tNum lines: " << num_lines << endl;

#ifdef DEBUG
	//for testing purposes...
	if(gVERBOSE > 1) {
		cout << "Matrix: " << endl;
		cout << "\t0\t1\t2\t3" << endl;
		for(int i=0; i<5; i++) {
			cout << i << "\t";
			for(int j=0; j<4; j++) 
				cout << temp_ALIGN_SCORES[i][j] << "\t";
			cout << endl;
		}
		cout << endl;
		throw new Exception("Passed.");
	}
#endif

	for(unsigned int i=0; i<4; i++) {
		gALIGN_SCORES[(unsigned int)'a'][i] = temp_ALIGN_SCORES[0][i];
		gALIGN_SCORES[(unsigned int)'c'][i] = temp_ALIGN_SCORES[1][i];
		gALIGN_SCORES[(unsigned int)'g'][i] = temp_ALIGN_SCORES[2][i];
		gALIGN_SCORES[(unsigned int)'t'][i] = temp_ALIGN_SCORES[3][i];
	}
}

void printPWM(Read* read) {
	printf("Name: %s\n",read->name);
	printf("Length: %d\n",read->length);
	for(unsigned int i=0; i<read->length; i++) {
		printf("%f %f %f %f\n",read->pwm[i][0],read->pwm[i][1],read->pwm[i][2],read->pwm[i][3]);
	}
}

int main(const int argc, const char* argv[]) {
	//Print command line args
	cerr << "This is GNUMAP, Version 1.5.75, for public and private use." << endl;

	memset(g_gen_CONVERSION,4,256);
	g_gen_CONVERSION[(int)'a'] = g_gen_CONVERSION[(int)'A'] = 0;
	g_gen_CONVERSION[(int)'c'] = g_gen_CONVERSION[(int)'C'] = 1;
	g_gen_CONVERSION[(int)'g'] = g_gen_CONVERSION[(int)'G'] = 2;
	g_gen_CONVERSION[(int)'t'] = g_gen_CONVERSION[(int)'T'] = 3;
	g_gen_CONVERSION[(int)'n'] = g_gen_CONVERSION[(int)'N'] = 4;
	g_gen_CONVERSION[(int)'\n'] = g_gen_CONVERSION[(int)'\r'] = 5;
	// Strange white-space characters
	g_gen_CONVERSION[10] = g_gen_CONVERSION[11] = g_gen_CONVERSION[12] = g_gen_CONVERSION[13] = 5;
	g_gen_CONVERSION[(int)'\0'] = 6;
	g_gen_CONVERSION[(int)'>'] = 7;

	memset(g_bs_CONVERSION,4,256);
	g_bs_CONVERSION[(unsigned int)'a'] = g_bs_CONVERSION[(unsigned int)'A'] = 0;
	g_bs_CONVERSION[(unsigned int)'c'] = g_bs_CONVERSION[(unsigned int)'C'] = 1;
	g_bs_CONVERSION[(unsigned int)'g'] = g_bs_CONVERSION[(unsigned int)'G'] = 2;
	g_bs_CONVERSION[(unsigned int)'t'] = g_bs_CONVERSION[(unsigned int)'T'] = 3;
	g_bs_CONVERSION[(unsigned int)'n'] = g_bs_CONVERSION[(unsigned int)'N'] = 4;
	g_bs_CONVERSION[(unsigned int)'\n'] = g_bs_CONVERSION[(unsigned int)'\r'] = 5;
	// Strange white-space characters
	g_bs_CONVERSION[10] = g_bs_CONVERSION[11] = g_bs_CONVERSION[12] = g_bs_CONVERSION[13] = 5;
	g_bs_CONVERSION[(unsigned int)'\0'] = 6;
	g_bs_CONVERSION[(unsigned int)'>'] = 7;

	memset(gINT2BASE,'?',16);
	gINT2BASE[0] = 'a';
	gINT2BASE[1] = 'c';
	gINT2BASE[2] = 'g';
	gINT2BASE[3] = 't';
	gINT2BASE[4] = 'n';


#ifdef DEBUG
	printf("int\tbs\tgen\n");
	for(int i=0; i<256; i++) {
		printf("%d\t%c\t%d\t%d\n",i,(char)i,g_bs_CONVERSION[i],g_gen_CONVERSION[i]);
	}
	printf("int2base\n");
	for(int i=0; i<16; i++) {
		printf("%d\t%c\n",i,gINT2BASE[i]);
	}
#endif

	cout << endl << "Command Line Arguments:  ";
	for(int i=0; i<argc; i++)
		cout << argv[i] << ' ';
	cout << endl;
	
	gVERBOSE = 0;

	if(argc == 1) //means only the bin/gnumap parameter
		usage(0,"");

	if(argc < 2) {
		usage(1, "Need at least a genome and a file to map.");
		exit(1);
	}

	int rc = ParseCmdLine(argc,argv);
	// an error occurred during option processing
	if(rc != 0) {
		GetParseError(cerr, argv);
		return -1;
	}


	//set up alignment matrix

	for(unsigned int i=0; i<256; i++)
		for(unsigned int j=0; j<4; j++)
			gALIGN_SCORES[i][j] = gGAP;
	gALIGN_SCORES[(int)'a'][0] = gMATCH;
	gALIGN_SCORES[(int)'a'][1] = gMISMATCH;
	gALIGN_SCORES[(int)'a'][2] = gMISMATCH;
	gALIGN_SCORES[(int)'a'][3] = gMISMATCH;

	gALIGN_SCORES[(int)'c'][1] = gMATCH;
	gALIGN_SCORES[(int)'c'][0] = gMISMATCH;
	gALIGN_SCORES[(int)'c'][2] = gMISMATCH;
	gALIGN_SCORES[(int)'c'][3] = gMISMATCH;

	gALIGN_SCORES[(int)'g'][2] = gMATCH;
	gALIGN_SCORES[(int)'g'][0] = gMISMATCH;
	gALIGN_SCORES[(int)'g'][1] = gMISMATCH;
	gALIGN_SCORES[(int)'g'][3] = gMISMATCH;

	gALIGN_SCORES[(int)'t'][3] = gMATCH;
	gALIGN_SCORES[(int)'t'][0] = gMISMATCH;
	gALIGN_SCORES[(int)'t'][1] = gMISMATCH;
	gALIGN_SCORES[(int)'t'][2] = gMISMATCH;

/*
	for(int i=0; i<4; i++) {
		for(int j=0; j<4; j++) {
			if(i == j)
				gALIGN_SCORES[i][j] = MATCH;
			else
				gALIGN_SCORES[i][j] = MISMATCH;
		}
		gALIGN_SCORES[4][i] = gGAP;
	}
*/

	/* Manage errors that might occur */	
	// If they don't include a fasta file for sequences
	// If they don't tell us how long the sequences are
	if( (seq_file == NULL) )
		usage(1, "Specify a single file e.g., sequences.fa\n");
	if(!genome_file)
		usage(1, "Specify a genome to map to with the -g flag\n");

	if(pos_matrix != NULL) {
		try {
			getPWM(pos_matrix);
		}
		catch(const char* err) {
			cerr << "ERROR: \n\t" << err << endl;
			return -1;
		}
		catch(Exception *e) {
			cerr << "Error: \n\t" << e->GetMessage() << endl;
			delete e;
			return -1;
		}
	}
	
	ostringstream params;
	params << "# Parameters: " << endl;
	params << "#\tVerbose: " << gVERBOSE << endl;
	if(gREAD)
		params << "#\tGenome file to read from: " << gREAD_FN << endl;
	else 
		params << "#\tGenome file(s): " << genome_file << endl;
	if(gSAVE)
		params << "#\tGenome file to save to: " << gSAVE_FN << endl;
	params << "#\tOutput file: " << output_file << endl;
	params << "#\tSequence file: " << seq_file << endl;
	if(!perc)
		params << "#\tAlign score: " << gALIGN_SCORE << endl;
	else
		params << "#\tAlign percentage: " << gALIGN_SCORE*100 << "%" << endl;
	params << "#\tNumber of threads: " << gNUM_THREADS << endl;
	if(fast) {
		if(!gMER_SIZE)	//if the user didn't specify one
			gMER_SIZE = 14;
		params << "#\tUsing FAST alignment mode" << endl;
	}
	if(gGEN_SKIP != 0) {
		if(gGEN_SKIP < 0)
			usage(1, "-s: Invalid Genome Skip size\n");
		params << "#\tSkipping " << gGEN_SKIP << " bases during hash step" << endl;
	}
	gGEN_SKIP++;	// We're mod'ing in the function, so we want to mod by GEN_SKIP+1
	if(gMAX_HASH_SIZE > 0)
		params << "#\tLargest Hash Size: " << gMAX_HASH_SIZE << endl;
	if(!gMER_SIZE)	//if the user didn't specify one
		gMER_SIZE = 10;
	params << "#\tMer size: " << gMER_SIZE << endl;
	if(gJUMP_SIZE) {
		if(gJUMP_SIZE < 1)
			usage(1, "--jump: Invalid jump size\n");
	}
	else {
		if(fast)
			gJUMP_SIZE = gMER_SIZE;
		else
			gJUMP_SIZE = gMER_SIZE/2;
	}
	params << "#\tUsing jump size of " << gJUMP_SIZE << endl;
	if(pos_matrix != NULL)
		params << "#\tUsing User-Defined Alignment Scores" << endl;
	else 
		params << "#\tUsing Default Alignment Scores" << endl;
	params << "#\tGap score: " << gGAP << endl;
	params << "#\tMaximum Gaps: " << gMAX_GAP << endl;
	if(g_adaptor) {
		params << "#\tUsing Adaptor Sequence: " << g_adaptor << endl;
	}
	if(gBISULFITE) {
		gPRINT_FULL = true;
		params << "#\tUsing BISULFITE conversion" << endl;
		g_bs_CONVERSION[(int)'c'] = g_bs_CONVERSION[(int)'C'] = 3;	//change it so every 'c' will be *hashed* as a 't'
		//gALIGN_SCORES[3][1] = MATCH;
		gALIGN_SCORES[(int)'c'][3] = gMATCH;
	}
	if(gATOG) {
		gPRINT_FULL = true;
		params << "#\tUsing A to G conversion" << endl;
		g_bs_CONVERSION[(int)'a'] = g_bs_CONVERSION[(int)'A'] = 2;	//change it so every 'a' will be *hashed* as a 'g'
		//gALIGN_SCORES[2][0] = MATCH;
		gALIGN_SCORES[(int)'a'][2] = gMATCH;
	}
	if(gGEN_SIZE != 8) {
		if(gGEN_SIZE > 8)
			usage(1,"Invalid bin size (must be 8 or less)\n");
		params << "#\tUsing irregular bin size of " << gGEN_SIZE << endl;
	}

	//determine the size of the hash
	gHASH_SIZE = (unsigned int) pow(4.0,gMER_SIZE);

#ifdef DEBUG
	printf("Alignment scores:\n");
	printf("\tA\tC\tG\tT\n");
	printf("A\t%f\t%f\t%f\t%f\n",gALIGN_SCORES[(int)'a'][0],gALIGN_SCORES[(int)'a'][1],
			gALIGN_SCORES[(int)'a'][2],gALIGN_SCORES[(int)'a'][3]);
	printf("C\t%f\t%f\t%f\t%f\n",gALIGN_SCORES[(int)'c'][0],gALIGN_SCORES[(int)'c'][1],
			gALIGN_SCORES[(int)'c'][2],gALIGN_SCORES[(int)'c'][3]);
	printf("G\t%f\t%f\t%f\t%f\n",gALIGN_SCORES[(int)'g'][0],gALIGN_SCORES[(int)'g'][1],
			gALIGN_SCORES[(int)'g'][2],gALIGN_SCORES[(int)'g'][3]);
	printf("T\t%f\t%f\t%f\t%f\n",gALIGN_SCORES[(int)'t'][0],gALIGN_SCORES[(int)'t'][1],
			gALIGN_SCORES[(int)'t'][2],gALIGN_SCORES[(int)'t'][3]);
	printf("N\t%f\t%f\t%f\t%f\n",gALIGN_SCORES[(int)'n'][0],gALIGN_SCORES[(int)'n'][1],
			gALIGN_SCORES[(int)'n'][2],gALIGN_SCORES[(int)'n'][3]);
#endif

	if(gVERBOSE) {
		cerr << params.str() << endl;
	}
	
	ofstream of(output_file);
	// Check to make sure the output file/directory exists
	if(of.bad() || of.fail()) {
		cerr << "ERROR:\n\t";
		perror("Error in reading output file");
		return -1;
	}
	
	of << params.str();

	// Create the output-writer thread
	//pthread_create(writer,NULL,write_thread,NULL);

	pthread_t pthread_hand[gNUM_THREADS];
	void *pthread_ret[gNUM_THREADS];
	pthread_mutex_init(&lock,NULL);
	pthread_mutex_init(&read_lock,NULL);
	pthread_mutex_init(&write_lock,NULL);
	
	if(gVERBOSE)
		cerr << "Hashing the genome." << endl;
	Genome gen; 
	try {
		gen.use(genome_file);
		gen.LoadGenome();
	}
	catch(const char* err) {
		cerr << "ERROR: \n\t" << err << endl;
		return -1;
	}
	catch(bad_alloc&) {
		perror("ERROR HERE");
		cerr << "ERROR: \n\tUnable to allocate enough memory.\n\t(Pick a smaller mer size?)" << endl;
		return -1;
	}
	catch(exception &e) {
		cerr << "ERROR: \n\t" << e.what() << endl;
		return -1;
	}
	catch(...) {
		cerr << "ERROR: \n\tUnknown error." << endl;
		return -1;
	}

	if(gVERBOSE > 1) 
		cerr << "\nTime to hash: " << (1.0*clock())/CLOCKS_PER_SEC << endl;			

	SeqReader sr(seq_file);
	unsigned int total_seqs = 0;

	if(gSAM)
		of << "# <QNAME>\t<FLAG>\t<RNAME>\t<POS>\t<MAPQ>\t<CIGAR>\t<MRNM>\t<MPOS>\t<ISIZE>\t<SEQ>\t<QUAL>\n";
	else 
		of << "# "
		   << "Consensus\tAlignScoreWithSelf\tAlignScoreCutoff\tNumUniqueMatches\t"
		   << "MAX:AlignScore\tMAX:NumIdenticalPositions\tMAX:ScoreAddedToPosition\tMAX:Sequence\t"
		   << "MAX:Locations...\n";

	
	//Here is where we'll begin the process of matching and writing to the 
	//output file.
	vector<string> file_vec = sr.GetSeqFileVector();
	
	vector<string>::iterator file_it = file_vec.begin();
	if(gVERBOSE) 
		cerr << "Matching " << sr.GetNumSeqFiles() << " file(s):" << endl;
		
	unsigned int seqs_matched = 0;
	unsigned int seqs_not_matched = 0;

	//we'll iterate through all of the files that were included on the commandline.
	for(; file_it!=file_vec.end(); ++file_it) {
		
		if(gVERBOSE)
			cerr << "Matching sequences of: " << *file_it << endl;
		try {
			//SequencReader will throw an error if the files are not in the proper
			//format.  This is to catch the error.
			sr.use(*file_it);		
		}
		catch(const char* err) {
			cerr << "ERROR: \n\t" << err << endl;
			if(gVERBOSE) 
				cerr << "\tIn file: " << *file_it << endl;
			continue;
		}
		catch(Exception *e) {
			cerr << "ERROR: \n\t" << e->GetMessage() << endl;
			cerr << "\tIn file " << *file_it << endl;
			delete e;
			continue;
		}
		catch(...) {
			cerr << "ERROR: \n\tImproper sequence length or file." << endl;
			cerr << "\tIn file " << *file_it << endl;
			continue;
		}		
		
		unsigned int num_seqs = sr.GetNumSeqs();
		total_seqs+=num_seqs;
		if(gVERBOSE)
			cerr << "Total Sequences: " << num_seqs << endl;

		unsigned int each_thread = num_seqs/gNUM_THREADS + 1;

		if(gVERBOSE > 1)
			cout << "Sequences for each thread: " << each_thread << endl;
		
		//Here is where we set the threading information.
		pthread_setconcurrency(gNUM_THREADS);
		pthread_mutex_init(&lock, NULL);
		pthread_attr_init(&attr);
		pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);

		//An array of pointers so we can delete them when we're done.
		//Read** seq_ptr[gNUM_THREADS];
		thread_opts* t_o_ptr[gNUM_THREADS];
		 
		//Give each thread an equal amount of sequences.
		for(int i=0; i<gNUM_THREADS; i++) {
						
			/*
			//We'll create a new vector pointer here.  Will delete it later.
			//vector<vector<double> > *seqs_to_match = new vector<vector<double> >[each_thread];
			Read** seqs_to_match = new Read*[each_thread];
			//Read** seqs_to_match = (Read**)malloc(sizeof(Read*) * each_thread);
			seq_ptr[i] = seqs_to_match;
			
			for(unsigned int j=0; j<each_thread; j++) {
				try {
					//SequencReader will throw an error if the files are not in the proper
					//format.  This is to catch the error.
					seqs_to_match[j] = sr.GetNextSequence();
				}
				catch(const char* err) {
					cerr << "ERROR: \n\t" << err << endl;
					if(gVERBOSE) 
						cerr << "\tIn file: " << *file_it << endl;
					continue;
				}
			}*/
					

			/*
			printf("**************Printing PWMS*********\n");
			for(unsigned int j=0; j<each_thread; j++) {
				if(!seqs_to_match[j])
					continue;
				printf("\n%d: ",j);
				printPWM(seqs_to_match[j]);
			}
			//*/

			//The parameters to pass to the thread
			//thread_opts* t_o = new thread_opts(gen,seqs_to_match,each_thread);
			thread_opts* t_o = new thread_opts(gen,sr,of,i);
			t_o->num_seqs = each_thread;
			t_o_ptr[i] = t_o;

			//pthread_create(&pthread_hand[i], &attr, thread_print, (void*)&t_o);
			pthread_create(&pthread_hand[i], NULL, thread_print, (void*)t_o);
		}

		
		//wait for all the threads to join back together...
		for (int i=0; i<gNUM_THREADS; i++) {
			pthread_join(pthread_hand[i], &pthread_ret[i]);
		}
		
		//delete the threads and the data so we can reserve memory.
		for(int i=0; i<gNUM_THREADS; i++) {
			//delete[] seq_ptr[i];
			delete t_o_ptr[i];
		}
		

		//print their return values to the file.
		for(int i=0; i<gNUM_THREADS; i++) {
			//of << (((thread_rets*)pthread_ret[i])->of)->str();
			seqs_matched += ((thread_rets*)pthread_ret[i])->good_seqs;
			seqs_not_matched += ((thread_rets*)pthread_ret[i])->bad_seqs;
			//delete ((thread_rets*)pthread_ret[i])->of;
			delete (thread_rets*)pthread_ret[i];
		}
	
		if(gVERBOSE) 
			cerr << endl;
	}
	
	
	if(gVERBOSE)
		cerr << "\nPrinting output." << endl;
		
	//Print to the .sgr file
	gen.PrintFinal(output_file);
	
	ostringstream stats;
	stats 	<< "\n#Finished." << endl 
			<< "#\tTotal Time: " << (1.0*clock())/CLOCKS_PER_SEC << " seconds." << endl
			<< "#\tFound " << total_seqs << " sequences." << endl
			//<< "\tSequences matched: " << total_seqs - bad_files << endl
			//<< "\tSequences not matched: " << bad_files << endl
			<< "#\tSequences matched: " << seqs_matched << endl
			<< "#\tSequences not matched: " << seqs_not_matched << endl
			<< "#\tOutput written to " << output_file << endl;
						
	cerr << stats.str();
	of << stats.str();
	of.close();
		
	return 0;
}


void* thread_print(void* t_opts) {
	//cout << "New Thread" << endl;
	unsigned int num_seqs = ((thread_opts*)t_opts)->num_seqs;
	unsigned int progress = num_seqs/100;
	unsigned int read_buffer = gREAD_BUFFER > num_seqs ? num_seqs : gREAD_BUFFER;
	//unsigned int num_seqs = gREAD_BUFFER;
	unsigned int good_seqs = 0;
	unsigned int bad_seqs = 0;

	unsigned int seqs_processed = 0;
	
	SeqReader *sr = &((thread_opts*)t_opts)->sr;

	bool finished = false;

	if(progress < 1)
		progress = 1;

	// Fill the read buffer
	while(!finished) {
		stringstream output;

		//printf("Getting more reads (total=%d and read buffer=%d) for thread %d!\n",num_seqs,read_buffer,((thread_opts*)t_opts)->thread_id);
		Read* reads[read_buffer];

		// Obtain a lock on the read generator
		pthread_mutex_lock(&read_lock);
		for(unsigned int k=0; k<read_buffer; k++) {
			try {
				// If it returns a NULL, it's because there are no more reads.
				// Just process the reads we have
				if(!(reads[k] = sr->GetNextSequence())) {
					finished = true;
					break;
				}
				//printf("Thread %d getting read %s\n",((thread_opts*)t_opts)->thread_id,reads[k]->name);
			}
			catch(const char* err) {
				cerr << "ERROR: \n\t" << err << endl;
				if(gVERBOSE) 
					cerr << "\tIn File: " << sr->GetFilename() << endl;
				finished = true;
				break;
			}
			catch(Exception* e) {
				cerr << "ERROR: \n\t" << e->GetMessage() << endl;
				if(gVERBOSE)
					cerr << "\tIn File: " << sr->GetFilename() << endl;
				finished = true;
				delete e;
				break;
			}
		}
		// Unlock the lock on the Sequence Reader
		pthread_mutex_unlock(&read_lock);

	
		for(unsigned int k=0; k < read_buffer; k++, seqs_processed++) {
			//vector<vector<double> > temp_vect = ((thread_opts*)t_opts)->seqs[k];
			Read* temp_read = reads[k];
			if(!temp_read) {
				//cout << "Found end" << endl;
				//break;
				//printf("Thread %d: Found null Read at pos %u (out of %u)\n",
				//	((thread_opts*)t_opts)->thread_id,k,read_buffer);
				break;
			}
			
			string consensus = GetConsensus(*temp_read);

			if(gVERBOSE)
				if( (seqs_processed % progress ) == 0) {
					cerr << "\r" << (int)(((double)seqs_processed/num_seqs)*100) << "% complete"; cerr.flush();
					//fprintf(stderr,"\r%d%% complete (seqs_processed=%d and num_seqs=%d)",
					//	(int)(((double)seqs_processed/num_seqs)*100),seqs_processed,num_seqs);
				}
			
			//if(temp_read->name)
				//output << temp_read->name << "\t";
			output << print_top_matches( ((thread_opts*)t_opts)->gen, *temp_read, consensus, bad_seqs, good_seqs);

			//done with the read.  Delete it.
			for(unsigned int i=0; i<temp_read->length; i++) {
				delete[] temp_read->pwm[i];
			}
			delete[] temp_read->pwm;
			if(temp_read->name)
				delete[] temp_read->name;

			delete temp_read;
			reads[k] = 0;
		}		 

		// Write this portion out to memory
		// Get a lock on the mutex to ensure correct writing
		pthread_mutex_lock(&write_lock);
		((thread_opts*)t_opts)->of << output.str();
		// Unlock the mutex
		pthread_mutex_unlock(&write_lock);
	}
		

	//delete[] ((thread_opts*)t_opts)->seqs;

	struct thread_rets* ret = new thread_rets;
	ret->good_seqs = good_seqs;
	ret->bad_seqs = bad_seqs;
	
	return (void*)ret;
}


int ParseCmdLine(const int argc, const char* argv[]) {
	
	bool found_other = false;

	for(int i=1; i<argc; i++) {
		if(*argv[i] == '-') { //check for normal
			int set_ret;

			if(*((argv[i])+1) == '-')  { //check for extended
				//printf("Found extended\n");
				set_ret = set_arg_ext(argv[i],argv[i+1],i);
			}
			else {
				set_ret = set_arg(argv[i],argv[i+1],i);
				i++;	//increment past the next expression
			}
			
			if(set_ret != 0) {
				this_cmd.errnum = set_ret;
				this_cmd.errpos = i;
				return -1;
			}

		}
		else {
			//if we've already found the the other cmd arg, it doesn't work. 
			if(found_other) {
				this_cmd.errnum = INVALID_ARG;
				this_cmd.errpos = i;
				return INVALID_ARG;
			}
			else {	//set the other arg to this pointer
				seq_file = argv[i];
			}
		}
	}

	return 0;

}

void GetParseError(ostream &os, const char* argv[]) {
	if(this_cmd.errnum == PARSE_ERROR) {
		os << "Irregular Parameter in: " << argv[this_cmd.errpos] << endl;
	}
	else if(this_cmd.errnum == NO_MATCHING_ARG) {
		os << "No matching arg in: " << argv[this_cmd.errpos] << endl;
	}
	else if(this_cmd.errnum == INVALID_ARG) {
		os << "Please specify only a single sequence file, or multiple files in \"quotations\" ("
		   << argv[this_cmd.errpos] << ")" << endl;
	}
	else {
		os << "Unknown error " << this_cmd.errnum << " in: " << argv[this_cmd.errpos] << endl;
	}

}


int set_arg(const char* param, const char* assign, int &count) {
	double temp_dbl;

	switch(*(param+1)) {	//switch on the first character
		case 'g':
			genome_file = assign;
			break;
		case 'l':
			if(sscanf(assign,"%d",&gSEQ_LENGTH) < 1)
				return PARSE_ERROR;
			break;
		case 'o':
			output_file = assign;
			break;
		case 'a':
			//if(sscanf(assign,"%lf",&ALIGN_SCORE) < 1)
			if(sscanf(assign,"%lf",&temp_dbl)< 1)
				return PARSE_ERROR;
			//	return (PARSE_ERROR * (int)(*param));
			gALIGN_SCORE = (float)temp_dbl;
			break;
		case 'p':
			count--;
			perc = true;
			break;
		case 'v':
			if(sscanf(assign,"%d",&gVERBOSE) < 1)
				return PARSE_ERROR;
			break;
		case 'c':
			if(sscanf(assign,"%d",&gNUM_THREADS) < 1)
				return PARSE_ERROR;
			break;
		case 'm': 
			if(sscanf(assign,"%d",&gMER_SIZE) < 1)
				return PARSE_ERROR;
			break;
		case 'B':
			if(sscanf(assign,"%d",&gBUFFER_SIZE) < 1)
				return PARSE_ERROR;
			break;
		case 'T':
			if(sscanf(assign,"%d",&gMAX_MATCHES) < 1)
				return PARSE_ERROR;
			break;
		case 'h':
			if(sscanf(assign,"%d",&gMAX_HASH_SIZE) < 1)
				return PARSE_ERROR;
			break;
		case 'S':
			pos_matrix = assign;
			break;
		case 'G':
			//if(sscanf(assign,"%lf",&gGAP) < 1)
			if(sscanf(assign,"%lf",&temp_dbl) < 1)
				return PARSE_ERROR;
			gGAP = (float)temp_dbl;
			break;
		case 'M':
			if(sscanf(assign,"%d",&gMAX_GAP) < 1)
				return PARSE_ERROR;
			break;
		case 'A':
			g_adaptor = new char[strlen(assign)+1];
			unsigned int i;
			for(i=0; i<strlen(assign); i++)
				g_adaptor[i] = tolower((char)assign[i]);
			// Null-terminate the string
			g_adaptor[i] = '\0';
			//g_adaptor = assign;
			break;
		case '0':
			count--;
			gPRINT_FULL = 1;
			break;
		case 'b':
			count--;
			gBISULFITE = 1;
			break;
		case 'd':
			count--;
			gATOG = 1;
			break;
		case 's':
			if(sscanf(assign,"%d",&gGEN_SKIP) < 1)
				return PARSE_ERROR;
			break;
		case 'j':
			if(sscanf(assign,"%d",&gJUMP_SIZE) < 1)
				return PARSE_ERROR;
			break;
		case '?':
			usage(0,"");
		default:
			return PARSE_ERROR;
	}

	return 0;

}

enum {
	PAR_ORIG=256,
	PAR_GENOME,
	PAR_LENGTH,
	PAR_OUTPUT,
	PAR_ALIGN_SCORE,
	PAR_PERCENT,
	PAR_VERBOSE,
	PAR_NUM_PROC,
	PAR_MER_SIZE,
	PAR_BUFFER,
	PAR_MAX_MATCH,
	PAR_MAX_HASH,
	PAR_SUBST_FILE,
	PAR_GAP_PENALTY,
	PAR_MAX_GAP,
	PAR_ADAPTOR,
	PAR_PRINT_FULL,
	PAR_BS_SEQ,
	PAR_A_TO_G,
	PAR_FAST,
	PAR_GEN_SKIP,
	PAR_READ,
	PAR_SAVE,
	PAR_GEN_SIZE,
	PAR_JUMP,
	PAR_SNP,
	PAR_ILL,
	PAR_SAM
};

int set_arg_ext(const char* param, const char* assign, int &count) {
	double temp_dbl;
	int which = 0;
	int adjust=0;

	if(strncmp(param+2,"genome=",7) == 0) {	//check to see if it's the genome
		//which = 'g';
		which = PAR_GENOME;
		adjust = 8;	//the size of "genome" plus 2
	}
	else if(strncmp(param+2,"length=",7) == 0) {
		//which = 'l';
		which = PAR_LENGTH;
		adjust = 8;
	}
	else if(strncmp(param+2,"output=",7) == 0) {
		//which = 'o';
		which = PAR_OUTPUT;
		adjust = 8;
	}
	else if(strncmp(param+2,"align_score=",12) == 0) {
		//which = 'a';
		which = PAR_ALIGN_SCORE;
		adjust = 13;
	}
	else if(strncmp(param+2,"percent",7) == 0) {
		//which = 'p';
		which = PAR_PERCENT;
		adjust = 9;
	}
	else if(strncmp(param+2,"verbose=",8) == 0) {
		//which = 'v';
		which = PAR_VERBOSE;
		adjust = 9;
	}
	else if(strncmp(param+2,"num_proc=",9) == 0) {
		//which = 'c';
		which = PAR_NUM_PROC;
		adjust = 10;
	}
	else if(strncmp(param+2,"mer_size=",9) == 0) {
		//which = 'm';
		which = PAR_MER_SIZE;
		adjust = 10;
	}
	else if(strncmp(param+2,"buffer=",7) == 0) {
		//which = 'B';
		which = PAR_BUFFER;
		adjust = 8;
	}
	else if(strncmp(param+2,"max_match=",9) == 0) {
		//which = 'T';
		which = PAR_MAX_MATCH;
		adjust = 10;
	}
	else if(strncmp(param+2,"max_hash=",10) == 0) {
		//which = 'h';
		which = PAR_MAX_HASH;
		adjust = 11;
	}
	else if(strncmp(param+2,"subst_file=",9) == 0) {
		//which = 'S';
		which = PAR_SUBST_FILE;
		adjust = 10;
	}
	else if(strncmp(param+2,"gap_penalty=",12) == 0) {
		//which = 'G';
		which = PAR_GAP_PENALTY;
		adjust = 13;
	}
	else if(strncmp(param+2,"max_gap=",8) == 0) {
		//which = 'M';
		which = PAR_MAX_GAP;
		adjust = 9;
	}
	else if(strncmp(param+2,"adaptor=",8) == 0) {
		//which = 'A';
		which = PAR_ADAPTOR;
		adjust = 9;
	}
	else if(strncmp(param+2,"print_full",10) == 0) {
		//which = '0';
		which = PAR_PRINT_FULL;
		adjust = 12;
	}
	else if(strncmp(param+2,"bs_seq",6) == 0) {
		//which = 'b';
		which = PAR_BS_SEQ;
		adjust = 8;
	}
	else if(strncmp(param+2,"a_to_g",6) == 0) {
		//which = 'd';
		which = PAR_A_TO_G;
		adjust = 8;
	}
	else if(strncmp(param+2,"fast",4) == 0) {
		//which = 'f';
		which = PAR_FAST;
		adjust = 6;
	}
	else if(strncmp(param+2,"gen_skip=",9) == 0) {
		//which = 's';
		which = PAR_GEN_SKIP;
		adjust = 10;
	}
	else if(strncmp(param+2,"read=",5) == 0) {
		//which = 1;
		which = PAR_READ;
		adjust = 6;
		gREAD=true;
	}
	else if(strncmp(param+2,"save=",5) == 0) {
		//which = 2;
		which = PAR_SAVE;
		adjust = 6;
		gSAVE=true;
	}
	else if(strncmp(param+2,"bin_size=",9) == 0) {
		//which = 3;
		which = PAR_GEN_SIZE;
		adjust = 10;
	}
	else if(strncmp(param+2,"jump",4) == 0) {
		which = PAR_JUMP;
		adjust = 6;
	}
	else if(strncmp(param+2,"snp",3) == 0) {
		which = PAR_SNP;
		adjust = 5;
	}
	else if(strncmp(param+2,"illumina",8) == 0) {
		which = PAR_ILL;
		adjust = 10;
	}
	else if(strncmp(param+2,"SAM",3) == 0) {
		which = PAR_SAM;
		adjust = 5;
	}
	else if(strcmp(param+2,"help") == 0) {
		usage(0,""); //usage will exit immediately
	}
	else 
		return NO_MATCHING_ARG;

	param += ++adjust;

	switch(which) {	//switch on the first character
		//case 'g':
		case PAR_GENOME:
			genome_file = param;
			break;
		//case 'l':
		case PAR_LENGTH:
			if(sscanf(param,"%d",&gSEQ_LENGTH) < 1)
				return PARSE_ERROR;
			break;
		//case 'o':
		case PAR_OUTPUT:
			output_file = param;
			break;
		//case 'a':
		case PAR_ALIGN_SCORE:
			//if(sscanf(param,"%lf",&ALIGN_SCORE) < 1)
			if(sscanf(param,"%lf",&temp_dbl) < 1)
				return PARSE_ERROR;
			//	return (PARSE_ERROR * (int)(*param));
			gALIGN_SCORE = (float)temp_dbl;
			break;
		//case 'p':
		case PAR_PERCENT:
			//count--;
			perc = true;
			break;
		//case 'v':
		case PAR_VERBOSE:
			if(sscanf(param,"%d",&gVERBOSE) < 1)
				return PARSE_ERROR;
			break;
		//case 'c':
		case PAR_NUM_PROC:
			if(sscanf(param,"%d",&gNUM_THREADS) < 1)
				return PARSE_ERROR;
			break;
		//case 'm': 
		case PAR_MER_SIZE:
			if(sscanf(param,"%d",&gMER_SIZE) < 1)
				return PARSE_ERROR;
			break;
		//case 'B':
		case PAR_BUFFER:
			if(sscanf(param,"%d",&gBUFFER_SIZE) < 1)
				return PARSE_ERROR;
			break;
		//case 'T':
		case PAR_MAX_MATCH:
			if(sscanf(param,"%d",&gMAX_MATCHES) < 1)
				return PARSE_ERROR;
			break;
		//case 'h':
		case PAR_MAX_HASH:
			if(sscanf(param,"%d",&gMAX_HASH_SIZE) < 1)
				return PARSE_ERROR;
			break;
		//case 'S':
		case PAR_SUBST_FILE:
			pos_matrix = param;
			break;
		//case 'G':
		case PAR_GAP_PENALTY:
			//if(sscanf(param,"%lf",&gGAP) < 1)
			if(sscanf(param,"%lf",&temp_dbl) < 1)
				return PARSE_ERROR;
			gGAP = (float)temp_dbl;
			break;
		//case 'M':
		case PAR_MAX_GAP:
			if(sscanf(param,"%d",&gMAX_GAP) < 1)
				return PARSE_ERROR;
			break;
		//case 'A':
		case PAR_ADAPTOR:
			g_adaptor = new char[strlen(assign)];
			for(unsigned int i=0; i<strlen(assign); i++)
				g_adaptor[i] = tolower((char)param[i]);
			//g_adaptor = (char*)param;
			break;
		//case '0':
		case PAR_PRINT_FULL:
			//count--;	//don't decrement in the extended version
			gPRINT_FULL = 1;
			break;
		//case 'b':
		case PAR_BS_SEQ:
			//count--;
			gBISULFITE = 1;
			break;
		//case 'd':
		case PAR_A_TO_G:
			//count--;
			gATOG = 1;
			break;
		//case 'f':
		case PAR_FAST:
			//count--;
			fast = true;
			break;
		//case 's':
		case PAR_GEN_SKIP:
			if(sscanf(param,"%d",&gGEN_SKIP) < 1)
				return PARSE_ERROR;
			break;
		//case 1:	//read
		case PAR_READ:
			gREAD_FN = (char*)param;
			if(!gREAD_FN)
				return PARSE_ERROR;
			break;
		//case 2:	//save
		case PAR_SAVE:
			gSAVE_FN = (char*)param;
			if(!gSAVE_FN)
				return PARSE_ERROR;
			break;
		case PAR_GEN_SIZE:
			if(sscanf(param,"%ud",&gGEN_SIZE) < 1)
				return PARSE_ERROR;
			break;
		case PAR_JUMP: 
			if(sscanf(param,"%ud",&gJUMP_SIZE) < 1)
				return PARSE_ERROR;
			break;
		case PAR_SNP:
			gSNP = true;
			gGEN_SIZE = 1;
			gPRINT_FULL = 1;
			break;
		case PAR_SAM:
			gSAM = 1;
			break;
		case PAR_ILL:
			gILLUMINA = true;
			break;
		default:
			return NO_MATCHING_ARG;
	}

	return 0;

}
