#include "bin_seq.h"
	
#define NEG_INF -1000;
inline float bin_seq::p(char x, char y) {
	if(x == y) return 10*4/5; else return 10*1/5;
}
	
inline float bin_seq::p_seq(float x[4], char y) {
	float sum = 0;
	sum += x[A_POS] * p('a',tolower(y));
	sum += x[C_POS] * p('c',tolower(y));
	sum += x[G_POS] * p('g',tolower(y));
	sum += x[T_POS] * p('t',tolower(y));
	return sum;
}
							
							
float** bin_seq::pairHMM(const Read &r, const string &consensus, const string &genome) {
	int i,j;
	int n = r.length;
	int m = genome.size();

	// Forward
	// Create Matrices
	double fMXY[n+1][(m+1)*3];
	double fM[n+1][m+1];
	double fX[n+1][m+1];
	double fY[n+1][m+1];
	// backward matrices
	double bM[n+1][m+1];
	double bX[n+1][m+1];
	double bY[n+1][m+1];
	
	// final matrices
	double pM[n][m];
	float** pGenScore = new float*[n];
	double pX[n][m];
	double pY[n][m];

	// set them to zero
	for(i=0; i<n+1; i++) {
		for(j=0; j<m+1; j++) {
			fMXY[i][3*j] = 0;
			fMXY[i][3*j+1] = 0;
			fMXY[i][3*j+2] = 0;
			fM[i][j] = 0;
			fX[i][j] = 0;
			fY[i][j] = 0;

			bM[i][j] = 0;
			bX[i][j] = 0;
			bY[i][j] = 0;

		}
	}
	for(i=0; i<n; i++) {
		pGenScore[i] = new float[5];
		pGenScore[i][A_POS] = 0;
		pGenScore[i][C_POS] = 0;
		pGenScore[i][G_POS] = 0;
		pGenScore[i][T_POS] = 0;
		pGenScore[i][N_POS] = 0;
		
		for(j=0; j<m; j++) {
			pM[i][j] = 0;
			pX[i][j] = 0;
			pY[i][j] = 0;
		}
	}

	// NOTE:  Indices are different because R is 1-based
	fM[0][0] = 1;

	// Forward

	for(i=1; i<n+1; i++) {
		for(j=1; j<m+1; j++) {
			fM[i][j] = p_seq(r.pwm[i-1],genome[j-1]) * (PHMM_Tmm*fM[i-1][j-1] + PHMM_Tgm*fX[i-1][j-1] + PHMM_Tgm*fY[i-1][j-1]);

			fX[i][j] = PHMM_q*(PHMM_Tmg*fM[i-1][j]+PHMM_Tgg*fX[i-1][j]);
			fY[i][j] = PHMM_q*(PHMM_Tmg*fM[i][j-1]+PHMM_Tgg*fY[i][j-1]);
		}
	}

	double fE=PHMM_t*(fM[n][m]+fX[n][m]+fY[n][m]);


	// Backward
	bM[n-1][m-1] = bX[n-1][m-1] = bY[n-1][m-1] = PHMM_t;

	for(i=n-1; i>=0; i--) {
		for(j=m-1; j>=0; j--) {
			if(j == (m-1) && i == (n-1))
				continue;
			if(j == m-1) {
				bM[i][j] = PHMM_q*PHMM_Tmg*bX[i+1][j];
				bX[i][j] = PHMM_q*PHMM_Tgg*bX[i+1][j];
				bY[i][j] = 0;
				continue;
			}
			if(i == n-1) {
				bM[i][j] = PHMM_q*PHMM_Tmg*bY[i][j+1];
				bY[i][j] = PHMM_q*PHMM_Tgg*bY[i][j+1];
				bX[i][j] = 0;
				continue;
			}
			
			bM[i][j]= p_seq(r.pwm[i+1],genome[j+1])*PHMM_Tmm*bM[i+1][j+1] + PHMM_q*PHMM_Tmg*bX[i+1][j] + PHMM_q*PHMM_Tmg*bY[i][j+1];
			bX[i][j]= p_seq(r.pwm[i+1],genome[j+1])*PHMM_Tgm*bM[i+1][j+1] + PHMM_q*PHMM_Tgg*bX[i+1][j];
			bY[i][j]= p_seq(r.pwm[i+1],genome[j+1])*PHMM_Tgm*bM[i+1][j+1] + PHMM_q*PHMM_Tgg*bY[i][j+1];

		}
	}

	for(i=1; i<n+1; i++) {
		for(j=1; j<m+1; j++) {
			pM[i-1][j-1] = fM[i][j]*bM[i-1][j-1]/fE;
			pX[i-1][j-1] = fX[i][j]*bX[i-1][j-1]/fE;
			pY[i-1][j-1] = fY[i][j]*bY[i-1][j-1]/fE;
		}
	}


	//apply(pM+pX,1,sum)
	//apply(pM+pY,2,sum)
	/*
	for(i=0; i<n; i++) {
		for(j=0; j<m; j++) {
			pX[i][j] = pM[i][j] + pX[i][j];
		}
	}
	*/

	for(i=0; i<m; i++) {
		for(j=0; j<n; j++) {
			pGenScore[i][g_gen_CONVERSION[(unsigned)consensus[j]]] += pY[j][i] + pM[j][i];
		}
	}


	/*
	for(i=0; i<n; i++) {
		for(j=0; j<4; j++) {
			printf("%.2f\t",pGenScore[i][j]);
		}
		printf("\n");
	}
	printf("\n");
	//*/
	
	return pGenScore;
}




/*
 * The following functions are used for hashing the genome.  Thus the name "binary sequence"
 */
unsigned int bin_seq::str2int(string &a) {
	unsigned int temp = 0;
	for(unsigned int i=0; i<16 && i<a.size(); i++) {
		temp <<= 2;
		//temp += base2int(a[i]);
		temp |= g_bs_CONVERSION[(int)a[i]];
	}
	
	return temp;
}

string bin_seq::hash2str(const unsigned int h, int length) {
	char to_return[length];
	unsigned int temp = h;

	for(int i=1; i<=length; i++) {
		//to_return[length-i] = int2base((temp % 4));
		to_return[length-i] = gINT2BASE[(temp % 4)];
		temp >>= 2;
	}
	
	to_return[length] = '\0';
	
	return to_return;
}

pair<bool,unsigned long> bin_seq::get_hash(const char* str) {
	string str_str = str;
	
	return get_hash(str_str);
}

pair<bool,unsigned long> bin_seq::get_hash(string str) {
	unsigned long temp = 0;

	//ensure that the sequence doesn't overflow the possible size of an unsinged int,
	//as well as not overflowing the gMER_SIZE.
	for(unsigned int i=0; i<(sizeof(unsigned long) * 4) && i<str.length(); i++) {
		
		//int spot = base2int(str[i]); 
		int spot = g_bs_CONVERSION[(int)str[i]]; 
		if(spot > 3)	//if it's not a valid, hashable character
			return pair<bool,unsigned long>(false,0);
			
		temp <<= 2;
		temp |= spot;
	}
		
	//temp %= HASH_SIZE;
	return pair<bool,unsigned long>(true,temp);
}

pair<int,unsigned long> bin_seq::get_hash(unsigned int &pos, const unsigned char* genome) {
	unsigned int temp = 0;

	int found = 0;
	// Ensure that the sequence doesn't overflow the possible size of a unsigned int,
	// as well as not overflowing the MER_SIZE.  (We won't allow MER_SIZE to
	// overflow the size of an unsigned long--64 bits.  This means we can allow up to 32bp MER_SIZE)
	// We want to make sure we find exactly MER_SIZE bases (found), but the int i helps us determine
	// where on the genome we look for them--while ignoring newline characters.
	//for(int i=0,found=0; found<gMER_SIZE && found<64 && pos+i<BUFFER_SIZE; found++, i++) {
	for(int i=0; found<gMER_SIZE; i++) {
		//unsigned int spot = base2int(genome[pos+i]);
		unsigned int spot = g_bs_CONVERSION[(int)genome[pos+i]];
		
		if(spot < 4) {
			temp <<= 2;
			temp |= spot;
		
			found++;
		}
		else if(spot == 5) {	//if there's an newline reached, this isn't an error: just ignore.
			if(i == 0) {	//if there's a newline reached on the first of a sequence...
				pos++;
				i=-1;
			}
		}
		else if(spot == 6) {
			//pos = BUFFER_SIZE;	// 5 comes from the flag for the end of the stream.
			return pair<int,unsigned int>(-1,0);
		}
		else if(spot == 7) {
			pos = pos+i;
			//return a flag for a new genome, with a number representing the number of bases skipped.
			return pair<int,unsigned int>(-2,found); 
		}
		else { // if(spot == 4) 
			//cout << "Unknown character: <" << genome[pos+i] << ">" << endl;
			return pair<int,unsigned int>(0,0);
		}
	}
		
	return pair<bool,unsigned int>(1,temp);
}
pair<int,unsigned long> bin_seq::get_hash(unsigned int &pos, const unsigned char* genome, const unsigned int max_pos) {
	unsigned int temp = 0, spot = 0;

	while(genome[pos] == '\n')
		pos++;

	int found = 0;
	// Ensure that the sequence doesn't overflow the possible size of a unsigned int,
	// as well as not overflowing the MER_SIZE.  (We won't allow MER_SIZE to
	// overflow the size of an unsigned long--64 bits.  This means we can allow up to 32bp MER_SIZE)
	// We want to make sure we find exactly MER_SIZE bases (found), but the int i helps us determine
	// where on the genome we look for them--while ignoring newline characters.
	//for(int i=0,found=0; found<gMER_SIZE && found<64 && pos+i<BUFFER_SIZE; found++, i++) {
	int i;
	for(i=0; found<gMER_SIZE && pos+i<max_pos; i++) {
		//unsigned int spot = base2int(genome[pos+i]);
		spot = g_bs_CONVERSION[(unsigned int)genome[pos+i]];
		
		if(spot < 4) {
			temp <<= 2;
			temp |= spot;
		
			found++;
		}
		else if(spot == 4) {
			//cout << "Unknown character: <" << genome[pos+i] << ">" << endl;
			return pair<int,unsigned int>(0,0);
		}
		else if(spot == 5) {	//if there's an newline reached, this isn't an error: just ignore.
			if(i == 0) {	//if there's a newline reached on the first of a sequence...
				pos++;
				i=-1;
				//fprintf(stderr,"Made it here");
			}
			//fprintf(stderr, ": did I?\n");
		}
		else if(spot == 6) {
			//pos = BUFFER_SIZE;	// 5 comes from the flag for the end of the stream.
			return pair<int,unsigned int>(-1,0);
		}
		else if(spot == 7) {
			pos = pos+i;
			//return a flag for a new genome, with a number representing the number of bases skipped.
			return pair<int,unsigned int>(-2,found); 
		}
	}
	
	// This is because it couldn't read enough bases
	if(pos+i >= max_pos)
		return pair<bool,unsigned int>(0,0);
	

	if(spot == 7) {
		//fprintf(stderr,"The Size is larger AGAIN\n");
		return pair<int,unsigned int>(-2,found);
	}

	return pair<bool,unsigned int>(1,temp);
}

inline unsigned int bin_seq::base2int(char c) {
	switch(c) {
		case 'A' : return 0;
		case 'C' : return 1;
		case 'G' : return 2;
		case 'T' : return 3;
		case 'a' : return 0;
		case 'c' : return 1;
		case 'g' : return 2;
		case 't' : return 3; 
		case 'N' : return 4;
		case 'n' : return 4;
		case '\n' : return 5;
		case '\r' : return 5;
		case '\0' : return 6;	//5 is the flag for the end of the genome stream.
		case '>' : return 7;	//7 means there's a new genome.
	}
	return 6;
} 
inline char bin_seq::int2base(unsigned int i) {
	switch(i) {
		case 0 : return 'a';
		case 1 : return 'c';
		case 2 : return 'g';
		case 3 : return 't';
	}
	return 'N';
} 



/*
 * The next few functions are for producing a Needleman-Wunsch alignment
 * for two given sequences.
 */
void bin_seq::print_align(string str1, string align, string str2) {
	int s1_size = str1.size();
	for(int i=1; i<=s1_size; i++) 
		cout << str1[s1_size-i]; 
	cout << endl;
	
	int a_size = align.size(); 
	for(int i=1; i<=a_size; i++)
		cout << align[a_size-i];
	cout << endl;
	
	int s2_size = str2.size();
	for(int i=1; i<=s2_size; i++)
		cout << str2[s2_size-i];
	cout << endl;
}

//will return the alignment string
pair<string,string> bin_seq::get_align_score_w_traceback(const Read &read, const string &consense, const string &gen) {
	/* 
	cout << "\naligning " << gen << endl
	     <<   "with     " << consense << endl;
	cout << "Scores: " << get_val(read.pwm[0],'a') << ' '
		 			   << get_val(read.pwm[0],'c') << ' '
		 			   << get_val(read.pwm[0],'g') << ' '		 			   
		 			   << get_val(read.pwm[0],'t') << endl; 
	//*/

	int SEQ_LENGTH = read.length;

	unsigned int seq1Length = SEQ_LENGTH;
	unsigned int seq2Length = gen.size();
	

	// Check to see if we need to extend the default array
	unsigned int squareLen = (seq1Length+1) * (seq2Length+1);
	if(squareLen > def_arr_size) {
		delete[] def_arr;
		def_arr = new float[squareLen];
		def_arr_size = squareLen;
	}
	if(squareLen > def_moves_size) {
		if(def_moves_arr)
			delete[] def_moves_arr;
		def_moves_arr = new char[squareLen];
		def_moves_size = squareLen;
	}
	
	float* nm = def_arr;
	char* moves = def_moves_arr;
	
	// The size of a single row
	unsigned int row_size = seq1Length+1;

	
	//Initialize the entire array to zeros.
	for(int i=0; i<=SEQ_LENGTH; i++)
		//for(int j=0; j<=gen.size(); j++) {
		for(int j=i-gMAX_GAP-1; j<=i+gMAX_GAP+1; j++) {
			if(j < 0 || j > (int)gen.size() )
				continue;
			nm[i*row_size + j] = NEG_INF;
			moves[i*row_size + j] = 'D';
		}
	
	//Initialize the first row to have the gap penalty.
	for(int i=0; i<=gMAX_GAP+2; i++) {
		nm[i*row_size] = gGAP * i;
		moves[i*row_size] = 'U';
	}
	//Initialize the first column to have the gap penalty.
	for(int j=0; j<=gMAX_GAP+2; j++) {
		nm[j] = gGAP * j;
		moves[j] = 'L';
	}
	moves[0] = 'D';

	//Set up the NW matrix	
	for(int i=1; i<=SEQ_LENGTH; i++) {
		//for(unsigned int j=1; j<=gen.size(); j++) {
		for(int j=i-gMAX_GAP; j<=i+gMAX_GAP; j++) {
			if(j <= 0)
				continue;
			if (j>(int)gen.size())
				break;
			
			float m_mm1 = nm[(i-1)*row_size + (j-1)];
			float m_mm2 = get_val(read.pwm[i-1],gen[j-1]);
			float m_mm = m_mm1 + m_mm2;
			float gap1 = nm[(i-1)*row_size + j] + gGAP;
			float gap2 = nm[i*row_size + (j-1)] + gGAP;

			
			
			float max_score = max_flt(moves[i*row_size + j],m_mm,gap1,gap2);
			
			nm[i*row_size + j] = max_score;

		}	
	}
	
#ifdef DEBUG
	// Print the NW matrix
 	for(int i=0; i<=SEQ_LENGTH; i++) {
		for(int j=0; j<=gen.size(); j++) {
			cout << nm[i*row_size + j] << '\t';
			
		}
		cout << endl;
	}
	cout << endl;
	
	// Print the "moves" matrix
	for(int i=0; i<=SEQ_LENGTH; i++) {
		for(int j=0; j<=gen.size(); j++) {
			cout << moves[i*row_size + j] << '\t';
			
		}
		cout << endl;
	}
	cout << endl;
#endif
	
	//Trace back though the NW matrix to find the best alignment.
	string aligned = "";
	string aligned_gen = "";
	char CIGAR[1024] = "";
	
	int gaps=0;

	int i = SEQ_LENGTH;
	int j = gen.size();

	CIGAR_TYPE c_type = MATCH;
	int c_counter = 0;

		
	while( (i != 0) && (j != 0) ) {
		char temp[1024];
		switch(moves[i*row_size + j]) {
			case 'D': // Made a diagonal transition
				aligned += consense[i-1];
				aligned_gen += gen[j-1];

				// If it's a mismatch or a match
				//if(consense[i-1] != gen[j-1]) {
					if(c_type == MATCH)
						c_counter++;
					else {
						// prepend the CIGAR format
						strcpy(temp,CIGAR);
						//sprintf(CIGAR,"%c%d%s",'M',j,temp);
						sprintf(CIGAR,"%d%c%s",c_counter,CIGAR_TO_STR(c_type),temp);
						c_type = MATCH;
						c_counter = 1;
					}
				//}

				i -= 1;
				j -= 1;

				break;
			case 'U': // Made a transition from the upper sequence
				aligned_gen += "-";
				aligned += consense[i];

				// prepend the CIGAR format
				if(c_type == INS)
					c_counter++;
				else {
					strcpy(temp,CIGAR);
					sprintf(CIGAR,"%d%c%s",c_counter,CIGAR_TO_STR(c_type),temp);
					c_type = INS;
					c_counter = 1;
				}
				
				i -= 1;
				gaps++;

				break;
			case 'L': // Made a transition from the left
				aligned_gen += gen[j];
				aligned += '-';

				if(c_type == DEL)
					c_counter++;
				else {
					// prepend the CIGAR format
					strcpy(temp,CIGAR);
					sprintf(CIGAR,"%d%c%s",c_counter,CIGAR_TO_STR(c_type),temp);

					c_type = DEL;
					c_counter = 1;
				}

				j -= 1;
				gaps++;
				
				break;
			default: 
				cerr << "ERROR at " << i << "," << j << "." << endl;
				cerr << "\tValue is: " << moves[i*row_size+j] << "(" << (int)moves[i*row_size+j] << ")" << endl;
				return pair<string,string>("","");
		}
	}

	char temp[1024];
	// Clean up the CIGAR from the front of the sequence
	strcpy(temp,CIGAR);
	sprintf(CIGAR,"%d%c%s",c_counter,CIGAR_TO_STR(c_type),temp);

	// Reverse the sequences
	string temp_aligned = aligned; 
	for(unsigned int i=0; i<aligned.size(); i++) {
		aligned[i] = temp_aligned[aligned.size()-1-i];
	}
	string temp_gen = aligned_gen;
	for(unsigned int i=0; i<aligned_gen.size(); i++) {
		aligned_gen[i] = temp_gen[aligned_gen.size()-1-i];
	}
	
	//return pair<string,string>(aligned,aligned_gen);
	return pair<string,string>(aligned,CIGAR);
}

/*
 * Puts all the begin, mid and end functions together and returns the value incorporated here.
 */
float bin_seq::get_align_score(const Read &read, const string &gen,
		unsigned int begin, unsigned int end) {
		
	float value = 0;


	// Check to see if we need a larger pointer length
	unsigned int squareLen = (read.length+1) * (gen.size()+1);
	if(squareLen > def_arr_size) {
		delete[] def_arr;
		def_arr = new float[squareLen];
		def_arr_size = squareLen;
	}
	
	value += get_align_score_begin(read,gen,begin);
	value += get_align_score_mid(read,gen,begin,end);
	value += get_align_score_end(read,gen,end);
	
	return value;
		
}

/*
 * This is meant to align the first piece of the read.  Aligning from the (exact) matching
 * portion which was produced from the hash piece to the beginning of the read.
 *
 *  *
 *	  *		<- we want to match this first piece.
 *		*
 *        *
 *          X	<- end.  We know it must be an exact match
 *			  X
 *
 */
float bin_seq::get_align_score_begin(const Read &read, const string &gen, unsigned int end) {
	if(end == 0)	//trivial case
		return 0;
		
	//float nm[end+1][end+1];
	unsigned int size = end+1;
	//float* nm = new float[size*size];
	// Use the default array
	float* nm = def_arr;
	
	//Initialize the middle of the array to zeros.
	for(unsigned int i=0; i<=end; i++) {
		for(int j=(signed int)i-gMAX_GAP-1; j<=(signed int)i+gMAX_GAP+1 && j<=(signed int)end; j++) {
			if(j < 0)
				continue;
			//nm[i][j] = NEG_INF;
			nm[i*size+j] = NEG_INF;
		}
	}
	
	//Initialize the last row to have the gap penalty.
	//for(int i=gMAX_GAP+2; i<=(signed int)end; i++) {
	for(int i=0; i<=(signed int)end; i++) {
		//nm[i][end] = gGAP * (end-i);
		nm[i*size+end] = gGAP * (end-i);
	}
	//Initialize the last column to have the gap penalty.
	//for(int j=gMAX_GAP+2; j<=(signed int)end; j++) {
	for(int j=0; j<=(signed int)end; j++) {
		//nm[end][j] = gGAP * (end-j);
		nm[end*size+j] = gGAP * (end-j);
	}
	
	
	//Set up the NW matrix	
	for(int i=end-1; i>=0; i--) {	//needs to be signed when it goes to 0...
		for(int j=i+gMAX_GAP; j>=i-gMAX_GAP; j--) {	//j needs to be signed
			if (j>=(signed int)end) 
				continue;
			if (j<0)
				break;


#ifdef DEBUG
/*	
	for(int q=0; q<=end; q++) {
		for(int p=0; p<=end; p++) {
			cout << nm[q][p] << '\t';
			
		}
		cout << endl;
	}
	cout << endl;
*/
#endif

			//float m_mm1 = nm[i+1][j+1];
			float m_mm1 = nm[(i+1)*size+(j+1)];
			float m_mm2 = get_val(read.pwm[i],gen[j]);
			float m_mm = m_mm1 + m_mm2;
			//float gap2 = nm[i+1][j] + gGAP;
			float gap2 = nm[(i+1)*size+j] + gGAP;
			//float gap1 = nm[i][j+1] + gGAP;
			float gap1 = nm[i*size+(j+1)] + gGAP;
			//double gap1 = nm[i][j-1] + get_val(pwm[i],gen[j]);
			//double gap2 = nm[i-1][j] + get_val(pwm[i-1],gen[j-1]);
			
			float max_score = max_flt(m_mm,gap1,gap2);
			
			//nm[i][j] = max_score;
			nm[i*size+j] = max_score;
		}	
	}

#ifdef DEBUG
/*
	cout << "\n\nend: " << end << endl;
	for(int i=0; i<=end; i++) {
		for(int j=0; j<=end; j++) {
			cout << nm[i][j] << '\t';
			
		}
		cout << endl;
	}
	cout << endl;
*/
#endif	

	//float align_score = nm[0][0];
	float align_score = nm[0];
	// Don't delete.  Keep in memory
	//delete[] nm;
	return align_score;
}

/*
 * We know there's an exact match, so we just compute the alignment score with itself
 * from start to end.
 * start is the first position of the exact match.  end is the last position of the 
 * exact match (not the first position of a possible match).
 */
float bin_seq::get_align_score_mid(const Read &read, const string &consense, 
			unsigned int start, unsigned int end) {
	float score = 0;
	for(unsigned int i=start; i<=end; i++) {
		score += (read.pwm[i][0] * gALIGN_SCORES[(unsigned int)consense[i]][0])
		   + (read.pwm[i][1] * gALIGN_SCORES[(unsigned int)consense[i]][1])
		   + (read.pwm[i][2] * gALIGN_SCORES[(unsigned int)consense[i]][2])
		   + (read.pwm[i][3] * gALIGN_SCORES[(unsigned int)consense[i]][3]);
	}
	
	return score;
}

/*
 * This is meant to align the last piece of the read.  Aligning from the (exact) matching
 * portion which was produced from the hash piece to the end of the read.
 *
 *  X
 *	  X		<- start.  We know it must be an exact match
 *		*
 *        *
 *          *	<- we want to match this last piece.
 *			  *
 *
 */
float bin_seq::get_align_score_end(const Read &read, const string &gen, unsigned int start) {
	if(start == read.length-1)	//trivial case
		return 0;
		
	unsigned int length = read.length - start;	//start+length == read.length
	
	//float nm[length][length];
	//float* nm = new float[length*length];
	// Use the default array
	float* nm = def_arr;
	
	//Initialize the middle of the array to zeros.
	for(unsigned int i=0; i<length; i++) {
		for(int j=i-gMAX_GAP-1; j<=(signed int)i+gMAX_GAP+1 && j < (signed int)length; j++) {	//j needs to be signed
			if(j < 0)
				continue;
			//nm[i][j] = NEG_INF;
			nm[i*length+j] = NEG_INF;
		}
	}
	
	//Initialize the first row to have the gap penalty.
	for(unsigned int i=0; i<=(unsigned int)gMAX_GAP+1 && i<length; i++) {
		//nm[i][0] = gGAP * i;
		nm[i*length+0] = gGAP * i;
	}
	//Initialize the first column to have the gap penalty.
	for(unsigned int j=0; j<=(unsigned int)gMAX_GAP+1 && j<length; j++) {
		//nm[0][j] = gGAP * j;
		nm[0*length+j] = gGAP * j;
	}
	//Set up the NW matrix	
	for(unsigned int i=1; i<length; i++) {
		for(int j=i-gMAX_GAP; j<=(signed int)i+gMAX_GAP && j<(signed int)length; j++) {
			if(j <= 0)
				continue;
			if (j>=(signed int)length)
				break;
			
			//float m_mm1 = nm[i-1][j-1];
			float m_mm1 = nm[(i-1)*length+(j-1)];
			float m_mm2 = get_val(read.pwm[i+start],gen[j+start]);
			float m_mm = m_mm1 + m_mm2;
			//float gap1 = nm[i][j-1] + gGAP;
			float gap1 = nm[i*length+(j-1)] + gGAP;
			//float gap2 = nm[i-1][j] + gGAP;
			float gap2 = nm[(i-1)*length+j] + gGAP;
			//double gap1 = nm[i][j-1] + get_val(pwm[i],gen[j]);
			//double gap2 = nm[i-1][j] + get_val(pwm[i-1],gen[j-1]);
			
			float max_score = max_flt(m_mm,gap1,gap2);
			
			//nm[i][j] = max_score;
			nm[i*length+j] = max_score;

		}	
	}
	
	//float align_score = nm[length-1][length-1];
	float align_score = nm[(length-1)*length+(length-1)];

	// Keep the default array in memory
	//delete[] nm;
	return align_score;
}




float bin_seq::get_align_score(Read &read, string &gen) {
//double bin_seq::get_align_score(vector< vector<double> > &pwm, string &gen) {
	/* 
	cout << "aligning " << gen << " with " << pwm[0][0] << "\t" << pwm[0][1] << endl;
	cout << "Scores: " << get_val(pwm[0],'a') << ' '
		 			   << get_val(pwm[0],'c') << ' '
		 			   << get_val(pwm[0],'g') << ' '		 			   
		 			   << get_val(pwm[0],'t') << endl; 
	//*/

	//int SEQ_LENGTH = pwm.size();
	int SEQ_LENGTH = read.length;

	float nm[SEQ_LENGTH+1][gen.size()+1];
	
	//Initialize the middle of the array to zeros.
	for(int i=0; i<=SEQ_LENGTH; i++) {
		for(int j=i-gMAX_GAP-1; j<=i+gMAX_GAP+1; j++) {
			if(j < 0 || j > (int)gen.size() )
				continue;
			nm[i][j] = NEG_INF;
		}
	}
	
	//Initialize the first row to have the gap penalty.
	for(int i=0; i<=gMAX_GAP+2; i++) {
		nm[i][0] = gGAP * i;
	}
	//Initialize the first column to have the gap penalty.
	for(int j=0; j<=gMAX_GAP+2; j++) {
		nm[0][j] = gGAP * j;
	}

	//Set up the NW matrix	
	for(int i=1; i<=SEQ_LENGTH; i++) {
		for(int j=i-gMAX_GAP; j<=i+gMAX_GAP; j++) {
			if(j <= 0)
				continue;
			if (j>(int)gen.size())
				break;
			
			float m_mm1 = nm[i-1][j-1];
			float m_mm2 = get_val(read.pwm[i-1],gen[j-1]);
			float m_mm = m_mm1 + m_mm2;
			float gap1 = nm[i][j-1] + gGAP;
			float gap2 = nm[i-1][j] + gGAP;
			//double gap1 = nm[i][j-1] + get_val(pwm[i],gen[j]);
			//double gap2 = nm[i-1][j] + get_val(pwm[i-1],gen[j-1]);
			
			float max_score = max_flt(m_mm,gap1,gap2);
			
			nm[i][j] = max_score;

		}	
	}
	
#ifdef DEBUG
	//Uncomment this to print the NW matrix
/* 	for(int i=0; i<=SEQ_LENGTH; i++) {
		for(int j=0; j<=gen.size(); j++) {
			cout << nm[i][j] << '\t';
			
		}
		cout << endl;
	}
	cout << endl;
*/	
	//if(VERBOSE > 3) {
	//	cout << "final answer: " << traceback << endl;
	//}
#endif

	float align_score = nm[SEQ_LENGTH][gen.size()];
	
	return align_score;
}


//inline double bin_seq::get_val(const vector<double> &a, const char b) {
inline float bin_seq::get_val(const float* a, const char b) {
	//if(VERBOSE > 4)
		//cout << "comparing " << b << endl;
	float score = 0;
	//score += (a[0] * gALIGN_SCORES[(unsigned int)b][0])
	//	   + (a[1] * gALIGN_SCORES[(unsigned int)b][1])
	//	   + (a[2] * gALIGN_SCORES[(unsigned int)b][2])
	//	   + (a[3] * gALIGN_SCORES[(unsigned int)b][3]);
	//
	//return score;

	float* scoreptr = gALIGN_SCORES[(unsigned int)b];
	score = (a[0] * scoreptr[0]
		   + a[1] * scoreptr[1]
		   + a[2] * scoreptr[2]
		   + a[3] * scoreptr[3]);
	return score;
/*
	int gen_char;
	switch(b) {
		case 'a': 
			gen_char = 0;
			break;
		case 'c':
			gen_char = 1;
			break;
		case 'g':
			gen_char = 2;
			break;
		case 't':
			gen_char = 3;
			break;
		case 'n':
			gen_char = 4;
			break;
		default:
			printf("Using default here.  Try another method.\n");
			gen_char = 4;
			break;
	}

	float score = 0;
	score += (a[0] * gALIGN_SCORES[gen_char][0])
		   + (a[1] * gALIGN_SCORES[gen_char][1])
		   + (a[2] * gALIGN_SCORES[gen_char][2])
		   + (a[3] * gALIGN_SCORES[gen_char][3]);

	return score;
*/
} 

inline float bin_seq::get_val(char a, char b) {
#ifdef DEBUG
	if(gVERBOSE > 4)
		cout << "comparing " << a << " and " << b << endl;
#endif

	return gALIGN_SCORES[(unsigned int)a][(unsigned int)b];
	/*
	int first=-1,second=-1;	

	if(tolower(a) == 'a')
		first = 0;
	else if(tolower(a) == 'c')
		first = 1;
	else if(tolower(a) == 't')
		first = 2;
	else if(tolower(a) == 'g')
		first = 3;
	else
		first = 4;

	if(tolower(b) == 'a')
		second = 0;
	else if(tolower(b) == 'c')
		second = 1;
	else if(tolower(b) == 't')
		second = 2;
	else if(tolower(b) == 'g')
		second = 3;
	else 
		second = 4;

	return gALIGN_SCORES[first][second];
	*/
}

float bin_seq::max_flt(char &path, float diag, float upgap, float leftgap) {
	//cout << diag << ' '<< upgap << ' ' << c << endl;
	if(diag >= upgap) {
		if(diag >= leftgap) {
			path = 'D';
			return diag;
		}
		else {
			path = 'L';
			return leftgap;	//implies leftgap>diag, which means leftgap>upgap.
		}
	}
	else {		//we know know upgap>diag.
		if(upgap >= leftgap) {
			path = 'U';
			return upgap;
		}
		else {
			path = 'L';
			return leftgap;
		}
	}
}

inline unsigned int bin_seq::max_flt(float array[], unsigned int array_size) {
	float* pos = max_element(array, array+array_size);
	unsigned int max_pos = pos - &array[0];
	return max_pos;
}

inline float bin_seq::max_flt(float a, float b, float c) {
	//cout << a << ' '<< b << ' ' << c << endl;
	if(a >= b) {
		if(a >= c)
			return a;
		else return c;	//implies c>a, which means c>b.
	}
	else {		//we know know b>a.
		if(b >= c)
			return b;
		else
			return c;
	}
}

inline int bin_seq::max(int a, int b, int c) {
	//cout << a << ' '<< b << ' ' << c << endl;
	if(a >= b) {
		if(a >= c)
			return a;
		else return c;
	}
	else {
		if(b >= c)
			return b;
		else
			return c;
	}
}

bool bin_seq::Test(ostream &os) {
	bin_seq bs;
	bool success = true;
	
	Read r1;
	r1.length = 35;
	string consense = "acgtcgatcgtggctaatcgttcgtagatcgatta";
	string gen_str =  "acgtcgatcgtggctaatcgttgtagatccgatta";
	r1.pwm = new float*[35];
	for(unsigned int i=0; i<35; i++) {
		r1.pwm[i] = new float[4];
	}
	unsigned int row = 0;
	r1.pwm[row][0] = 1.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 1.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 1.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 1.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 1.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 1.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 1.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 1.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 1.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 1.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 1.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 1.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 1.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 1.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 1.0; row++;
	r1.pwm[row][0] = 1.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 1.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 1.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 1.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 1.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 1.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 1.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 1.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 1.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 1.0; row++;
	r1.pwm[row][0] = 1.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 1.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 1.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 1.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 1.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 1.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 1.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 0.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 1.0; row++;
	r1.pwm[row][0] = 0.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 1.0; row++;
	r1.pwm[row][0] = 1.0; r1.pwm[row][1] = 0.0; r1.pwm[row][2] = 0.0; r1.pwm[row][3] = 0.0; row++;
	
	pair<string,string> result = bs.get_align_score_w_traceback(r1, consense, gen_str);
	printf("CONSENSE: %s\nGENOME:   %s\nRESULT:   %s\n          %s\n",
		consense.c_str(),gen_str.c_str(),result.first.c_str(),result.second.c_str());

	TEST(result.first == "acgtcgatcgtggctaatcgttggtagat-cgatta");
	//TEST(result.second == "acgtcgatcgtggctaatcgtt-gtagatccgatta");
	TEST(result.second == "I22D29");

	//consense = "acgtcgatcgtggctaatcgttcgtagatcgatta";
	   gen_str = "acgtcgatcgtggctaatcgttcggagatcgatta";
	//gen_str =  "acgtcgtttatcgtggctaatcgttccgattaccc";
	// Check timing on this alignment
	for(unsigned int i=0; i<1000; i++) {
		result = bs.get_align_score_w_traceback(r1, consense, gen_str);
	}
	printf("CONSENSE: %s\nGENOME:   %s\nRESULT:   %s\n CIGAR:   %s\n",
		consense.c_str(),gen_str.c_str(),result.first.c_str(),result.second.c_str());
	TEST(result.first == "acgtcg---atcgtggctaatcgttctaagatggatta");
	TEST(result.second == "D7D8D9I26I26M27I30M31M33M34M35");

	/************PHMM****************/
	float** phmm = bs.pairHMM(r1, consense, gen_str);
	for(unsigned int i=0; i<r1.length; i++) {
		printf("[%u, %c]\t",i,gen_str[i]);
		for(int j=0; j<5; j++) {
			printf("%3.2f\t",phmm[i][j]);
		}
		printf("\n");
		delete[] phmm[i];
	}

	printf("\n\n");

   gen_str = "acgtcgatcgtggctaatcgttcgagatcgattaa";
	phmm = bs.pairHMM(r1, consense, gen_str);
	for(unsigned int i=0; i<r1.length; i++) {
		printf("[%u, %c]\t",i,gen_str[i]);
		for(int j=0; j<5; j++) {
			printf("%3.2f\t",phmm[i][j]);
		}
		printf("\n");
		delete[] phmm[i];
	}


	// Clean up after ourselves
	for(unsigned int i=0; i<35; i++) {
		delete[] r1.pwm[i];
	}
	delete[] r1.pwm;
	
	
	unsigned int GEN_SIZE = 100000;
	unsigned char* genome = new unsigned char[GEN_SIZE];

	ifstream in;	
	in.open("test/short.fa");
	in.seekg(7);
	in.read((char*)genome, GEN_SIZE);
	
	string dna = "aaaagaaag";
	string dna2 = "aaaaaaaaa";
	string dnat = "ttttttttt";

	unsigned int bs2_t_good = 6285;
	unsigned int bs2_t_bad = 33;
	unsigned int good = 7057;
	bin_seq bs2;
	
	//string loads = "ccccccccc";
	//cout << "highest " <<  bs.str2int(loads) << endl;
	
	unsigned int temp_MER = gMER_SIZE;
	gMER_SIZE = 11;
	const char* test_genome = "aaaaaaaaaataaaatNtttt";
	unsigned int pos = 0;
	TEST(bs.get_hash(pos,(unsigned char*)test_genome,21u).second == 3);
	TEST(bs.hash2str(3,11) == "aaaaaaaaaat");
	os << bs.hash2str(3,11) << endl;
	
	gMER_SIZE = temp_MER;
	TEST(!bs.get_hash(bs2_t_bad,genome).first);
	TEST(bs.get_hash(bs2_t_good,genome).first);
	
	TEST(bs.get_hash(good,genome).second == 65712);
	TEST(bs.str2int(dna) == 514);
	//TEST(bs2.get_hash(dna) == bs.get_hash(bs2_t_good,genome));
	cout << "get hash: " << bs2.get_hash(bs2_t_good,genome).second << endl;
	cout << "dna: " << bs2.get_hash(dna).second << endl;
	TEST(bs.hash2str(514,9) == dna);
	TEST(bs.hash2str(0,9) == dna2);
	TEST(bs.hash2str(262143,9) == dnat);
	TEST(bs.get_hash(dnat).second == 262143);
	TEST(bs.get_hash("caaaagtaa").second == 65712);
	TEST(bs.get_hash("atcgccaga").second == 55624);
	
	//Used for testing the N-W alignement
	TEST(bs.max(1,2,3) == 3);
	TEST(bs.max(1,3,2) == 3);
	TEST(bs.max(3,2,1) == 3);
	TEST(bs.max(3,2,3) == 3);
	
	delete[] genome;
	in.close();
	return success;
}
